@@ -1,6 +1,6 @@
 /*
  * mod_ldap - LDAP password lookup module for ProFTPD
- * Copyright (c) 1999, 2000-7, John Morrissey <jwm@horde.net>
+ * Copyright (c) 1999, 2000-9, John Morrissey <jwm@horde.net>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -22,7 +22,7 @@
  */
 
 /*
- * mod_ldap v2.8.17
+ * mod_ldap v2.8.20-20090124
  *
  * Thanks for patches go to (in alphabetical order):
  *
@@ -48,7 +48,7 @@
  *                                                   LDAPDefaultAuthScheme
  *
  *
- * $Id: mod_ldap.c,v 1.58 2007/09/07 16:13:41 jwm Exp $
+ * $Id: mod_ldap.c,v 1.75 2009/01/25 03:24:17 jwm Exp $
  * $Libraries: -lldap -llber$
  */
 
@@ -56,21 +56,10 @@
  * with the --enable-openssl argument to configure.
  */
 
-/* Uncomment this to use LDAP TLS. If enabled, we will try to enable TLS
- * after connecting to the LDAP server. If TLS cannot be enabled, the LDAP
- * connection will fail.
- */
-/* #define USE_LDAP_TLS */
-
-/*
- * If you have to edit anything below this line, it's a bug. Report it
- * at http://bugs.proftpd.org/.
- */
-
 #include "conf.h"
 #include "privs.h"
 
-#define MOD_LDAP_VERSION	"mod_ldap/2.8.17"
+#define MOD_LDAP_VERSION	"mod_ldap/2.8.20-20090124"
 
 #if PROFTPD_VERSION_NUMBER < 0x0001021002
 # error MOD_LDAP_VERSION " requires ProFTPD 1.2.10rc2 or later"
@@ -82,10 +71,6 @@
 
 #include <errno.h>
 #include <ctype.h>     /* isdigit()   */
-#include <stdio.h>     /* snprintf()  */
-#include <string.h>    /* various :-) */
-#include <sys/types.h> /* seteuid()   */
-#include <unistd.h>    /* seteuid()   */
 
 #include <lber.h>
 #include <ldap.h>
@@ -100,13 +85,16 @@
 # define LDAP_SEARCH(ld, base, scope, filter, attrs, timeout, sizelimit, res) \
    ldap_search_ext_s(ld, base, scope, filter, attrs, 0, NULL, NULL, \
                      timeout, sizelimit, res)
-#else
+#else /* LDAP_API_VERSION >= 2000 */
+static char *ldap_server;
+static int ldap_port = LDAP_PORT;
+
 # define LDAP_VALUE_T char
 # define LDAP_GET_VALUES(ld, entry, attr) ldap_get_values(ld, entry, attr)
 # define LDAP_VALUE(values, i) (values[i])
 # define LDAP_COUNT_VALUES(values) (ldap_count_values(values))
 # define LDAP_VALUE_FREE(values) (ldap_value_free(values))
-# define LDAP_UNBIND(ld) (ldap_unbind_s(ld, NULL, NULL))
+# define LDAP_UNBIND(ld) (ldap_unbind_s(ld))
 
 static void
 pr_ldap_set_sizelimit(LDAP *limit_ld, int limit)
@@ -123,48 +111,29 @@ pr_ldap_set_sizelimit(LDAP *limit_ld, int limit)
   pr_log_debug(DEBUG3, MOD_LDAP_VERSION ": set search size limit to %d", limit);
 }
 
-# define LDAP_SEARCH(ld, base, scope, filter, attrs, timeout, sizelimit, res) \
-   { \
-     pr_ldap_set_sizelimit(sizelimit);
-     ldap_search_s(ld, base, scope, filter, attrs, 0, res); \
-   }
-#endif
+static int
+LDAP_SEARCH(LDAP *ld, char *base, int scope, char *filter, char *attrs[],
+            struct timeval *timeout, int sizelimit, LDAPMessage **res)
+{
+  pr_ldap_set_sizelimit(ld, sizelimit);
+  return ldap_search_st(ld, base, scope, filter, attrs, 0, timeout, res);
+}
+#endif /* LDAP_API_VERSION >= 2000 */
 
 /* Thanks, Sun. */
 #ifndef LDAP_OPT_SUCCESS
 # define LDAP_OPT_SUCCESS LDAP_SUCCESS
 #endif
+#ifndef LDAP_SCOPE_DEFAULT
+# define LDAP_SCOPE_DEFAULT LDAP_SCOPE_SUBTREE
+#endif
 
 #if defined(HAVE_OPENSSL) || defined(PR_USE_OPENSSL)
 # include <openssl/evp.h>
 #endif
 
-#define HASH_TABLE_SIZE 10
-
-/* FIXME: does proftpd proper have uesr/group caching? remove ours if so? */
-typedef union pr_idauth {
-  uid_t uid;
-  gid_t gid;
-} pr_idauth_t;
-
-typedef struct _idmap {
-  struct _idmap *next, *prev;
-
-  /* This is a union because different OSs may give different types/sizes to
-   * UIDs and GIDs. This presents a far more portable way to deal with this
-   * reality.
-   */
-  pr_idauth_t id;
-
-  char *name;                  /* user or group name */
-  unsigned short int negative; /* have we gotten a negative answer before? */
-} pr_idmap_t;
-
-static xaset_t *uid_table[HASH_TABLE_SIZE];
-static xaset_t *gid_table[HASH_TABLE_SIZE];
-
 /* Config entries */
-static char *ldap_server, *ldap_dn, *ldap_dnpass,
+static char *ldap_server_url, *ldap_dn, *ldap_dnpass,
             *ldap_auth_filter, *ldap_uid_filter,
             *ldap_group_gid_filter, *ldap_group_name_filter,
             *ldap_group_member_filter, *ldap_quota_filter,
@@ -181,10 +150,9 @@ static char *ldap_server, *ldap_dn, *ldap_dnpass,
             *ldap_attr_cn = "cn",
             *ldap_attr_memberuid = "memberUid",
             *ldap_attr_ftpquota = "ftpQuota";
-static int ldap_port = LDAP_PORT,
-           ldap_doauth = 0, ldap_douid = 0, ldap_dogid = 0, ldap_doquota = 0,
-           ldap_authbinds = 1, ldap_negcache = 1,
-           ldap_querytimeout = 0, ldap_genhdir = 0, ldap_genhdir_prefix_nouname = 0,
+static int ldap_doauth = 0, ldap_douid = 0, ldap_dogid = 0, ldap_doquota = 0,
+           ldap_authbinds = 1, ldap_querytimeout = 0,
+           ldap_genhdir = 0, ldap_genhdir_prefix_nouname = 0,
            ldap_forcedefaultuid = 0, ldap_forcedefaultgid = 0,
            ldap_forcegenhdir = 0, ldap_protocol_version = 3,
            ldap_dereference = LDAP_DEREF_NEVER,
@@ -194,7 +162,7 @@ static struct timeval ldap_querytimeout_tp;
 static uid_t ldap_defaultuid = -1;
 static gid_t ldap_defaultgid = -1;
 
-#ifdef USE_LDAP_TLS
+#ifdef LDAP_OPT_X_TLS
 static int ldap_use_tls = 0;
 #endif
 #ifdef LDAP_OPT_X_TLS_HARD
@@ -202,19 +170,9 @@ static int ldap_use_ssl = 0;
 #endif
 
 static LDAP *ld = NULL;
-static struct passwd *pw = NULL;
-static struct group *gr = NULL;
 array_header *cached_quota = NULL;
 
 
-static int
-pr_ldap_module_init(void)
-{
-  memset(uid_table, 0, sizeof(uid_table));
-  memset(gid_table, 0, sizeof(gid_table));
-  return 0;
-}
-
 static void
 pr_ldap_unbind(void)
 {
@@ -235,7 +193,7 @@ pr_ldap_unbind(void)
 }
 
 static int
-pr_ldap_connect(LDAP **conn_ld, int bind)
+pr_ldap_connect(LDAP **conn_ld, int do_bind)
 {
   int ret, version;
 #ifdef LDAP_OPT_X_TLS_HARD
@@ -245,12 +203,23 @@ pr_ldap_connect(LDAP **conn_ld, int bind)
   struct berval bindcred;
 #endif
 
-  *conn_ld = ldap_init(ldap_server, ldap_port);
-  if (!conn_ld) {
-    pr_log_pri(PR_LOG_ERR, MOD_LDAP_VERSION ": pr_ldap_connect(): ldap_init() to %s failed: %s", ldap_server, strerror(errno));
+#if LDAP_API_VERSION >= 2000
+  ret = ldap_initialize(conn_ld, ldap_server_url);
+  if (ret != LDAP_SUCCESS) {
+    pr_log_pri(PR_LOG_ERR, MOD_LDAP_VERSION ": pr_ldap_connect(): ldap_initialize() to %s failed: %s", ldap_server_url, ldap_err2string(ret));
+    *conn_ld = NULL;
     return -1;
   }
-  pr_log_debug(DEBUG3, MOD_LDAP_VERSION ": connected to %s:%d", ldap_server, LDAP_PORT);
+  pr_log_debug(DEBUG3, MOD_LDAP_VERSION ": connected to %s", ldap_server_url);
+#else /* LDAP_API_VERSION >= 2000 */
+  *conn_ld = ldap_init(ldap_server, ldap_port);
+  if (!conn_ld) {
+    pr_log_pri(PR_LOG_ERR, MOD_LDAP_VERSION ": pr_ldap_connect(): ldap_init() to %s:%d failed: %s", ldap_server, ldap_port, strerror(errno));
+    return -1;
+  }
+  pr_log_debug(DEBUG3, MOD_LDAP_VERSION ": connected to %s:%d", ldap_server, ldap_port);
+#endif
+
 
   version = -1;
   switch (ldap_protocol_version) {
@@ -283,7 +252,7 @@ pr_ldap_connect(LDAP **conn_ld, int bind)
   }
 #endif /* LDAP_OPT_X_TLS_HARD */
 
-#ifdef USE_LDAP_TLS
+#ifdef LDAP_OPT_X_TLS
   if (ldap_use_tls == 1) {
     ret = ldap_start_tls_s(*conn_ld, NULL, NULL);
     if (ret != LDAP_SUCCESS) {
@@ -293,9 +262,9 @@ pr_ldap_connect(LDAP **conn_ld, int bind)
     }
     pr_log_debug(DEBUG3, MOD_LDAP_VERSION ": enabled TLS.");
   }
-#endif /* USE_LDAP_TLS */
+#endif /* LDAP_OPT_X_TLS */
 
-  if (bind == TRUE) {
+  if (do_bind == TRUE) {
 #if LDAP_API_VERSION >= 2000
     bindcred.bv_val = ldap_dnpass;
     bindcred.bv_len = ldap_dnpass != NULL ? strlen(ldap_dnpass) : 0;
@@ -370,6 +339,7 @@ pr_ldap_user_lookup(pool *p,
 {
   char *filter, *dn;
   int i = 0, ret;
+  struct passwd *pw;
   LDAPMessage *result, *e;
   LDAP_VALUE_T **values;
 
@@ -430,12 +400,7 @@ pr_ldap_user_lookup(pool *p,
     return NULL; /* No LDAP entries for this user */
   }
 
-  if (!pw) {
-    pw = pcalloc(session.pool, sizeof(struct passwd));
-  } else {
-    memset(pw, '\0', sizeof(struct passwd));
-  }
-
+  pw = pcalloc(session.pool, sizeof(struct passwd));
   while (ldap_attrs[i] != NULL) {
     pr_log_debug(DEBUG3, MOD_LDAP_VERSION ": fetching value(s) for attr %s", ldap_attrs[i]);
 
@@ -596,6 +561,7 @@ pr_ldap_group_lookup(pool *p,
 {
   char *filter, *dn;
   int i = 0, value_count, value_offset, ret;
+  struct group *gr;
   LDAPMessage *result, *e;
   LDAP_VALUE_T **values;
 
@@ -647,12 +613,7 @@ pr_ldap_group_lookup(pool *p,
     return NULL; /* No LDAP entries for this user */
   }
 
-  if (!gr) {
-    gr = pcalloc(session.pool, sizeof(struct group));
-  } else {
-    memset(gr, '\0', sizeof(struct group));
-  }
-
+  gr = pcalloc(session.pool, sizeof(struct group));
   while (ldap_attrs[i] != NULL) {
     pr_log_debug(DEBUG3, MOD_LDAP_VERSION ": fetching value(s) for attr %s", ldap_attrs[i]);
 
@@ -816,7 +777,8 @@ pr_ldap_quota_lookup(pool *p, char *filter_template, const char *replace,
 static struct group *
 pr_ldap_getgrnam(pool *p, const char *group_name)
 {
-  char *group_attrs[] = {ldap_attr_cn, ldap_attr_gidnumber, ldap_attr_memberuid, NULL};
+  char *group_attrs[] = {ldap_attr_cn, ldap_attr_gidnumber,
+                         ldap_attr_memberuid, NULL};
 
   return pr_ldap_group_lookup(p, ldap_group_name_filter,
     group_name, group_attrs);
@@ -885,77 +847,6 @@ pr_ldap_getpwuid(pool *p, uid_t uid)
     ldap_uid_basedn, uid_attrs, ldap_authbinds ? &ldap_authbind_dn : NULL);
 }
 
-static int
-_compare_uid(pr_idmap_t *m1, pr_idmap_t *m2)
-{
-  if (m1->id.uid < m2->id.uid) {
-    return -1;
-  }
-
-  if (m1->id.uid > m2->id.uid) {
-    return 1;
-  }
-
-  return 0;
-}
-
-static int
-_compare_gid(pr_idmap_t *m1, pr_idmap_t *m2)
-{
-  if (m1->id.gid < m2->id.gid) {
-    return -1;
-  }
-
-  if (m1->id.gid > m2->id.gid) {
-    return 1;
-  }
-
-  return 0;
-}
-
-static int
-_compare_id(xaset_t **table, pr_idauth_t id, pr_idauth_t idcomp)
-{
-  if (table == uid_table) {
-    return id.uid == idcomp.uid;
-  }
-  return id.gid == idcomp.gid;
-}
-
-static pr_idmap_t *
-_auth_lookup_id(xaset_t **id_table, pr_idauth_t id)
-{
-  int hash = ((id_table == uid_table) ? id.uid : id.gid) % HASH_TABLE_SIZE;
-  pr_idmap_t *m;
-
-  if (!id_table[hash]) {
-    id_table[hash] = xaset_create(permanent_pool,
-      (id_table == uid_table) ? (XASET_COMPARE) _compare_uid :
-                                (XASET_COMPARE) _compare_gid);
-  }
-
-  for (m = (pr_idmap_t *) id_table[hash]->xas_list; m; m = m->next) {
-    if (_compare_id(id_table, m->id, id)) {
-      break;
-    }
-  }
-
-  if (!m || !_compare_id(id_table, m->id, id)) {
-    /* Isn't in the table. */
-    m = (pr_idmap_t *) pcalloc(id_table[hash]->pool, sizeof(pr_idmap_t));
-
-    if (id_table == uid_table) {
-      m->id.uid = id.uid;
-    } else {
-      m->id.gid = id.gid;
-    }
-
-    xaset_insert_sort(id_table[hash], (xasetmember_t *) m, FALSE);
-  }
-
-  return m;
-}
-
 MODRET
 handle_ldap_quota_lookup(cmd_rec *cmd)
 {
@@ -992,8 +883,6 @@ handle_ldap_endpwent(cmd_rec *cmd)
 {
   if (ldap_doauth || ldap_douid || ldap_dogid) {
     pr_ldap_unbind();
-    pw = NULL;
-    gr = NULL;
     return PR_HANDLED(cmd);
   }
 
@@ -1003,11 +892,13 @@ handle_ldap_endpwent(cmd_rec *cmd)
 MODRET
 handle_ldap_getpwuid(cmd_rec *cmd)
 {
+  struct passwd *pw;
+
   if (!ldap_douid) {
     return PR_DECLINED(cmd);
   }
 
-  pw = pr_ldap_getpwuid(cmd->tmp_pool, (uid_t)cmd->argv[0]);
+  pw = pr_ldap_getpwuid(cmd->tmp_pool, *((uid_t *) cmd->argv[0]));
   if (pw) {
     return mod_create_data(cmd, pw);
   }
@@ -1018,15 +909,12 @@ handle_ldap_getpwuid(cmd_rec *cmd)
 MODRET
 handle_ldap_getpwnam(cmd_rec *cmd)
 {
+  struct passwd *pw;
+
   if (!ldap_doauth) {
     return PR_DECLINED(cmd);
   }
 
-  if (pw && pw->pw_name && strcasecmp(pw->pw_name, cmd->argv[0]) == 0) {
-    pr_log_debug(DEBUG3, MOD_LDAP_VERSION ": getpwnam: returning cached data for %s", pw->pw_name);
-    return mod_create_data(cmd, pw);
-  }
-
   pw = pr_ldap_getpwnam(cmd->tmp_pool, cmd->argv[0]);
   if (pw) {
     return mod_create_data(cmd, pw);
@@ -1038,14 +926,12 @@ handle_ldap_getpwnam(cmd_rec *cmd)
 MODRET
 handle_ldap_getgrnam(cmd_rec *cmd)
 {
+  struct group *gr;
+
   if (!ldap_dogid) {
     return PR_DECLINED(cmd);
   }
 
-  if (gr && strcasecmp(gr->gr_name, cmd->argv[0]) == 0) {
-    pr_log_debug(DEBUG3, MOD_LDAP_VERSION ": getgrnam: returning cached data for %s", gr->gr_name);
-    return mod_create_data(cmd, gr);
-  }
   gr = pr_ldap_getgrnam(cmd->tmp_pool, cmd->argv[0]);
   if (gr) {
     return mod_create_data(cmd, gr);
@@ -1057,20 +943,18 @@ handle_ldap_getgrnam(cmd_rec *cmd)
 MODRET
 handle_ldap_getgrgid(cmd_rec *cmd)
 {
+  struct group *gr;
+
   if (!ldap_dogid) {
     return PR_DECLINED(cmd);
   }
 
-  if (gr && gr->gr_gid == (gid_t)cmd->argv[0]) {
-    pr_log_debug(DEBUG3, MOD_LDAP_VERSION ": getgrgid: returning cached data for %lu", (unsigned long)gr->gr_gid);
-    return mod_create_data(cmd, gr);
-  }
-  gr = pr_ldap_getgrgid(cmd->tmp_pool, (gid_t)cmd->argv[0]);
-  if (gr) {
-    return mod_create_data(cmd, gr);
+  gr = pr_ldap_getgrgid(cmd->tmp_pool, *((gid_t *) cmd->argv[0]));
+  if (!gr) {
+    return PR_DECLINED(cmd);
   }
 
-  return PR_DECLINED(cmd);
+  return mod_create_data(cmd, gr);
 }
 
 MODRET
@@ -1198,7 +1082,10 @@ MODRET
 handle_ldap_is_auth(cmd_rec *cmd)
 {
   const char *username = cmd->argv[0];
-  char *pass_attrs[] = {ldap_attr_userpassword, ldap_attr_homedirectory, NULL};
+  char *pass_attrs[] = {ldap_attr_userpassword, ldap_attr_uid,
+                        ldap_attr_uidnumber, ldap_attr_gidnumber,
+                        ldap_attr_homedirectory, ldap_attr_loginshell, NULL};
+  struct passwd *pw;
 
   if (!ldap_doauth) {
     return PR_DECLINED(cmd);
@@ -1212,19 +1099,12 @@ handle_ldap_is_auth(cmd_rec *cmd)
    * this way, then by all means, let me know.
    */
 
-  /* If we don't have a cached entry, or if the cached entry isn't for this
-   * user, fetch the entry.
-   */
-  if (!pw || (pw && pw->pw_name && strcasecmp(pw->pw_name, username) != 0)) {
-    if ((pw = pr_ldap_user_lookup(cmd->tmp_pool, ldap_auth_filter, username,
-                                  pr_ldap_generate_filter(cmd->tmp_pool, ldap_auth_basedn, username),
-                                  ldap_authbinds ? pass_attrs + 1 : pass_attrs,
-                                  ldap_authbinds ? &ldap_authbind_dn : NULL)) == NULL)
-    {
-      return PR_DECLINED(cmd); /* Can't find the user in the LDAP directory. */
-    }
-  } else {
-    pr_log_debug(DEBUG3, MOD_LDAP_VERSION ": auth: using cached data for %s", pw->pw_name);
+  pw = pr_ldap_user_lookup(cmd->tmp_pool, ldap_auth_filter, username,
+    pr_ldap_generate_filter(cmd->tmp_pool, ldap_auth_basedn, username),
+    ldap_authbinds ? pass_attrs + 1 : pass_attrs,
+    ldap_authbinds ? &ldap_authbind_dn : NULL);
+  if (!pw) {
+    return PR_DECLINED(cmd); /* Can't find the user in the LDAP directory. */
   }
 
   if (!ldap_authbinds && !pw->pw_passwd) {
@@ -1269,10 +1149,16 @@ handle_ldap_check(cmd_rec *cmd)
 #if defined(HAVE_OPENSSL) || defined(PR_USE_OPENSSL)
   EVP_MD_CTX EVP_Context;
   const EVP_MD *md;
-  int md_len;
+  unsigned int md_len;
   unsigned char md_value[EVP_MAX_MD_SIZE];
   EVP_ENCODE_CTX EVP_Encode;
-  char buff[EVP_MAX_KEY_LENGTH];
+
+  /* According to RATS, the output buffer (buff) for EVP_EncodeBlock() needs
+   * to be 4/3 the size of the input buffer (md_val).  Let's make it easy, and
+   * use an output buffer that's twice the size of the input buffer.
+   */
+  unsigned char buff[EVP_MAX_MD_SIZE * 2];
+
 #endif /* !HAVE_OPENSSL and !PR_USE_OPENSSL */
 
   if (!ldap_doauth) {
@@ -1393,10 +1279,11 @@ handle_ldap_check(cmd_rec *cmd)
     EVP_DigestFinal(&EVP_Context, md_value, &md_len);
 
     /* Base64 Encoding */
+    memset(buff, '\0', sizeof(buff));
     EVP_EncodeInit(&EVP_Encode);
-    EVP_EncodeBlock(buff, md_value, md_len);
+    EVP_EncodeBlock(buff, md_value, (int) md_len);
 
-    if (strcmp(buff, cryptpass + encname_len + 2) != 0) {
+    if (strcmp((char *) buff, cryptpass + encname_len + 2) != 0) {
       return PR_ERROR(cmd);
     }
   }
@@ -1413,117 +1300,71 @@ handle_ldap_check(cmd_rec *cmd)
 MODRET
 handle_ldap_uid_name(cmd_rec *cmd)
 {
-  pr_idmap_t *m;
-  pr_idauth_t id;
+  struct passwd *pw;
 
   if (!ldap_douid) {
     return PR_DECLINED(cmd);
   }
 
-  id.uid = *((uid_t *) cmd->argv[0]);
-  m = _auth_lookup_id(uid_table, id);
-
-  if (m->name) {
-    pr_log_debug(DEBUG3, MOD_LDAP_VERSION ": uid2name: returning cached data for %lu (%s)", (unsigned long)id.uid, m->name);
-  } else {
-    if (ldap_negcache && m->negative) {
-      return PR_DECLINED(cmd);
-    }
-
-    /* Wasn't cached and we've haven't seen this one, so perform a lookup.
-     * If we don't have a cached entry, or if the cached entry isn't for
-     * this user, fetch the entry.
-     */
-    if (!pw || pw->pw_uid != id.uid) {
-      if (! (pw = pr_ldap_getpwuid(cmd->tmp_pool, id.uid))) {
-        if (ldap_negcache) {
-          m->negative = 1;
-        }
-        return PR_DECLINED(cmd); /* Can't find the user in the LDAP directory. */
-      }
-    }
-
-    m->name = pstrdup(permanent_pool, pw->pw_name);
+  pw = pr_ldap_getpwuid(cmd->tmp_pool, *((uid_t *) cmd->argv[0]));
+  if (!pw) {
+    /* Can't find the user in the LDAP directory. */
+    return PR_DECLINED(cmd);
   }
 
-  return mod_create_data(cmd, m->name);
+  return mod_create_data(cmd, pstrdup(permanent_pool, pw->pw_name));
 }
 
 MODRET
 handle_ldap_gid_name(cmd_rec *cmd)
 {
-  pr_idmap_t *m;
-  pr_idauth_t id;
+  struct group *gr;
 
   if (!ldap_dogid) {
     return PR_DECLINED(cmd);
   }
 
-  id.gid = *((gid_t *) cmd->argv[0]);
-  m = _auth_lookup_id(gid_table, id);
-
-  if (m->name) {
-    pr_log_debug(DEBUG3, MOD_LDAP_VERSION ": uid2name: returning cached data for %lu (%s)", (unsigned long)id.gid, m->name);
-  } else {
-    if (ldap_negcache && m->negative) {
-      return PR_DECLINED(cmd);
-    }
-
-    /* Wasn't cached and we've haven't seen this one, so perform a lookup.
-     * If we don't have a cached entry, or if the cached entry isn't for
-     * this group, fetch the entry.
-     */
-    if (!gr || gr->gr_gid != id.gid) {
-      if (! (gr = pr_ldap_getgrgid(cmd->tmp_pool, id.gid))) {
-        if (ldap_negcache) {
-          m->negative = 1;
-        }
-        return PR_DECLINED(cmd); /* Can't find the user in the LDAP directory. */
-      }
-    }
-
-    m->name = pstrdup(permanent_pool, gr->gr_name);
+  gr = pr_ldap_getgrgid(cmd->tmp_pool, *((gid_t *) cmd->argv[0]));
+  if (!gr) {
+    /* Can't find the user in the LDAP directory. */
+    return PR_DECLINED(cmd);
   }
 
-  return mod_create_data(cmd, m->name);
+  return mod_create_data(cmd, pstrdup(permanent_pool, gr->gr_name));
 }
 
 MODRET
 handle_ldap_name_uid(cmd_rec *cmd)
 {
+  struct passwd *pw;
+
   if (!ldap_doauth) {
     return PR_DECLINED(cmd);
   }
 
-  if (pw && pw->pw_name && strcasecmp(pw->pw_name, cmd->argv[0]) == 0) {
-    pr_log_debug(DEBUG3, MOD_LDAP_VERSION ": name2uid: returning cached data for %s", pw->pw_name);
-    return mod_create_data(cmd, (void *) &pw->pw_uid);
-  }
   pw = pr_ldap_getpwnam(cmd->tmp_pool, cmd->argv[0]);
-  if (pw) {
-    return mod_create_data(cmd, (void *) &pw->pw_uid);
+  if (!pw) {
+    return PR_DECLINED(cmd);
   }
 
-  return PR_DECLINED(cmd);
+  return mod_create_data(cmd, (void *) &pw->pw_uid);
 }
 
 MODRET
 handle_ldap_name_gid(cmd_rec *cmd)
 {
+  struct group *gr;
+
   if (!ldap_dogid) {
     return PR_DECLINED(cmd);
   }
 
-  if (gr && strcasecmp(gr->gr_name, cmd->argv[0]) == 0) {
-    pr_log_debug(DEBUG3, MOD_LDAP_VERSION ": name2gid: returning cached data for %s", gr->gr_name);
-    return mod_create_data(cmd, (void *) &gr->gr_gid);
-  }
   gr = pr_ldap_getgrnam(cmd->tmp_pool, cmd->argv[0]);
-  if (gr) {
-    return mod_create_data(cmd, (void *) &gr->gr_gid);
+  if (!gr) {
+    return PR_DECLINED(cmd);
   }
 
-  return PR_DECLINED(cmd);
+  return mod_create_data(cmd, (void *) &gr->gr_gid);
 }
 
 
@@ -1544,6 +1385,13 @@ set_ldap_server(cmd_rec *cmd)
       CONF_ERROR(cmd, "LDAPServer: must be supplied with a valid LDAP URL.");
     }
 
+    if (find_config(main_server->conf, CONF_PARAM, "LDAPSearchScope", FALSE)) {
+      CONF_ERROR(cmd, "LDAPSearchScope cannot be used when LDAPServer specifies a URL; specify a search scope in the LDAPServer URL instead.");
+    }
+    if (find_config(main_server->conf, CONF_PARAM, "LDAPUseSSL", FALSE)) {
+      CONF_ERROR(cmd, "LDAPUseSSL cannot be used when LDAPServer specifies a URL; specify the desired scheme (ldap:// or ldaps://) in the LDAPServer URL instead.");
+    }
+
 #ifdef LDAP_OPT_X_TLS_HARD
     if (strncasecmp(cmd->argv[1], "ldap:", strlen("ldap:")) != 0 &&
         strncasecmp(cmd->argv[1], "ldaps:", strlen("ldaps:")) != 0) {
@@ -1560,14 +1408,13 @@ set_ldap_server(cmd_rec *cmd)
       CONF_ERROR(cmd, "A base DN may not be specified by an LDAPServer URL, only by LDAPDoAuth, LDAPDoUIDLookups, LDAPDoGIDLookups, or LDAPDoQuotaLookups.");
     }
     if (url->lud_filter && strcmp(url->lud_filter, "") != 0) {
-      CONF_ERROR(cmd, "A base DN may not be specified by an LDAPServer URL, only by LDAPDoAuth, LDAPDoUIDLookups, LDAPDoGIDLookups, or LDAPDoQuotaLookups.");
+      CONF_ERROR(cmd, "A search filter may not be specified by an LDAPServer URL, only by LDAPDoAuth, LDAPDoUIDLookups, LDAPDoGIDLookups, or LDAPDoQuotaLookups.");
     }
 
-    add_config_param_str(cmd->argv[0], 2, "url", url);
-  } else {
-    add_config_param_str(cmd->argv[0], 2, "host", cmd->argv[1]);
+    ldap_free_urldesc(url);
   }
 
+  add_config_param_str(cmd->argv[0], 1, cmd->argv[1]);
   return PR_HANDLED(cmd);
 }
 
@@ -1625,9 +1472,16 @@ set_ldap_querytimeout(cmd_rec *cmd)
 MODRET
 set_ldap_searchscope(cmd_rec *cmd)
 {
+  config_rec *c;
+
   CHECK_ARGS(cmd, 1);
   CHECK_CONF(cmd, CONF_ROOT|CONF_VIRTUAL|CONF_GLOBAL);
 
+  c = find_config(main_server->conf, CONF_PARAM, "LDAPServer", FALSE);
+  if (c && ldap_is_ldap_url(c->argv[0])) {
+    CONF_ERROR(cmd, "LDAPSearchScope cannot be used when LDAPServer specifies a URL; specify a search scope in the LDAPServer URL instead.");
+  }
+
   add_config_param_str(cmd->argv[0], 1, cmd->argv[1]);
   return PR_HANDLED(cmd);
 }
@@ -1854,19 +1708,7 @@ MODRET set_ldap_forcedefaultgid(cmd_rec *cmd)
 MODRET
 set_ldap_negcache(cmd_rec *cmd)
 {
-  int b;
-  config_rec *c;
-
-  CHECK_ARGS(cmd, 1);
-  CHECK_CONF(cmd, CONF_ROOT|CONF_VIRTUAL|CONF_GLOBAL);
-
-  if ((b = get_boolean(cmd, 1)) == -1) {
-    CONF_ERROR(cmd, "LDAPNegativeCache: expected a boolean value for first argument.");
-  }
-
-  c = add_config_param(cmd->argv[0], 1, NULL);
-  c->argv[0] = pcalloc(c->pool, sizeof(int));
-  *((int *) c->argv[0]) = b;
+  pr_log_pri(PR_LOG_ERR, MOD_LDAP_VERSION ": LDAPNegativeCache has no effect and is deprecated; please remove it from your configuration");
   return PR_HANDLED(cmd);
 }
 
@@ -1949,9 +1791,9 @@ set_ldap_defaultauthscheme(cmd_rec *cmd)
 MODRET
 set_ldap_usetls(cmd_rec *cmd)
 {
-#ifndef USE_LDAP_TLS
-  CONF_ERROR(cmd, "LDAPUseTLS: You must edit mod_ldap.c and recompile with USE_LDAP_TLS enabled in order to use TLS.");
-#else /* USE_LDAP_TLS */
+#ifndef LDAP_OPT_X_TLS
+  CONF_ERROR(cmd, "LDAPUseTLS: Your LDAP libraries do not appear to support TLS.");
+#else /* LDAP_OPT_X_TLS */
   int b;
   config_rec *c;
 
@@ -1970,7 +1812,7 @@ set_ldap_usetls(cmd_rec *cmd)
   c->argv[0] = pcalloc(c->pool, sizeof(int));
   *((int *) c->argv[0]) = b;
   return PR_HANDLED(cmd);
-#endif /* USE_LDAP_TLS */
+#endif /* LDAP_OPT_X_TLS */
 }
 
 MODRET
@@ -1988,6 +1830,10 @@ set_ldap_usessl(cmd_rec *cmd)
   if (find_config(main_server->conf, CONF_PARAM, "LDAPUseTLS", FALSE)) {
     CONF_ERROR(cmd, "LDAPUseSSL: cannot be used with LDAPUseTLS.");
   }
+  c = find_config(main_server->conf, CONF_PARAM, "LDAPServer", FALSE);
+  if (c && ldap_is_ldap_url(c->argv[0])) {
+    CONF_ERROR(cmd, "LDAPUseSSL cannot be used when LDAPServer specifies a URL; specify the desired scheme (ldap:// or ldaps://) in the LDAPServer URL instead.");
+  }
 
   if ((b = get_boolean(cmd, 1)) == -1) {
     CONF_ERROR(cmd, "LDAPUseSSL: expected a boolean value for first argument.");
@@ -2202,11 +2048,6 @@ ldap_getconf(void)
     ldap_forcegenhdir = *((int *) ptr);
   }
 
-  ptr = get_param_ptr(main_server->conf, "LDAPNegativeCache", FALSE);
-  if (ptr) {
-    ldap_negcache = *((int *) ptr);
-  }
-
   ptr = get_param_ptr(main_server->conf, "LDAPGenerateHomedir", FALSE);
   if (ptr) {
     ldap_genhdir = *((int *) ptr);
@@ -2227,12 +2068,12 @@ ldap_getconf(void)
     ldap_protocol_version = *((int *) ptr);
   }
 
-#ifdef USE_LDAP_TLS
+#ifdef LDAP_OPT_X_TLS
   ptr = get_param_ptr(main_server->conf, "LDAPUseTLS", FALSE);
   if (ptr) {
     ldap_use_tls = *((int *) ptr);
   }
-#endif /* LDAP_USE_TLS */
+#endif /* LDAP_OPT_X_TLS */
 
 #ifdef LDAP_OPT_X_TLS_HARD
   ptr = get_param_ptr(main_server->conf, "LDAPUseSSL", FALSE);
@@ -2241,48 +2082,61 @@ ldap_getconf(void)
   }
 #endif /* LDAP_OPT_X_TLS_HARD */
 
-  /* If LDAPServer isn't present, we'll leave ldap_server NULL, and
-   * ldap_init() will connect to the LDAP SDK's default.
+  /* If LDAPServer isn't present, we'll leave ldap_server_url NULL, and
+   * ldap_initialize() will connect to the LDAP SDK's default.
    */
   if ((c = find_config(main_server->conf, CONF_PARAM, "LDAPServer", FALSE)) != NULL) {
-    if (strcmp(c->argv[0], "url") == 0) {
-      url = c->argv[1];
-      if (url) {
+    if (ldap_is_ldap_url(c->argv[0])) {
+      if (ldap_url_parse(c->argv[0], &url) != LDAP_SUCCESS) {
+        pr_log_pri(PR_LOG_ERR, MOD_LDAP_VERSION ": ldap_getconf(): url %s was valid during ProFTPD startup, but is no longer valid?!", (char *)c->argv[0]);
+        return -1;
+      }
+
+      ldap_server_url = pstrdup(session.pool, c->argv[0]);
+
 #ifdef LDAP_OPT_X_TLS_HARD
-        if (strcmp(url->lud_scheme, "ldaps") == 0) {
-          ldap_use_ssl = 1;
-        }
+      if (strcmp(url->lud_scheme, "ldaps") == 0) {
+        ldap_use_ssl = 1;
+      }
 #endif /* LDAP_OPT_X_TLS_HARD */
 
-        if (url->lud_host != NULL) {
-          ldap_server = pstrdup(session.pool, url->lud_host);
-        }
-        if (url->lud_port != 0) {
-          ldap_port = url->lud_port;
-        }
-        if (url->lud_scope != LDAP_SCOPE_DEFAULT) {
-          ldap_search_scope = url->lud_scope;
-        }
-
-        /* We intentionally avoid ldap_free_urldesc()ing url, since it's
-         * attached to the LDAPServer configuration directive and will be used
-         * by other/future callers.
-         */
+#if LDAP_API_VERSION < 2000
+      /* Need to keep parsed host and port for pre-2000 ldap_init(). */
+      if (url->lud_host != NULL) {
+        ldap_server = pstrdup(session.pool, url->lud_host);
       }
-    } else if (strcmp(c->argv[0], "host") == 0) {
-      ldap_server = c->argv[1];
+      if (url->lud_port != 0) {
+        ldap_port = url->lud_port;
+      }
+#endif /* LDAP_API_VERSION < 2000 */
+
+      if (url->lud_scope != LDAP_SCOPE_DEFAULT) {
+        ldap_search_scope = url->lud_scope;
+      }
+
+      ldap_free_urldesc(url);
     } else {
-      /* This should never happen, since the configuration handler for
-       * LDAPServer only passes url or host, but we'll be defensive.
-       */
-      pr_log_pri(PR_LOG_ERR, MOD_LDAP_VERSION ": unexpected LDAPServer type.");
-      return -1;
+      ldap_server_url = pstrcat(session.pool,
+        "ldap://", c->argv[0], "/", NULL);
+
+#if LDAP_API_VERSION < 2000
+      ldap_server = pstrdup(session.pool, c->argv[0]);
+      ldap_port = LDAP_PORT;
+#endif /* LDAP_API_VERSION < 2000 */
     }
   }
 
   return 0;
 }
 
+static int ldap_mod_init(void) {
+  pr_log_debug(DEBUG2, MOD_LDAP_VERSION
+    ": compiled using LDAP vendor '%s', LDAP API version %lu",
+    LDAP_VENDOR_NAME, (unsigned long) LDAP_API_VERSION);
+
+  return 0;
+}
+
 static conftable ldap_config[] = {
   { "LDAPServer",                          set_ldap_server,               NULL },
   { "LDAPDNInfo",                          set_ldap_dninfo,               NULL },
@@ -2336,12 +2190,13 @@ static authtable ldap_auth[] = {
 };
 
 module ldap_module = {
-  NULL, NULL,                        /* Always NULL */
-  0x20,                              /* API Version 2.0 */
+  NULL, NULL,          /* Always NULL */
+  0x20,                /* API Version 2.0 */
   "ldap",
-  ldap_config,                       /* Configuration directive table */
-  ldap_cmdtab,                       /* Command handlers */
-  ldap_auth,                         /* Authentication handlers */
-  pr_ldap_module_init, ldap_getconf, /* Initialization functions */
+  ldap_config,         /* Configuration directive table */
+  ldap_cmdtab,         /* Command handlers */
+  ldap_auth,           /* Authentication handlers */
+  ldap_mod_init,       /* Initialization functions */
+  ldap_getconf,
   MOD_LDAP_VERSION
 };
