@@ -1,7 +1,7 @@
 /*
  * ProFTPD: mod_sql_mysql -- Support for connecting to MySQL databases.
  * Copyright (c) 2001 Andrew Houghton
- * Copyright (c) 2004-2005 TJ Saunders
+ * Copyright (c) 2004-2009 TJ Saunders
  *  
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -22,7 +22,7 @@
  * the resulting executable, without including the source code for OpenSSL in
  * the source distribution.
  *
- * $Id: mod_sql_mysql.c,v 1.39 2005/06/27 20:41:14 castaglia Exp $
+ * $Id: mod_sql_mysql.c,v 1.51 2009/02/05 18:53:54 castaglia Exp $
  */
 
 /*
@@ -128,14 +128,15 @@
  * Internal define used for debug and logging.  All backends are encouraged
  * to use the same format.
  */
-#define MOD_SQL_MYSQL_VERSION		"mod_sql_mysql/4.05"
+#define MOD_SQL_MYSQL_VERSION		"mod_sql_mysql/4.0.8"
 
 #define _MYSQL_PORT "3306"
 
-#include <mysql.h>
 #include "conf.h"
 #include "../contrib/mod_sql.h"
 
+#include <mysql.h>
+
 /* 
  * timer-handling code adds the need for a couple of forward declarations
  */
@@ -158,6 +159,7 @@ struct db_conn_struct {
   char *pass;
   char *db;
   char *port;
+  char *unix_sock;
 
   MYSQL *mysql;
 
@@ -236,7 +238,7 @@ static void *_sql_add_connection(pool *p, char *name, db_conn_t *conn)
     return NULL;
   }
 
-  entry = (conn_entry_t *) pcalloc( p, sizeof( conn_entry_t ));
+  entry = (conn_entry_t *) pcalloc(p, sizeof(conn_entry_t));
   entry->name = name;
   entry->data = conn;
 
@@ -263,17 +265,16 @@ static void _sql_check_cmd(cmd_rec *cmd, char *msg) {
 }
 
 /*
- * _sql_timer_callback: when a timer goes off, this is the function
- *  that gets called.  This function makes assumptions about the 
- *  db_conn_t members.
+ * sql_timer_cb: when a timer goes off, this is the function that gets called.
+ * This function makes assumptions about the db_conn_t members.
  */
-static int _sql_timer_callback(CALLBACK_FRAME) {
+static int sql_timer_cb(CALLBACK_FRAME) {
   conn_entry_t *entry = NULL;
-  int cnt = 0;
+  int i = 0;
   cmd_rec *cmd = NULL;
  
-  for (cnt=0; cnt < conn_cache->nelts; cnt++) {
-    entry = ((conn_entry_t **) conn_cache->elts)[cnt];
+  for (i = 0; i < conn_cache->nelts; i++) {
+    entry = ((conn_entry_t **) conn_cache->elts)[i];
 
     if (entry->timer == p2) {
       sql_log(DEBUG_INFO, "timer expired for connection '%s'", entry->name);
@@ -296,11 +297,11 @@ static modret_t *_build_error(cmd_rec *cmd, db_conn_t *conn) {
   char num[20] = {'\0'};
 
   if (!conn)
-    return ERROR_MSG(cmd, MOD_SQL_MYSQL_VERSION, "badly formed request");
+    return PR_ERROR_MSG(cmd, MOD_SQL_MYSQL_VERSION, "badly formed request");
 
-  snprintf(num, 20, "%u", mysql_errno(conn->mysql) );
+  snprintf(num, 20, "%u", mysql_errno(conn->mysql));
 
-  return ERROR_MSG(cmd, num, (char *) mysql_error(conn->mysql));
+  return PR_ERROR_MSG(cmd, num, (char *) mysql_error(conn->mysql));
 }
 
 /*
@@ -320,7 +321,7 @@ static modret_t *_build_data(cmd_rec *cmd, db_conn_t *conn) {
   unsigned long i = 0;
 
   if (!conn) 
-    return ERROR_MSG(cmd, MOD_SQL_MYSQL_VERSION, "badly formed request");
+    return PR_ERROR_MSG(cmd, MOD_SQL_MYSQL_VERSION, "badly formed request");
 
   mysql = conn->mysql;
 
@@ -328,9 +329,9 @@ static modret_t *_build_data(cmd_rec *cmd, db_conn_t *conn) {
    * the number of rows returned we can't presize the data[] array.
    */
 
-  result = mysql_store_result( mysql );
+  result = mysql_store_result(mysql);
   if (!result) {
-    return _build_error( cmd, conn );
+    return _build_error(cmd, conn);
   }
   
   sd = (sql_data_t *) pcalloc(cmd->tmp_pool, sizeof(sql_data_t));
@@ -338,9 +339,9 @@ static modret_t *_build_data(cmd_rec *cmd, db_conn_t *conn) {
   sd->fnum = (unsigned long) mysql_num_fields(result);
   cnt = sd->rnum * sd->fnum;
 
-  data = (char **) pcalloc( cmd->tmp_pool, sizeof(char *) * (cnt + 1) );
+  data = (char **) pcalloc(cmd->tmp_pool, sizeof(char *) * (cnt + 1));
   
-  while ((row = mysql_fetch_row( result ))) {
+  while ((row = mysql_fetch_row(result))) {
     for (cnt = 0; cnt < sd->fnum; cnt++)
       data[i++] = pstrdup(cmd->tmp_pool, row[cnt]);
   }
@@ -348,9 +349,9 @@ static modret_t *_build_data(cmd_rec *cmd, db_conn_t *conn) {
   /* at this point either we finished correctly or an error occurred in the
    * fetch.  Do the right thing.
    */
-  if ( mysql_errno( mysql ) ) {
-    mr = _build_error( cmd, conn );
-    mysql_free_result( result );
+  if (mysql_errno(mysql)) {
+    mr = _build_error(cmd, conn);
+    mysql_free_result(result);
     return mr;
   }
 
@@ -384,18 +385,18 @@ MODRET cmd_open(cmd_rec *cmd) {
 
   sql_log(DEBUG_FUNC, "%s", "entering \tmysql cmd_open");
 
-  _sql_check_cmd(cmd, "cmd_open" );
+  _sql_check_cmd(cmd, "cmd_open");
 
   if (cmd->argc < 1) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tmysql cmd_open");
-    return ERROR_MSG(cmd, MOD_SQL_MYSQL_VERSION, "badly formed request");
+    return PR_ERROR_MSG(cmd, MOD_SQL_MYSQL_VERSION, "badly formed request");
   }    
 
   /* get the named connection */
 
   if (!(entry = _sql_get_connection( cmd->argv[0]))) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tmysql cmd_open");
-    return ERROR_MSG(cmd, MOD_SQL_MYSQL_VERSION, "unknown named connection");
+    return PR_ERROR_MSG(cmd, MOD_SQL_MYSQL_VERSION, "unknown named connection");
   } 
 
   conn = (db_conn_t *) entry->data;
@@ -411,7 +412,7 @@ MODRET cmd_open(cmd_rec *cmd) {
     sql_log(DEBUG_INFO, "connection '%s' count is now %d", entry->name,
       entry->connections);
     sql_log(DEBUG_FUNC, "%s", "exiting \tmysql cmd_open");
-    return HANDLED(cmd);
+    return PR_HANDLED(cmd);
   }
 
   /* Make sure we have a new conn struct */
@@ -431,21 +432,64 @@ MODRET cmd_open(cmd_rec *cmd) {
   mysql_options(conn->mysql, MYSQL_READ_DEFAULT_GROUP, "client");
 
   if (!mysql_real_connect(conn->mysql, conn->host, conn->user, conn->pass,
-      conn->db, (int) strtol(conn->port, (char **) NULL, 10), NULL,
-      CLIENT_INTERACTIVE)) {
+      conn->db, (int) strtol(conn->port, (char **) NULL, 10),
+      conn->unix_sock, CLIENT_INTERACTIVE)) {
 
     /* If it didn't work, return an error. */
     sql_log(DEBUG_FUNC, "%s", "exiting \tmysql cmd_open");
     return _build_error(cmd, conn);
   }
 
+#ifdef PR_USE_NLS
+  if (pr_encode_get_encoding() != NULL) {
+
+# if MYSQL_VERSION_ID >= 50007
+    /* Configure the connection for the current local character set.
+     *
+     * Note: the mysql_set_character_set() function appeared in MySQL 5.0.7,
+     * as per:
+     *
+     *  http://dev.mysql.com/doc/refman/5.0/en/mysql-set-character-set.html
+     */
+    if (!mysql_set_character_set(conn->mysql, pr_encode_get_charset())) {
+      sql_log(DEBUG_FUNC, "%s", "exiting \tmysql cmd_open");
+      return _build_error(cmd, conn);
+    }
+
+    sql_log(DEBUG_FUNC, "MySQL connection character set now '%s' (from '%s')",
+      mysql_character_set_name(conn->mysql), pr_encode_get_charset());
+
+# else
+    /* No mysql_set_character_set() API available.  But
+     * mysql_character_set_name() has been around for a while; we can use it
+     * to at least see whether there might be a character set discrepancy.
+     */
+
+    const char *local_charset = pr_encode_get_charset();
+    const char *mysql_charset = mysql_character_set_name(conn->mysql);
+
+    if (local_charset &&
+        mysql_charset &&
+        strcasecmp(local_charset, mysql_charset) != 0) {
+      pr_log_pri(PR_LOG_ERR, MOD_SQL_MYSQL_VERSION
+        ": local character set '%s' does not match MySQL character set '%s', "
+        "SQL injection possible, shutting down", local_charset, mysql_charset);
+      sql_log(DEBUG_WARN, "local character set '%s' does not match MySQL "
+        "character set '%s', SQL injection possible, shutting down",
+        local_charset, mysql_charset);
+      end_login(1);
+    }
+# endif /* older MySQL */
+  }
+#endif /* !PR_USE_NLS */
+
   /* bump connections */
   entry->connections++;
 
   /* set up our timer if necessary */
   if (entry->ttl > 0) {
     entry->timer = pr_timer_add(entry->ttl, -1, &sql_mysql_module,
-      _sql_timer_callback);
+      sql_timer_cb, "mysql connection ttl");
     sql_log(DEBUG_INFO, "connection '%s' - %d second timer started",
       entry->name, entry->ttl);
 
@@ -462,7 +506,7 @@ MODRET cmd_open(cmd_rec *cmd) {
     entry->connections);
 
   sql_log(DEBUG_FUNC, "%s", "exiting \tmysql cmd_open");
-  return HANDLED(cmd);
+  return PR_HANDLED(cmd);
 }
 
 /*
@@ -499,13 +543,13 @@ MODRET cmd_close(cmd_rec *cmd) {
 
   if ((cmd->argc < 1) || (cmd->argc > 2)) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tmysql cmd_close");
-    return ERROR_MSG(cmd, MOD_SQL_MYSQL_VERSION, "badly formed request");
+    return PR_ERROR_MSG(cmd, MOD_SQL_MYSQL_VERSION, "badly formed request");
   }
 
   /* get the named connection */
   if (!(entry = _sql_get_connection( cmd->argv[0] ))) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tmysql cmd_close");
-    return ERROR_MSG(cmd, MOD_SQL_MYSQL_VERSION, "unknown named connection");
+    return PR_ERROR_MSG(cmd, MOD_SQL_MYSQL_VERSION, "unknown named connection");
   }
 
   conn = (db_conn_t *) entry->data;
@@ -516,7 +560,7 @@ MODRET cmd_close(cmd_rec *cmd) {
       entry->connections);
 
     sql_log(DEBUG_FUNC, "%s", "exiting \tmysql cmd_close");
-    return HANDLED(cmd);
+    return PR_HANDLED(cmd);
   }
 
   /* decrement connections. If our count is 0 or we received a second arg
@@ -541,7 +585,7 @@ MODRET cmd_close(cmd_rec *cmd) {
     entry->connections);
   sql_log(DEBUG_FUNC, "%s", "exiting \tmysql cmd_close");
   
-  return HANDLED(cmd);
+  return PR_HANDLED(cmd);
 }
 
 /*
@@ -567,8 +611,7 @@ MODRET cmd_close(cmd_rec *cmd) {
  *  a number or ttl is negative, the connection will be assumed to have no
  *  associated timer.
  */
-MODRET cmd_defineconnection(cmd_rec *cmd)
-{
+MODRET cmd_defineconnection(cmd_rec *cmd) {
   char *info = NULL;
   char *name = NULL;
 
@@ -586,12 +629,27 @@ MODRET cmd_defineconnection(cmd_rec *cmd)
 
   _sql_check_cmd(cmd, "cmd_defineconnection");
 
-  if ((cmd->argc < 4) || (cmd->argc > 5) || (!cmd->argv[0])) {
+  if (cmd->argc < 4 ||
+      cmd->argc > 5 ||
+      !cmd->argv[0]) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tmysql cmd_defineconnection");
-    return ERROR_MSG(cmd, MOD_SQL_MYSQL_VERSION, "badly formed request");
+    return PR_ERROR_MSG(cmd, MOD_SQL_MYSQL_VERSION, "badly formed request");
   }
 
-  conn = (db_conn_t *) palloc(conn_pool, sizeof(db_conn_t));
+  if (!conn_pool) {
+    pr_log_pri(PR_LOG_WARNING, "warning: the mod_sql_mysql module has not been "
+      "properly initialized.  Please make sure your --with-modules configure "
+      "option lists mod_sql *before* mod_sql_mysql, and recompile.");
+
+    sql_log(DEBUG_FUNC, "%s", "The mod_sql_mysql module has not been properly "
+      "initialized.  Please make sure your --with-modules configure option "
+      "lists mod_sql *before* mod_sql_mysql, and recompile.");
+    sql_log(DEBUG_FUNC, "%s", "exiting \tmysql cmd_defineconnection");
+
+    return PR_ERROR_MSG(cmd, MOD_SQL_MYSQL_VERSION, "uninitialized module");
+  }
+
+  conn = (db_conn_t *) pcalloc(conn_pool, sizeof(db_conn_t));
 
   name = pstrdup(conn_pool, cmd->argv[0]);
   conn->user = pstrdup(conn_pool, cmd->argv[1]);
@@ -615,6 +673,7 @@ MODRET cmd_defineconnection(cmd_rec *cmd)
   if (haveport) {
     port = haveport + 1;
     *haveport = '\0';
+
   } else {
     port = _MYSQL_PORT;
   }
@@ -622,18 +681,30 @@ MODRET cmd_defineconnection(cmd_rec *cmd)
   if (havehost) {
     host = havehost + 1;
     *havehost = '\0';
+
   } else {
     host = "localhost";
   }
 
-  conn->host = pstrdup(conn_pool, host);
+  /* Hack to support ability to configure path to Unix domain socket
+   * for MySQL: if the host string starts with a '/', assume it's
+   * a path to the Unix domain socket to use.
+   */
+  if (*host == '/') {
+    conn->unix_sock = pstrdup(conn_pool, host);
+
+  } else {
+    conn->host = pstrdup(conn_pool, host);
+  }
+
   conn->db   = pstrdup(conn_pool, db);
   conn->port = pstrdup(conn_pool, port);
 
-  /* insert the new conn_info into the connection hash */
-  if (!(entry = _sql_add_connection(conn_pool, name, (void *) conn))) {
+  /* Insert the new conn_info into the connection hash */
+  entry = _sql_add_connection(conn_pool, name, (void *) conn);
+  if (!entry) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tmysql cmd_defineconnection");
-    return ERROR_MSG(cmd, MOD_SQL_MYSQL_VERSION,
+    return PR_ERROR_MSG(cmd, MOD_SQL_MYSQL_VERSION,
       "named connection already exists");
   }
 
@@ -645,15 +716,22 @@ MODRET cmd_defineconnection(cmd_rec *cmd)
   entry->timer = 0;
   entry->connections = 0;
 
-  sql_log(DEBUG_INFO, " name: '%s'", entry->name);
-  sql_log(DEBUG_INFO, " user: '%s'", conn->user);
-  sql_log(DEBUG_INFO, " host: '%s'", conn->host);
-  sql_log(DEBUG_INFO, "   db: '%s'", conn->db);
-  sql_log(DEBUG_INFO, " port: '%s'", conn->port);
-  sql_log(DEBUG_INFO, "  ttl: '%d'", entry->ttl);
+  sql_log(DEBUG_INFO, "  name: '%s'", entry->name);
+  sql_log(DEBUG_INFO, "  user: '%s'", conn->user);
+
+  if (conn->host) {
+    sql_log(DEBUG_INFO, "  host: '%s'", conn->host);
+
+  } else if (conn->unix_sock) {
+    sql_log(DEBUG_INFO, "socket: '%s'", conn->unix_sock);
+  }
+
+  sql_log(DEBUG_INFO, "    db: '%s'", conn->db);
+  sql_log(DEBUG_INFO, "  port: '%s'", conn->port);
+  sql_log(DEBUG_INFO, "   ttl: '%d'", entry->ttl);
 
   sql_log(DEBUG_FUNC, "%s", "exiting \tmysql cmd_defineconnection");
-  return HANDLED(cmd);
+  return PR_HANDLED(cmd);
 }
 
 /*
@@ -682,7 +760,7 @@ static modret_t *cmd_exit(cmd_rec *cmd) {
 
   sql_log(DEBUG_FUNC, "%s", "exiting \tmysql cmd_exit");
 
-  return HANDLED(cmd);
+  return PR_HANDLED(cmd);
 }
 
 /*
@@ -746,14 +824,14 @@ MODRET cmd_select(cmd_rec *cmd) {
 
   if (cmd->argc < 2) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tmysql cmd_select");
-    return ERROR_MSG(cmd, MOD_SQL_MYSQL_VERSION, "badly formed request");
+    return PR_ERROR_MSG(cmd, MOD_SQL_MYSQL_VERSION, "badly formed request");
   }
 
   /* get the named connection */
-  entry = _sql_get_connection( cmd->argv[0] );
+  entry = _sql_get_connection(cmd->argv[0]);
   if (!entry) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tmysql cmd_select");
-    return ERROR_MSG(cmd, MOD_SQL_MYSQL_VERSION, "unknown named connection");
+    return PR_ERROR_MSG(cmd, MOD_SQL_MYSQL_VERSION, "unknown named connection");
   }
  
   conn = (db_conn_t *) entry->data;
@@ -767,41 +845,47 @@ MODRET cmd_select(cmd_rec *cmd) {
   /* construct the query string */
   if (cmd->argc == 2) {
     query = pstrcat(cmd->tmp_pool, "SELECT ", cmd->argv[1], NULL);
+
   } else {
-    query = pstrcat( cmd->tmp_pool, cmd->argv[2], " FROM ", 
-		     cmd->argv[1], NULL );
-    if ((cmd->argc > 3) && (cmd->argv[3]))
-      query = pstrcat( cmd->tmp_pool, query, " WHERE ", cmd->argv[3], NULL );
-    if ((cmd->argc > 4) && (cmd->argv[4]))
-      query = pstrcat( cmd->tmp_pool, query, " LIMIT ", cmd->argv[4], NULL );
+    query = pstrcat(cmd->tmp_pool, cmd->argv[2], " FROM ", cmd->argv[1], NULL);
+
+    if (cmd->argc > 3 &&
+        cmd->argv[3])
+      query = pstrcat(cmd->tmp_pool, query, " WHERE ", cmd->argv[3], NULL);
+
+    if (cmd->argc > 4 &&
+        cmd->argv[4])
+      query = pstrcat(cmd->tmp_pool, query, " LIMIT ", cmd->argv[4], NULL);
+
     if (cmd->argc > 5) {
 
-      /* handle the optional arguments -- they're rare, so in this case
+      /* Handle the optional arguments -- they're rare, so in this case
        * we'll play with the already constructed query string, but in 
        * general we should probably take optional arguments into account 
        * and put the query string together later once we know what they are.
        */
     
-      for (cnt=5; cnt < cmd->argc; cnt++) {
-	if ((cmd->argv[cnt]) && !strcasecmp("DISTINCT",cmd->argv[cnt])) {
-	  query = pstrcat( cmd->tmp_pool, "DISTINCT ", query, NULL);
+      for (cnt = 5; cnt < cmd->argc; cnt++) {
+	if (cmd->argv[cnt] &&
+            strcasecmp("DISTINCT", cmd->argv[cnt]) == 0) {
+	  query = pstrcat(cmd->tmp_pool, "DISTINCT ", query, NULL);
 	}
       }
     }
 
-    query = pstrcat( cmd->tmp_pool, "SELECT ", query, NULL);    
+    query = pstrcat(cmd->tmp_pool, "SELECT ", query, NULL);
   }
 
-  /* log the query string */
-  sql_log( DEBUG_INFO, "query \"%s\"", query);
+  /* Log the query string */
+  sql_log(DEBUG_INFO, "query \"%s\"", query);
 
-  /* perform the query.  if it doesn't work, log the error, close the
+  /* Perform the query.  if it doesn't work, log the error, close the
    * connection then return the error from the query processing.
    */
-  if ( mysql_real_query( conn->mysql, query, strlen(query) ) ) {
-    dmr = _build_error( cmd, conn );
+  if (mysql_real_query(conn->mysql, query, strlen(query))) {
+    dmr = _build_error(cmd, conn);
 
-    close_cmd = _sql_make_cmd( cmd->tmp_pool, 1, entry->name );
+    close_cmd = _sql_make_cmd(cmd->tmp_pool, 1, entry->name);
     cmd_close(close_cmd);
     SQL_FREE_CMD(close_cmd);
 
@@ -809,14 +893,14 @@ MODRET cmd_select(cmd_rec *cmd) {
     return dmr;
   }
 
-  /* get the data. if it doesn't work, log the error, close the
+  /* Get the data. if it doesn't work, log the error, close the
    * connection then return the error from the data processing.
    */
-  dmr = _build_data( cmd, conn );
+  dmr = _build_data(cmd, conn);
   if (MODRET_ERROR(dmr)) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tmysql cmd_select");
 
-    close_cmd = _sql_make_cmd( cmd->tmp_pool, 1, entry->name );
+    close_cmd = _sql_make_cmd(cmd->tmp_pool, 1, entry->name);
     cmd_close(close_cmd);
     SQL_FREE_CMD(close_cmd);
 
@@ -824,10 +908,10 @@ MODRET cmd_select(cmd_rec *cmd) {
   }    
 
   /* close the connection, return the data. */
-  close_cmd = _sql_make_cmd( cmd->tmp_pool, 1, entry->name );
+  close_cmd = _sql_make_cmd(cmd->tmp_pool, 1, entry->name);
   cmd_close(close_cmd);
   SQL_FREE_CMD(close_cmd);
-
+ 
   sql_log(DEBUG_FUNC, "%s", "exiting \tmysql cmd_select");
   return dmr;
 }
@@ -877,14 +961,14 @@ MODRET cmd_insert(cmd_rec *cmd) {
 
   if ((cmd->argc != 2) && (cmd->argc != 4)) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tmysql cmd_insert");
-    return ERROR_MSG(cmd, MOD_SQL_MYSQL_VERSION, "badly formed request");
+    return PR_ERROR_MSG(cmd, MOD_SQL_MYSQL_VERSION, "badly formed request");
   }
 
   /* get the named connection */
   entry = _sql_get_connection(cmd->argv[0]);
   if (!entry) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tmysql cmd_insert");
-    return ERROR_MSG(cmd, MOD_SQL_MYSQL_VERSION, "unknown named connection");
+    return PR_ERROR_MSG(cmd, MOD_SQL_MYSQL_VERSION, "unknown named connection");
   }
 
   conn = (db_conn_t *) entry->data;
@@ -911,10 +995,10 @@ MODRET cmd_insert(cmd_rec *cmd) {
    * connection (and log any errors there, too) then return the error
    * from the query processing.
    */
-  if ( mysql_real_query( conn->mysql, query, strlen(query) ) ) {
-    dmr = _build_error( cmd, conn );
+  if (mysql_real_query(conn->mysql, query, strlen(query))) {
+    dmr = _build_error(cmd, conn);
 
-    close_cmd = _sql_make_cmd( cmd->tmp_pool, 1, entry->name );
+    close_cmd = _sql_make_cmd(cmd->tmp_pool, 1, entry->name);
     cmd_close(close_cmd);
     SQL_FREE_CMD(close_cmd);
 
@@ -923,12 +1007,12 @@ MODRET cmd_insert(cmd_rec *cmd) {
   }
 
   /* close the connection and return HANDLED. */
-  close_cmd = _sql_make_cmd( cmd->tmp_pool, 1, entry->name );
+  close_cmd = _sql_make_cmd(cmd->tmp_pool, 1, entry->name);
   cmd_close(close_cmd);
   SQL_FREE_CMD(close_cmd);
 
   sql_log(DEBUG_FUNC, "%s", "exiting \tmysql cmd_insert");
-  return HANDLED(cmd);
+  return PR_HANDLED(cmd);
 }
 
 /*
@@ -975,14 +1059,14 @@ MODRET cmd_update(cmd_rec *cmd) {
 
   if ((cmd->argc < 2) || (cmd->argc > 4)) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tmysql cmd_update");
-    return ERROR_MSG(cmd, MOD_SQL_MYSQL_VERSION, "badly formed request");
+    return PR_ERROR_MSG(cmd, MOD_SQL_MYSQL_VERSION, "badly formed request");
   }
 
   /* get the named connection */
   entry = _sql_get_connection(cmd->argv[0]);
   if (!entry) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tmysql cmd_update");
-    return ERROR_MSG(cmd, MOD_SQL_MYSQL_VERSION, "unknown named connection");
+    return PR_ERROR_MSG(cmd, MOD_SQL_MYSQL_VERSION, "unknown named connection");
   }
 
   conn = (db_conn_t *) entry->data;
@@ -1003,16 +1087,16 @@ MODRET cmd_update(cmd_rec *cmd) {
       query = pstrcat( cmd->tmp_pool, query, " WHERE ", cmd->argv[3], NULL );
   }
 
-  /* log the query string */
+  /* Log the query string */
   sql_log(DEBUG_INFO, "query \"%s\"", query);
 
-  /* perform the query.  if it doesn't work close the connection, then
+  /* Perform the query.  if it doesn't work close the connection, then
    * return the error from the query processing.
    */
-  if ( mysql_real_query( conn->mysql, query, strlen(query) ) ) {
-    dmr = _build_error( cmd, conn );
+  if (mysql_real_query(conn->mysql, query, strlen(query))) {
+    dmr = _build_error(cmd, conn);
 
-    close_cmd = _sql_make_cmd( cmd->tmp_pool, 1, entry->name );
+    close_cmd = _sql_make_cmd(cmd->tmp_pool, 1, entry->name);
     cmd_close(close_cmd);
     SQL_FREE_CMD(close_cmd);
 
@@ -1020,13 +1104,13 @@ MODRET cmd_update(cmd_rec *cmd) {
     return dmr;
   }
 
-  /* close the connection, return HANDLED.  */
-  close_cmd = _sql_make_cmd( cmd->tmp_pool, 1, entry->name );
+  /* Close the connection, return HANDLED.  */
+  close_cmd = _sql_make_cmd(cmd->tmp_pool, 1, entry->name);
   cmd_close(close_cmd);
   SQL_FREE_CMD(close_cmd);
 
   sql_log(DEBUG_FUNC, "%s", "exiting \tmysql cmd_update");
-  return HANDLED(cmd);
+  return PR_HANDLED(cmd);
 }
 
 /*
@@ -1054,14 +1138,14 @@ MODRET cmd_procedure(cmd_rec *cmd) {
 
   if (cmd->argc != 3) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tmysql cmd_procedure");
-    return ERROR_MSG(cmd, MOD_SQL_MYSQL_VERSION, "badly formed request");
+    return PR_ERROR_MSG(cmd, MOD_SQL_MYSQL_VERSION, "badly formed request");
   }
 
   /* MySQL does not support procedures.  Nothing to do. */
 
   sql_log(DEBUG_FUNC, "%s", "exiting \tmysql cmd_procedure");
 
-  return ERROR_MSG(cmd, MOD_SQL_MYSQL_VERSION,
+  return PR_ERROR_MSG(cmd, MOD_SQL_MYSQL_VERSION,
     "backend does not support procedures");
 }
 
@@ -1098,14 +1182,14 @@ MODRET cmd_query(cmd_rec *cmd) {
 
   if (cmd->argc != 2) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tmysql cmd_query");
-    return ERROR_MSG(cmd, MOD_SQL_MYSQL_VERSION, "badly formed request");
+    return PR_ERROR_MSG(cmd, MOD_SQL_MYSQL_VERSION, "badly formed request");
   }
 
   /* get the named connection */
   entry = _sql_get_connection(cmd->argv[0]);
   if (!entry) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tmysql cmd_query");
-    return ERROR_MSG(cmd, MOD_SQL_MYSQL_VERSION, "unknown named connection");
+    return PR_ERROR_MSG(cmd, MOD_SQL_MYSQL_VERSION, "unknown named connection");
   }
 
   conn = (db_conn_t *) entry->data;
@@ -1118,16 +1202,16 @@ MODRET cmd_query(cmd_rec *cmd) {
 
   query = pstrcat(cmd->tmp_pool, cmd->argv[1], NULL);
 
-  /* log the query string */
+  /* Log the query string */
   sql_log(DEBUG_INFO, "query \"%s\"", query);
 
-  /* perform the query.  if it doesn't work close the connection, then
+  /* Perform the query.  if it doesn't work close the connection, then
    * return the error from the query processing.
    */
-  if ( mysql_real_query( conn->mysql, query, strlen(query) ) ) {
-    dmr = _build_error( cmd, conn );
+  if (mysql_real_query(conn->mysql, query, strlen(query))) {
+    dmr = _build_error(cmd, conn);
     
-    close_cmd = _sql_make_cmd( cmd->tmp_pool, 1, entry->name );
+    close_cmd = _sql_make_cmd(cmd->tmp_pool, 1, entry->name);
     cmd_close(close_cmd);
     SQL_FREE_CMD(close_cmd);
     
@@ -1135,21 +1219,22 @@ MODRET cmd_query(cmd_rec *cmd) {
     return dmr;
   }
 
-  /* get data if necessary. if it doesn't work, log the error, close the
+  /* Get data if necessary. if it doesn't work, log the error, close the
    * connection then return the error from the data processing.
    */
 
-  if ( mysql_field_count( conn->mysql ) ) {
-    dmr = _build_data( cmd, conn );
+  if (mysql_field_count(conn->mysql)) {
+    dmr = _build_data(cmd, conn);
     if (MODRET_ERROR(dmr)) {
       sql_log(DEBUG_FUNC, "%s", "exiting \tmysql cmd_query");
     }
+
   } else {
-    dmr = HANDLED(cmd);
+    dmr = PR_HANDLED(cmd);
   }
   
   /* close the connection, return the data. */
-  close_cmd = _sql_make_cmd( cmd->tmp_pool, 1, entry->name );
+  close_cmd = _sql_make_cmd(cmd->tmp_pool, 1, entry->name);
   cmd_close(close_cmd);
   SQL_FREE_CMD(close_cmd);
 
@@ -1186,6 +1271,7 @@ MODRET cmd_query(cmd_rec *cmd) {
 MODRET cmd_escapestring(cmd_rec * cmd) {
   conn_entry_t *entry = NULL;
   db_conn_t *conn = NULL;
+  modret_t *cmr = NULL;
   char *unescaped = NULL;
   char *escaped = NULL;
 
@@ -1195,18 +1281,25 @@ MODRET cmd_escapestring(cmd_rec * cmd) {
 
   if (cmd->argc != 2) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tmysql cmd_escapestring");
-    return ERROR_MSG(cmd, MOD_SQL_MYSQL_VERSION, "badly formed request");
+    return PR_ERROR_MSG(cmd, MOD_SQL_MYSQL_VERSION, "badly formed request");
   }
 
   /* get the named connection */
   entry = _sql_get_connection(cmd->argv[0]);
   if (!entry) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tmysql cmd_escapestring");
-    return ERROR_MSG(cmd, MOD_SQL_MYSQL_VERSION, "unknown named connection");
+    return PR_ERROR_MSG(cmd, MOD_SQL_MYSQL_VERSION, "unknown named connection");
   }
 
   conn = (db_conn_t *) entry->data;
 
+  /* Make sure the connection is open. */
+  cmr = cmd_open(cmd);
+  if (MODRET_ERROR(cmr)) {
+    sql_log(DEBUG_FUNC, "%s", "exiting \tmysql cmd_escapestring");
+    return cmr;
+  }
+
   unescaped = cmd->argv[1];
   escaped = (char *) pcalloc(cmd->tmp_pool, sizeof(char) *
     (strlen(unescaped) * 2) + 1);
@@ -1236,12 +1329,12 @@ MODRET cmd_escapestring(cmd_rec * cmd) {
  *  cmd->argv[2]: hashed string
  *
  * Returns:
- *  HANDLED(cmd)                        -- passwords match
- *  ERROR_INT(cmd, PR_AUTH_NOPWD)       -- missing password
- *  ERROR_INT(cmd, PR_AUTH_BADPWD)      -- passwords don't match
- *  ERROR_INT(cmd, PR_AUTH_DISABLEPWD)  -- password is disabled
- *  ERROR_INT(cmd, PR_AUTH_AGEPWD)      -- password is aged
- *  ERROR(cmd)                          -- unknown error
+ *  PR_HANDLED(cmd)                        -- passwords match
+ *  PR_ERROR_INT(cmd, PR_AUTH_NOPWD)       -- missing password
+ *  PR_ERROR_INT(cmd, PR_AUTH_BADPWD)      -- passwords don't match
+ *  PR_ERROR_INT(cmd, PR_AUTH_DISABLEPWD)  -- password is disabled
+ *  PR_ERROR_INT(cmd, PR_AUTH_AGEPWD)      -- password is aged
+ *  PR_ERROR(cmd)                          -- unknown error
  *
  * Notes:
  *  If this backend does not provide this functionality, this cmd *must*
@@ -1261,19 +1354,19 @@ MODRET cmd_checkauth(cmd_rec * cmd) {
 
   if (cmd->argc != 3) {
     sql_log(DEBUG_FUNC, "exiting \tmysql cmd_checkauth");
-    return ERROR_MSG(cmd, MOD_SQL_MYSQL_VERSION, "badly formed request");
+    return PR_ERROR_MSG(cmd, MOD_SQL_MYSQL_VERSION, "badly formed request");
   }
 
   /* get the named connection -- not used in this case, but for consistency */
   entry = _sql_get_connection( cmd->argv[0] );
   if (!entry) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tmysql cmd_checkauth");
-    return ERROR_MSG(cmd, MOD_SQL_MYSQL_VERSION, "unknown named connection");
+    return PR_ERROR_MSG(cmd, MOD_SQL_MYSQL_VERSION, "unknown named connection");
   }
 
   if (cmd->argv[1] == NULL) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tmysql cmd_checkauth");
-    return ERROR_INT(cmd, PR_AUTH_NOPWD);
+    return PR_ERROR_INT(cmd, PR_AUTH_NOPWD);
   }
 
   conn = (db_conn_t *) entry->data;
@@ -1311,7 +1404,7 @@ MODRET cmd_checkauth(cmd_rec * cmd) {
 
   sql_log(DEBUG_FUNC, "%s", "exiting \tmysql cmd_checkauth");
 
-  return success ? HANDLED(cmd) : ERROR_INT(cmd, PR_AUTH_BADPWD);
+  return success ? PR_HANDLED(cmd) : PR_ERROR_INT(cmd, PR_AUTH_BADPWD);
 }
 
 /*
@@ -1351,24 +1444,65 @@ MODRET cmd_identify(cmd_rec * cmd) {
   return mod_create_data(cmd, (void *) sd);
 }  
 
+/*
+ * cmd_prepare: prepares this mod_sql_mysql module for running.
+ *
+ * Inputs:
+ *  cmd->argv[0]:  A pool to be used for any necessary preparations.
+ *
+ * Returns:
+ *  Success.
+ */
+MODRET cmd_prepare(cmd_rec *cmd) {
+  if (cmd->argc != 1) {
+    return PR_ERROR(cmd);
+  }
+
+  conn_pool = (pool *) cmd->argv[0];
+  conn_cache = make_array((pool *) cmd->argv[0], DEF_CONN_POOL_SIZE,
+    sizeof(conn_entry_t));
+
+  return mod_create_data(cmd, NULL);
+}
+
+/*
+ * cmd_cleanup: cleans up any initialisations made during module preparations
+ *  (see cmd_prepre).
+ *
+ * Inputs:
+ *  None.
+ *
+ * Returns:
+ *  Success.
+ */
+MODRET cmd_cleanup(cmd_rec *cmd) {
+  destroy_pool(conn_pool);
+  conn_pool = NULL;
+  conn_cache = NULL;
+
+  return mod_create_data(cmd, NULL);
+}
+
 /* SQL cmdtable: mod_sql requires each backend module to define a cmdtable
  *  with this exact name. ALL these functions must be defined; mod_sql checks
  *  that they all exist on startup and ProFTPD will refuse to start if they
  *  aren't defined.
  */
 static cmdtable sql_mysql_cmdtable[] = {
-  { CMD, "sql_open",             G_NONE, cmd_open,             FALSE, FALSE },
   { CMD, "sql_close",            G_NONE, cmd_close,            FALSE, FALSE },
+  { CMD, "sql_checkauth",        G_NONE, cmd_checkauth,        FALSE, FALSE },
+  { CMD, "sql_cleanup",          G_NONE, cmd_cleanup,          FALSE, FALSE },
   { CMD, "sql_defineconnection", G_NONE, cmd_defineconnection, FALSE, FALSE },
+  { CMD, "sql_escapestring",     G_NONE, cmd_escapestring,     FALSE, FALSE },
   { CMD, "sql_exit",             G_NONE, cmd_exit,             FALSE, FALSE },
-  { CMD, "sql_select",           G_NONE, cmd_select,           FALSE, FALSE },
+  { CMD, "sql_identify",         G_NONE, cmd_identify,         FALSE, FALSE },
   { CMD, "sql_insert",           G_NONE, cmd_insert,           FALSE, FALSE },
-  { CMD, "sql_update",           G_NONE, cmd_update,           FALSE, FALSE },
+  { CMD, "sql_open",             G_NONE, cmd_open,             FALSE, FALSE },
+  { CMD, "sql_prepare",          G_NONE, cmd_prepare,          FALSE, FALSE },
   { CMD, "sql_procedure",        G_NONE, cmd_procedure,        FALSE, FALSE },
   { CMD, "sql_query",            G_NONE, cmd_query,            FALSE, FALSE },
-  { CMD, "sql_escapestring",     G_NONE, cmd_escapestring,     FALSE, FALSE },
-  { CMD, "sql_checkauth",        G_NONE, cmd_checkauth,        FALSE, FALSE },
-  { CMD, "sql_identify",         G_NONE, cmd_identify,         FALSE, FALSE },
+  { CMD, "sql_select",           G_NONE, cmd_select,           FALSE, FALSE },
+  { CMD, "sql_update",           G_NONE, cmd_update,           FALSE, FALSE },
 
   { 0, NULL }
 };
