@@ -2,7 +2,7 @@
  * ProFTPD: mod_sql_postgres -- Support for connecting to Postgres databases.
  * Time-stamp: <1999-10-04 03:21:21 root>
  * Copyright (c) 2001 Andrew Houghton
- * Copyright (c) 2004-2007 TJ Saunders
+ * Copyright (c) 2004-2009 TJ Saunders
  *  
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -23,7 +23,7 @@
  * the resulting executable, without including the source code for OpenSSL in
  * the source distribution.
  *
- * $Id: mod_sql_postgres.c,v 1.32 2007/07/06 22:40:58 castaglia Exp $
+ * $Id: mod_sql_postgres.c,v 1.37 2009/02/05 18:53:54 castaglia Exp $
  */
 
 /*
@@ -34,7 +34,7 @@
  * Internal define used for debug and logging.  All backends are encouraged
  * to use the same format.
  */
-#define MOD_SQL_POSTGRES_VERSION	"mod_sql_postgres/4.03"
+#define MOD_SQL_POSTGRES_VERSION	"mod_sql_postgres/4.0.4"
 
 #define _POSTGRES_PORT "5432"
 
@@ -43,6 +43,21 @@
 
 #include <libpq-fe.h>
 
+/* For the pg_encoding_to_char() function, used for NLS support, we need
+ * to include the <mb/pg_wchar.h> file.  It's OK; the function has been
+ * declared in that file for a while, according to:
+ *
+ *  http://archives.postgresql.org/pgsql-bugs/2006-07/msg00125.php
+ *
+ * But it's a pain to quell all of the compiler warnings about redefined
+ * this, undefined that.  The linker finds the symbol without issue, so
+ * punt on including <mb/pg_wchar.h> file.  Instead, we'll copy the
+ * function declaration here.
+ */
+#ifdef PR_USE_NLS
+extern const char *pg_encoding_to_char(int encoding);
+#endif
+
 /* 
  * timer-handling code adds the need for a couple of forward declarations
  */
@@ -142,7 +157,7 @@ static void *_sql_add_connection(pool *p, char *name, db_conn_t *conn)
     return NULL;
   }
 
-  entry = (conn_entry_t *) pcalloc( p, sizeof( conn_entry_t ));
+  entry = (conn_entry_t *) pcalloc(p, sizeof(conn_entry_t));
   entry->name = name;
   entry->data = conn;
 
@@ -168,17 +183,16 @@ static void _sql_check_cmd(cmd_rec *cmd, char *msg) {
 }
 
 /*
- * _sql_timer_callback: when a timer goes off, this is the function
- *  that gets called.  This function makes assumptions about the 
- *  db_conn_t members.
+ * sql_timer_cb: when a timer goes off, this is the function that gets called.
+ * This function makes assumptions about the db_conn_t members.
  */
-static int _sql_timer_callback(CALLBACK_FRAME) {
+static int sql_timer_cb(CALLBACK_FRAME) {
   conn_entry_t *entry = NULL;
-  int cnt = 0;
+  int i = 0;
   cmd_rec *cmd = NULL;
  
-  for (cnt=0; cnt < conn_cache->nelts; cnt++) {
-    entry = ((conn_entry_t **) conn_cache->elts)[cnt];
+  for (i = 0; i < conn_cache->nelts; i++) {
+    entry = ((conn_entry_t **) conn_cache->elts)[i];
 
     if (entry->timer == p2) {
       sql_log(DEBUG_INFO, "timer expired for connection '%s'", entry->name);
@@ -228,8 +242,8 @@ static modret_t *_build_data(cmd_rec *cmd, db_conn_t *conn) {
   sd->rnum = (unsigned long) PQntuples(result);
   sd->fnum = (unsigned long) PQnfields(result);
 
-  data = (char **) pcalloc( cmd->tmp_pool, sizeof(char *) * 
-			    ((sd->rnum * sd->fnum) + 1) );
+  data = (char **) pcalloc(cmd->tmp_pool, sizeof(char *) * 
+			    ((sd->rnum * sd->fnum) + 1));
   
   for (row = 0; row < sd->rnum; row++) {
     for (field = 0; field < sd->fnum; field++) {
@@ -307,13 +321,46 @@ MODRET cmd_open(cmd_rec *cmd) {
     return _build_error( cmd, conn );
   }
 
+#ifdef PR_USE_NLS
+  if (pr_encode_get_encoding() != NULL) {
+    const char *encoding = pr_encode_get_encoding();
+
+    /* XXX Hack to deal with Postgres' incredibly broken behavior when
+     * handling the 'ASCII' encoding.  Specifically, Postgres chokes on
+     * 'ASCII', and instead insists on calling it 'SQL_ASCII' (which, of
+     * course, is not even close to being a valid encoding name according
+     * to libiconv.)
+     *
+     * Treat 'ANSI_X3.4-1968' (another common name/alias for ASCII) the
+     * same; rename it to 'SQL_ASCII' for Postgres' benefit.  And same for
+     * 'US-ASCII'.
+     */
+    if (strcmp(encoding, "ANSI_X3.4-1968") == 0 ||
+        strcmp(encoding, "ASCII") == 0 ||
+        strcmp(encoding, "US-ASCII") == 0) {
+      encoding = "SQL_ASCII";
+    }
+
+    /* Configure the connection for the current local character set. */
+    if (PQsetClientEncoding(conn->postgres, encoding) < 0) {
+      /* if it didn't work, return an error */
+      sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_open");
+      return _build_error(cmd, conn);
+    }
+
+    sql_log(DEBUG_FUNC, "Postgres connection character set now '%s' "
+      "(from '%s')", pg_encoding_to_char(PQclientEncoding(conn->postgres)),
+      pr_encode_get_charset());
+  }
+#endif /* !PR_USE_NLS */
+
   /* bump connections */
   entry->connections++;
 
   /* set up our timer if necessary */
   if (entry->ttl > 0) {
     entry->timer = pr_timer_add(entry->ttl, -1, &sql_postgres_module,
-      _sql_timer_callback);
+      sql_timer_cb, "postgres connection ttl");
     sql_log(DEBUG_INFO, "connection '%s' - %d second timer started",
       entry->name, entry->ttl);
 
@@ -474,7 +521,7 @@ MODRET cmd_defineconnection(cmd_rec *cmd) {
     return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION, "uninitialized module");
   }
 
-  conn = (db_conn_t *) palloc(conn_pool, sizeof(db_conn_t));
+  conn = (db_conn_t *) pcalloc(conn_pool, sizeof(db_conn_t));
 
   name = pstrdup(conn_pool, cmd->argv[0]);
   conn->user = pstrdup(conn_pool, cmd->argv[1]);
@@ -1102,6 +1149,7 @@ MODRET cmd_query(cmd_rec *cmd) {
 MODRET cmd_escapestring(cmd_rec * cmd) {
   conn_entry_t *entry = NULL;
   db_conn_t *conn = NULL;
+  modret_t *cmr = NULL;
   char *unescaped = NULL;
   char *escaped = NULL;
 
@@ -1124,6 +1172,13 @@ MODRET cmd_escapestring(cmd_rec * cmd) {
 
   conn = (db_conn_t *) entry->data;
 
+  /* Make sure the connection is open. */
+  cmr = cmd_open(cmd);
+  if (MODRET_ERROR(cmr)) {
+    sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_escapestring");
+    return cmr;
+  }
+
   /* Note: the PQescapeString() function appeared in the C API as of
    * Postgres-7.2.
    */
@@ -1214,8 +1269,8 @@ MODRET cmd_identify(cmd_rec * cmd) {
 
   _sql_check_cmd(cmd, "cmd_identify");
 
-  sd = (sql_data_t *) pcalloc( cmd->tmp_pool, sizeof(sql_data_t));
-  sd->data = (char **) pcalloc( cmd->tmp_pool, sizeof(char *) * 2);
+  sd = (sql_data_t *) pcalloc(cmd->tmp_pool, sizeof(sql_data_t));
+  sd->data = (char **) pcalloc(cmd->tmp_pool, sizeof(char *) * 2);
 
   sd->rnum = 1;
   sd->fnum = 2;
