@@ -1,6 +1,6 @@
 /*
  * ProFTPD - FTP server daemon
- * Copyright (c) 2004-2007 The ProFTPD Project team
+ * Copyright (c) 2004-2008 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -24,7 +24,7 @@
 
 /* ProFTPD Auth API
  *
- * $Id: auth.h,v 1.5 2007/04/17 21:33:40 castaglia Exp $
+ * $Id: auth.h,v 1.10 2008/05/14 05:51:38 castaglia Exp $
  */
 
 #ifndef PR_AUTH_H
@@ -85,6 +85,32 @@ int pr_auth_requires_pass(pool *, const char *);
  */
 config_rec *pr_auth_get_anon_config(pool *p, char **, char **, char **);
 
+/* Wrapper function around the chroot(2) system call, handles setting of
+ * appropriate environment variables if necessary.
+ */
+int pr_auth_chroot(const char *);
+
+/* Check the /etc/ftpusers file, as per the UseFtpUsers directive, to see
+ * if the given user is allowed.  Returns TRUE if the user is banned by
+ * /etc/ftpusers, FALSE if not banned, and -1 if there was an error.
+ */
+int pr_auth_banned_by_ftpusers(xaset_t *, const char *);
+
+/* Check the /etc/shells file, as per the RequireValidShell directive, to
+ * ensure that the given shell is valid.  Returns TRUE if the user has
+ * a valid shell, FALSE if an invalid shell, and -1 if there was an error.
+ */
+int pr_auth_is_valid_shell(xaset_t *, const char *);
+
+/* Add to the list of authenticating-only modules (e.g. PAM). */
+int pr_auth_add_auth_only_module(const char *);
+
+/* Enable caching of certain data within the Auth API. */
+int pr_auth_cache_set(int, unsigned int);
+#define PR_AUTH_CACHE_FL_UID2NAME	0x00001
+#define PR_AUTH_CACHE_FL_GID2NAME	0x00002
+#define PR_AUTH_CACHE_FL_AUTH_MODULE	0x00004
+
 /* For internal use only. */
 int init_auth(void);
 int set_groups(pool *, gid_t, array_header *);
