@@ -2,7 +2,7 @@
  * ProFTPD - FTP server daemon
  * Copyright (c) 1997, 1998 Public Flood Software
  * Copyright (c) 1999, 2000 MacGyver aka Habeeb J. Dihu <macgyver@tos.net>
- * Copyright (c) 2001, 2002, 2003 The ProFTPD Project team
+ * Copyright (c) 2001-2008 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -27,7 +27,7 @@
 /*
  * Configuration structure, server, command and associated prototypes.
  *
- * $Id: dirtree.h,v 1.60 2005/06/12 00:36:12 castaglia Exp $
+ * $Id: dirtree.h,v 1.71 2008/06/14 02:40:04 castaglia Exp $
  */
 
 #ifndef PR_DIRTREE_H
@@ -115,12 +115,14 @@ struct config_struc {
   struct config_struc *next,*prev;
 
   int config_type;
+  unsigned int config_id;
+
   pool *pool;			/* memory pool for this object */
   xaset_t *set;			/* The set we are stored in */
   char *name;
   int argc;
   void **argv;
-  long overrides;		/* Override classes */
+
   long flags;			/* Flags */
 
   server_rec *server;		/* Server this config element is attached to */
@@ -155,11 +157,6 @@ struct config_struc {
 #define ORDER_ALLOWDENY		0
 #define ORDER_DENYALLOW		1
 
-/* For the different types of expressions: AND, OR, and REGEX. */
-#define PR_EXPR_EVAL_AND	0
-#define PR_EXPR_EVAL_OR		1
-#define PR_EXPR_EVAL_REGEX	2
-
 /* The following macro determines the "highest" level available for
  * configuration directives.  If a current dir_config is available, it's
  * subset is used, otherwise anon config or main server
@@ -176,12 +173,9 @@ extern int			SocketBindTight;
 extern char			ServerType;
 extern int			ServerMaxInstances;
 extern int			ServerUseReverseDNS;
-extern int			TimeoutIdle;
-extern int			TimeoutNoXfer;
-extern int			TimeoutStalled;
 
 /* These macros are used to help handle configuration in modules */
-#define CONF_ERROR(x, s)	return ERROR_MSG((x),NULL,pstrcat((x)->tmp_pool, \
+#define CONF_ERROR(x, s)	return PR_ERROR_MSG((x),NULL,pstrcat((x)->tmp_pool, \
 				(x)->argv[0],": ",(s),NULL));
 
 #define CHECK_ARGS(x, n)	if((x)->argc-1 < n) \
@@ -200,14 +194,14 @@ extern int			TimeoutStalled;
 
 #define CHECK_CMD_ARGS(x, n)	\
   if ((x)->argc != (n)) { \
-    pr_response_add_err(R_501, "Invalid number of arguments."); \
-    return ERROR((x)); \
+    pr_response_add_err(R_501, _("Invalid number of arguments")); \
+    return PR_ERROR((x)); \
   }
 
 #define CHECK_CMD_MIN_ARGS(x, n)	\
   if ((x)->argc < (n)) { \
-    pr_response_add_err(R_501, "Invalid number of arguments."); \
-    return ERROR((x)); \
+    pr_response_add_err(R_501, _("Invalid number of arguments")); \
+    return PR_ERROR((x)); \
   }
 
 /* Prototypes */
@@ -219,7 +213,9 @@ extern int			TimeoutStalled;
 void kludge_disable_umask(void);
 void kludge_enable_umask(void);
 
-unsigned char define_exists(const char *);
+int pr_define_add(const char *);
+unsigned char pr_define_exists(const char *);
+
 void init_config(void);
 int fixup_servers(xaset_t *list);
 int parse_config_path(pool *, const char *);
@@ -236,27 +232,31 @@ config_rec *find_config(xaset_t *, int, const char *, int);
 void find_config_set_top(config_rec *);
 int remove_config(xaset_t *, const char *, int);
 
+/* Returns the assigned ID for the provided directive name, or zero
+ * if no ID mapping was found.
+ */
+unsigned int pr_config_get_id(const char *name);
+
+/* Returns the buffer size to use for data transfers.
+ */
+int pr_config_get_xfer_bufsz(void);
+
+/* Assigns a unique ID for the given configuration directive.  The
+ * mapping of directive to ID is stored in a lookup table, so that
+ * searching of the config database by directive name can be done using
+ * ID comparisons rather than string comparisons.
+ *
+ * Returns the ID assigned for the given directive, or zero if there was an
+ * error.
+ */
+unsigned int pr_config_set_id(const char *name);
+
 cmd_rec *pr_cmd_alloc(pool *, int, ...);
 
-/* Expression API.  (XXX These should be in their own header/src files). */
-array_header *pr_expr_create(pool *, int *, char **);
-unsigned char pr_expr_eval_class_and(char **);
-unsigned char pr_expr_eval_class_or(char **);
-unsigned char pr_expr_eval_group_and(char **);
-unsigned char pr_expr_eval_group_or(char **);
-unsigned char pr_expr_eval_user_and(char **);
-unsigned char pr_expr_eval_user_or(char **);
-
-long get_param_int(xaset_t *, const char *, int);
-long get_param_int_next(const char *, int);
 void *get_param_ptr(xaset_t *, const char *, int);
 void *get_param_ptr_next(const char *, int);
 xaset_t *get_dir_ctxt(pool *, char *);
 
-char *pr_str_get_word(char **, int);
-#define PR_STR_FL_PRESERVE_COMMENTS		0x0001
-#define PR_STR_FL_PRESERVE_WHITESPACE		0x0002
-
 config_rec *dir_match_path(pool *, char *);
 void build_dyn_config(pool *, char *, struct stat *, unsigned char);
 unsigned char dir_hide_file(const char *);
@@ -273,10 +273,13 @@ void resolve_deferred_dirs(server_rec *);
 void fixup_dirs(server_rec *, int);
 unsigned char check_context(cmd_rec *, int);
 char *get_context_name(cmd_rec *);
-int pr_is_boolean(const char *);
 int get_boolean(cmd_rec *, int);
 char *get_full_cmd(cmd_rec *);
 
 void pr_config_dump(void (*)(const char *, ...), xaset_t *, char *);
 
+#ifdef PR_USE_DEVEL
+void pr_dirs_dump(void (*)(const char *, ...), xaset_t *, char *);
+#endif /* PR_USE_DEVEL */
+
 #endif /* PR_DIRTREE_H */
