@@ -2,7 +2,7 @@
  * ProFTPD - FTP server daemon
  * Copyright (c) 1997, 1998 Public Flood Software
  * Copyright (c) 1999, 2000 MacGyver aka Habeeb J. Dihu <macgyver@tos.net>
- * Copyright (c) 2001-2004 The ProFTPD Project team
+ * Copyright (c) 2001-2008 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -25,7 +25,7 @@
  */
 
 /* BSD socket manipulation tools.
- * $Id: inet.h,v 1.27 2005/02/07 18:15:49 castaglia Exp $
+ * $Id: inet.h,v 1.30 2008/10/28 00:46:11 castaglia Exp $
  */
 
 #ifndef PR_INET_H
@@ -128,13 +128,13 @@ typedef struct conn_struc {
 
 } conn_t;
 
+/* XXX Find a better home in the headers for this function. */
+void pr_cmd_set_handler(void (*)(server_rec *, conn_t *)); 
 
 /* Prototypes */
 void pr_inet_clear(void);
 int pr_inet_reverse_dns(pool *, int);
 int pr_inet_getservport(pool *, const char *, const char *);
-char *pr_inet_validate(char *);
-char *pr_inet_fqdn(pool *, const char *);
 pr_netaddr_t *pr_inet_getaddr(pool *, const char *, array_header **);
 conn_t *pr_inet_copy_connection(pool *, conn_t*);
 conn_t *pr_inet_create_dup_connection(pool *, xaset_t *, int, pr_netaddr_t *);
@@ -158,8 +158,6 @@ int pr_inet_connect(pool *, conn_t *, pr_netaddr_t *, int);
 int pr_inet_connect_nowait(pool *, conn_t *, pr_netaddr_t *, int);
 int pr_inet_get_conn_info(conn_t *, int);
 conn_t *pr_inet_accept(pool *, conn_t *, conn_t *, int, int, unsigned char);
-conn_t *pr_inet_associate(pool *, conn_t *, pr_netaddr_t *,
-  pr_netio_stream_t *, pr_netio_stream_t *, int);
 conn_t *pr_inet_openrw(pool *, conn_t *, pr_netaddr_t *, int, int, int,
   int, int);
 
