@@ -2,7 +2,7 @@
  * ProFTPD - FTP server daemon
  * Copyright (c) 1997, 1998 Public Flood Software
  * Copyright (c) 1999, 2000 MacGyver aka Habeeb J. Dihu <macgyver@tos.net>
- * Copyright (c) 2001, 2002, 2003 The ProFTPD Project team
+ * Copyright (c) 2001-2008 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -27,7 +27,7 @@
 /* Logging, either to syslog or stderr, as well as debug logging
  * and debug levels.
  *
- * $Id: log.h,v 1.25 2005/06/14 01:23:44 castaglia Exp $
+ * $Id: log.h,v 1.28 2008/02/18 21:45:20 castaglia Exp $
  */
 
 #ifndef PR_LOG_H
@@ -65,11 +65,57 @@
 #define PR_LOG_SYSTEM_MODE	0640
 #define PR_LOG_XFER_MODE	0644
 
+#ifdef PR_USE_LASTLOG
+
+/* It is tempting to want to have these lastlog-related includes separated
+ * out into their own lastlog.h file.  However, on some systems, such a
+ * proftpd-specific lastlog.h file may collide with the system's lastlog.h
+ * file.  Ultimately it's an issue with the default search order of the
+ * system C preprocessor, not with us -- and not every installation has
+ * this problem.
+ *
+ * In the meantime, the most portable thing is to keep these lastlog-related
+ * includes in this file.  Yay portability.
+ */
+
+#ifdef HAVE_LASTLOG_H
+# include <lastlog.h>
+#endif
+
+#ifdef HAVE_LOGIN_H
+# include <login.h>
+#endif
+
+#ifdef HAVE_PATHS_H
+# include <paths.h>
+#endif
+
+#ifndef PR_LASTLOG_PATH
+# ifdef _PATH_LASTLOG
+#   define PR_LASTLOG_PATH      _PATH_LASTLOG
+# else
+#   ifdef LASTLOG_FILE
+#     define PR_LASTLOG_PATH    LASTLOG_FILE
+#   else
+#     error "Missing path to lastlog file (use --with-lastlog=PATH)"
+#   endif
+# endif
+#endif
+
+int log_lastlog(uid_t uid, const char *user_name, const char *tty,
+  pr_netaddr_t *remote_addr);
+#endif /* PR_USE_LASTLOG */
+
 int log_wtmp(char *, const char *, const char *, pr_netaddr_t *);
 
 /* file-based logging functions */
 int pr_log_openfile(const char *, int *, mode_t);
-int pr_log_writefile(int, const char *, const char *, ...);
+int pr_log_writefile(int, const char *, const char *, ...)
+#ifdef __GNUC__
+  __attribute__ ((format (printf, 3, 4)));
+#else
+  ;
+#endif
 
 /* syslog-based logging functions.  Note that the open/close functions are
  * not part of the public API; use the pr_log_pri() function to log via
@@ -79,9 +125,6 @@ void log_closesyslog(void);
 int log_opensyslog(const char *);
 void log_setfacility(int);
 
-/* Utilize gcc's __attribute__ pragma for signalling that it should perform
- * printf-style checking of this function's arguments.
- */
 void pr_log_pri(int, const char *, ...)
 #ifdef __GNUC__
        __attribute__ ((format (printf, 2, 3)));
@@ -89,11 +132,13 @@ void pr_log_pri(int, const char *, ...)
        ;
 #endif
 
-void pr_log_auth(int, const char *, ...);
+void pr_log_auth(int, const char *, ...)
+#ifdef __GNUC__
+       __attribute__ ((format (printf, 2, 3)));
+#else
+       ;
+#endif
 
-/* Utilize gcc's __attribute__ pragma for signalling that it should perform
- * printf-style checking of this function's arguments.
- */
 void pr_log_debug(int, const char *, ...)
 #ifdef __GNUC__
        __attribute__ ((format (printf, 2, 3)));
