@@ -2,7 +2,7 @@
  * ProFTPD: mod_delay -- a module for adding arbitrary delays to the FTP
  *                       session lifecycle
  *
- * Copyright (c) 2004-2007 TJ Saunders
+ * Copyright (c) 2004-2009 TJ Saunders
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -26,7 +26,7 @@
  * This is mod_delay, contrib software for proftpd 1.2.10 and above.
  * For more information contact TJ Saunders <tj@castaglia.org>.
  *
- * $Id: mod_delay.c,v 1.26 2007/10/01 21:41:28 castaglia Exp $
+ * $Id: mod_delay.c,v 1.30 2009/01/04 01:14:37 castaglia Exp $
  */
 
 #include "conf.h"
@@ -351,7 +351,7 @@ static int delay_table_init(void) {
 
       } else {
         pr_log_pri(PR_LOG_WARNING, MOD_DELAY_VERSION
-          "warning: unable to obtain write lock on DelayTable '%s': %s",
+          ": unable to obtain write lock on DelayTable '%s': %s",
           fh->fh_path, strerror(errno));
         pr_trace_msg("delay", 1,
           "unable to obtain write lock on DelayTable '%s': %s", fh->fh_path,
@@ -367,7 +367,21 @@ static int delay_table_init(void) {
      * mmap() call.
      */
     lseek(fh->fh_fd, tab_size-1, SEEK_SET);
-    (void) write(fh->fh_fd, "", 1);
+    if (write(fh->fh_fd, "", 1) != 1) {
+      int xerrno = errno;
+
+      pr_log_pri(PR_LOG_WARNING, MOD_DELAY_VERSION
+        ": error writing single byte to DelayTable '%s': %s", fh->fh_path,
+        strerror(errno));
+      pr_trace_msg("delay", 1,
+        "error writing single byte to DelayTable '%s': %s", fh->fh_path,
+        strerror(errno));
+
+      pr_fsio_close(fh);
+
+      errno = xerrno;
+      return -1;
+    }
 
     /* Truncate the table, in case we're shrinking an existing table. */
     pr_fsio_ftruncate(fh, tab_size);
@@ -388,12 +402,20 @@ static int delay_table_init(void) {
     MAP_SHARED, delay_tab.dt_fd, 0);
 
   if (delay_tab.dt_data == MAP_FAILED) {
+    int xerrno = errno;
+
+    delay_tab.dt_data = NULL;
+
     pr_log_pri(PR_LOG_ERR, MOD_DELAY_VERSION
       ": error mapping DelayTable '%s' into memory: %s", delay_tab.dt_path,
-      strerror(errno));
+      strerror(xerrno));
     pr_trace_msg("delay", 1, "error mapping DelayTable '%s' into memory: %s",
-      delay_tab.dt_path, strerror(errno));
+      delay_tab.dt_path, strerror(xerrno));
+
     pr_fsio_close(fh);
+    delay_tab.dt_fd = -1;
+
+    errno = xerrno;
     return -1;
   }
 
@@ -445,7 +467,7 @@ static int delay_table_init(void) {
 
       } else {
         pr_log_pri(PR_LOG_WARNING, MOD_DELAY_VERSION
-          "warning: unable to obtain write lock on DelayTable '%s': %s",
+          ": unable to obtain write lock on DelayTable '%s': %s",
           fh->fh_path, strerror(errno));
         pr_trace_msg("delay", 1,
           "unable to obtain write lock on DelayTable '%s': %s", fh->fh_path,
@@ -461,7 +483,21 @@ static int delay_table_init(void) {
      * mmap() call.
      */
     lseek(fh->fh_fd, tab_size-1, SEEK_SET);
-    (void) write(fh->fh_fd, "", 1);
+    if (write(fh->fh_fd, "", 1) != 1) {
+      int xerrno = errno;
+
+      pr_log_pri(PR_LOG_WARNING, MOD_DELAY_VERSION
+        ": error writing single byte to DelayTable '%s': %s", fh->fh_path,
+        strerror(xerrno));
+      pr_trace_msg("delay", 1,
+        "error writing single byte to DelayTable '%s': %s", fh->fh_path,
+        strerror(xerrno));
+
+      pr_fsio_close(fh);
+
+      errno = xerrno;
+      return -1;
+    }
 
     /* Truncate the table, in case we're shrinking an existing table. */
     pr_fsio_ftruncate(fh, tab_size);
@@ -492,6 +528,7 @@ static int delay_table_init(void) {
   }
 
   delay_tab.dt_data = NULL;
+  delay_tab.dt_fd = -1;
 
   if (pr_fsio_close(fh) < 0) {
     pr_log_pri(PR_LOG_WARNING, MOD_DELAY_VERSION
@@ -502,7 +539,6 @@ static int delay_table_init(void) {
     return -1;
   }
 
-  delay_tab.dt_fd = -1;
   return 0;
 }
 
@@ -527,15 +563,17 @@ static int delay_table_load(int lock_table) {
     }
   }
 
-  if (!delay_tab.dt_data) {
+  if (delay_tab.dt_data == NULL) {
     pr_trace_msg("delay", 8, "mapping DelayTable '%s' (%" PR_LU
       " bytes, fd %d) into memory", delay_tab.dt_path,
       (pr_off_t) delay_tab.dt_size, delay_tab.dt_fd);
     delay_tab.dt_data = mmap(NULL, delay_tab.dt_size, PROT_READ|PROT_WRITE,
       MAP_SHARED, delay_tab.dt_fd, 0);
 
-    if (delay_tab.dt_data == MAP_FAILED)
+    if (delay_tab.dt_data == MAP_FAILED) {
+      delay_tab.dt_data = NULL;
       return -1;
+    }
   }
 
   return 0;
@@ -678,16 +716,19 @@ static int delay_handle_info(pr_ctrls_t *ctrl, int reqargc,
   delay_tab.dt_data = NULL;
 
   if (delay_table_load(TRUE) < 0) {
+    int xerrno = errno;
+
     pr_ctrls_add_response(ctrl,
       "unable to load DelayTable '%s' into memory: %s",
-      delay_tab.dt_path, strerror(errno));
+      delay_tab.dt_path, strerror(xerrno));
     pr_trace_msg("delay", 1, "unable to load DelayTable '%s' into memory: %s",
-      delay_tab.dt_path, strerror(errno));
+      delay_tab.dt_path, strerror(xerrno));
 
     pr_fsio_close(fh);
     delay_tab.dt_fd = -1;
     delay_tab.dt_data = NULL;
 
+    errno = xerrno;
     return -1;
   }
 
@@ -710,7 +751,7 @@ static int delay_handle_info(pr_ctrls_t *ctrl, int reqargc,
       char buf[80];
 
       memset(buf, '\0', sizeof(buf));
-      sprintf(buf, "%10ld", row->d_vals[DELAY_NVALUES - i]);
+      snprintf(buf, sizeof(buf)-1, "%10ld", row->d_vals[DELAY_NVALUES - i]);
 
       vals = pstrcat(tmp_pool, vals, " ", buf, NULL);
 
@@ -736,7 +777,7 @@ static int delay_handle_info(pr_ctrls_t *ctrl, int reqargc,
       char buf[80];
 
       memset(buf, '\0', sizeof(buf));
-      sprintf(buf, "%10ld", row->d_vals[DELAY_NVALUES - i]);
+      snprintf(buf, sizeof(buf)-1, "%10ld", row->d_vals[DELAY_NVALUES - i]);
 
       vals = pstrcat(tmp_pool, vals, " ", buf, NULL);
 
@@ -946,11 +987,15 @@ MODRET delay_post_pass(cmd_rec *cmd) {
 
   /* Prepare for manipulating the table. */
   if (delay_table_load(FALSE) < 0) {
+    int xerrno = errno;
+
     pr_log_pri(PR_LOG_WARNING, MOD_DELAY_VERSION
-      "warning: unable to load DelayTable '%s' into memory: %s",
-      delay_tab.dt_path, strerror(errno));
+      ": unable to load DelayTable '%s' into memory: %s",
+      delay_tab.dt_path, strerror(xerrno));
     pr_trace_msg("delay", 1, "unable to load DelayTable '%s' into memory: %s",
-      delay_tab.dt_path, strerror(errno));
+      delay_tab.dt_path, strerror(xerrno));
+
+    errno = xerrno;
     return PR_DECLINED(cmd);
   }
 
@@ -973,17 +1018,18 @@ MODRET delay_post_pass(cmd_rec *cmd) {
     delay_table_add_interval(rownum, interval);
     delay_npass++;
 
-  } else
+  } else {
     /* Generate an event, in case a module (i.e. mod_ban) might want
      * to do something appropriate to such an ill-behaved client.
      */
     pr_event_generate("mod_delay.max-pass", session.c);
+  }
 
   /* Done with the table. */
   delay_table_unlock(rownum);
   if (delay_table_unload(FALSE) < 0) {
     pr_log_pri(PR_LOG_WARNING, MOD_DELAY_VERSION
-      "warning: unable to unload DelayTable '%s' from memory: %s",
+      ": unable to unload DelayTable '%s' from memory: %s",
       delay_tab.dt_path, strerror(errno));
   }
 
@@ -1019,11 +1065,15 @@ MODRET delay_post_user(cmd_rec *cmd) {
 
   /* Prepare for manipulating the table. */
   if (delay_table_load(FALSE) < 0) {
+    int xerrno = errno;
+
     pr_log_pri(PR_LOG_WARNING, MOD_DELAY_VERSION
-      "warning: unable to load DelayTable '%s' into memory: %s",
-      delay_tab.dt_path, strerror(errno));
+      ": unable to load DelayTable '%s' into memory: %s",
+      delay_tab.dt_path, strerror(xerrno));
     pr_trace_msg("delay", 1, "unable to load DelayTable '%s' into memory: %s",
-      delay_tab.dt_path, strerror(errno));
+      delay_tab.dt_path, strerror(xerrno));
+
+    errno = xerrno;
     return PR_DECLINED(cmd);
   }
 
@@ -1056,7 +1106,7 @@ MODRET delay_post_user(cmd_rec *cmd) {
   delay_table_unlock(rownum);
   if (delay_table_unload(FALSE) < 0) {
     pr_log_pri(PR_LOG_WARNING, MOD_DELAY_VERSION
-      "warning: unable to unload DelayTable '%s' from memory: %s",
+      ": unable to unload DelayTable '%s' from memory: %s",
       delay_tab.dt_path, strerror(errno));
   }
 
@@ -1098,7 +1148,7 @@ static void delay_exit_ev(const void *event_data, void *user_data) {
 
   if (!fh) {
     pr_log_pri(PR_LOG_WARNING, MOD_DELAY_VERSION
-      ": warning: unable to open DelayTable '%s': %s", delay_tab.dt_path,
+      ": unable to open DelayTable '%s': %s", delay_tab.dt_path,
       strerror(errno));
     return;
   }
@@ -1112,7 +1162,7 @@ static void delay_exit_ev(const void *event_data, void *user_data) {
 
     errno = xerrno;
     pr_log_pri(PR_LOG_WARNING, MOD_DELAY_VERSION
-      "warning: unable to load DelayTable '%s' into memory: %s",
+      ": unable to load DelayTable '%s' into memory: %s",
       delay_tab.dt_path, strerror(errno));
     pr_trace_msg("delay", 1, "unable to load DelayTable '%s' into memory: %s",
       delay_tab.dt_path, strerror(errno));
@@ -1125,19 +1175,20 @@ static void delay_exit_ev(const void *event_data, void *user_data) {
 
   if (delay_table_unload(TRUE) < 0) {
     pr_log_pri(PR_LOG_WARNING, MOD_DELAY_VERSION
-      ": warning: error unloading DelayTable '%s' from memory: %s",
+      ": error unloading DelayTable '%s' from memory: %s",
       delay_tab.dt_path, strerror(errno));
   }
 
   if (pr_fsio_write(fh, data, datalen) < 0) {
     pr_log_pri(PR_LOG_WARNING, MOD_DELAY_VERSION
-      ": warning: error updating DelayTable '%s': %s", delay_tab.dt_path,
+      ": error updating DelayTable '%s': %s", delay_tab.dt_path,
       strerror(errno));
   }
 
+  delay_tab.dt_fd = -1;
   if (pr_fsio_close(fh) < 0) {
     pr_log_pri(PR_LOG_WARNING, MOD_DELAY_VERSION
-      ": warning: error writing DelayTable '%s': %s", delay_tab.dt_path,
+      ": error writing DelayTable '%s': %s", delay_tab.dt_path,
       strerror(errno));
   }
 
@@ -1235,7 +1286,7 @@ static int delay_sess_init(void) {
 
   if (!fh) {
     pr_log_pri(PR_LOG_WARNING, MOD_DELAY_VERSION
-      ": warning: unable to open DelayTable '%s': %s", delay_tab.dt_path,
+      ": unable to open DelayTable '%s': %s", delay_tab.dt_path,
       strerror(errno));
     pr_trace_msg("delay", 1, "unable to open DelayTable '%s': %s",
       delay_tab.dt_path, strerror(errno));
