@@ -1,7 +1,7 @@
 /*
  * ProFTPD: mod_dso -- support for loading/unloading modules at run-time
  *
- * Copyright (c) 2004-2007 TJ Saunders <tj@castaglia.org>
+ * Copyright (c) 2004-2008 TJ Saunders <tj@castaglia.org>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -25,12 +25,14 @@
  * This is mod_dso, contrib software for proftpd 1.3.x.
  * For more information contact TJ Saunders <tj@castaglia.org>.
  *
- * $Id: mod_dso.c,v 1.13 2007/09/13 14:57:59 castaglia Exp $
+ * $Id: mod_dso.c,v 1.16 2008/10/14 23:23:09 castaglia Exp $
  */
 
 #include "conf.h"
 #include "mod_ctrls.h"
-#include "ltdl.h"
+
+/* Make sure we use the libltdl shipped with proftpd, not the system libltdl. */
+#include "lib/libltdl/ltdl.h"
 
 #define MOD_DSO_VERSION		"mod_dso/0.4"
 
@@ -87,6 +89,8 @@ static int dso_load_module(char *name) {
     return -1;
   }
 
+  pr_log_debug(DEBUG7, "loading '%s'", name);
+
   tmp = strrchr(name, '.');
   if (!tmp) {
     errno = EINVAL;
@@ -597,6 +601,17 @@ static void dso_restart_ev(const void *event_data, void *user_data) {
 /* Initialization routines
  */
 
+/* We should be using the LTDL_SET_PRELOADED_SYMBOLS macro provided by
+ * the ltdl.h header.  However, doing so resulted in compiler warnings
+ * about "nested extern declaraction of lt_preloaded_symbols".  To
+ * work around these warnings, we will use the expanded version of the
+ * macro directly.
+ *
+ * By the way, it appears that this lt_preloaded_symbols list is defined
+ * at link-time by the libtool script.
+ */
+extern const lt_dlsymlist lt_preloaded_symbols[];
+
 static int dso_init(void) {
 #ifdef PR_USE_CTRLS
   register unsigned int i = 0;
@@ -606,7 +621,7 @@ static int dso_init(void) {
   dso_pool = make_sub_pool(permanent_pool);
   pr_pool_tag(dso_pool, MOD_DSO_VERSION);
 
-  LTDL_SET_PRELOADED_SYMBOLS();
+  lt_dlpreload_default(lt_preloaded_symbols);
 
   /* Initialize libltdl. */
   if (lt_dlinit() < 0) {
