@@ -2,7 +2,7 @@
  * ProFTPD - FTP server daemon
  * Copyright (c) 1997, 1998 Public Flood Software
  * Copyright (c) 1999, 2000 MacGyver aka Habeeb J. Dihu <macgyver@tos.net>
- * Copyright (c) 2001-2007 The ProFTPD Project team
+ * Copyright (c) 2001-2009 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -25,7 +25,7 @@
  */
 
 /* Flexible logging module for proftpd
- * $Id: mod_log.c,v 1.83 2007/09/11 00:47:51 castaglia Exp $
+ * $Id: mod_log.c,v 1.88 2009/02/03 17:42:05 castaglia Exp $
  */
 
 #include "conf.h"
@@ -98,6 +98,9 @@ struct logfile_struc {
 #define META_RESPONSE_STR	26
 #define META_VERSION		27
 
+/* For tracking the size of deleted files. */
+static off_t log_dele_filesz = 0;
+
 static pool			*log_pool;
 static logformat_t		*formats = NULL;
 static xaset_t			*format_set = NULL;
@@ -566,10 +569,14 @@ static char *get_next_meta(pool *p, cmd_rec *cmd, unsigned char **f) {
 
     case META_BYTES_SENT:
       argp = arg;
-      if (session.xfer.p)
+      if (session.xfer.p) {
         snprintf(argp, sizeof(arg), "%" PR_LU,
           (pr_off_t) session.xfer.total_bytes);
-      else
+
+      } else if (strcmp(cmd->argv[0], C_DELE) == 0) {
+        snprintf(argp, sizeof(arg), "%" PR_LU, (pr_off_t) log_dele_filesz);
+
+      } else
         sstrncpy(argp, "-", sizeof(arg));
 
       m++;
@@ -593,9 +600,12 @@ static char *get_next_meta(pool *p, cmd_rec *cmd, unsigned char **f) {
           strcmp(cmd->argv[0], C_XCUP) == 0 ||
           strcmp(cmd->argv[0], C_XMKD) == 0 ||
           strcmp(cmd->argv[0], C_XRMD) == 0) {
-        char *tmp = strrchr(cmd->arg, '/');
+        char *path, *tmp;
 
-        sstrncpy(argp, tmp ? tmp : cmd->arg, sizeof(arg));
+        path = pr_fs_decode_path(p, cmd->arg);
+        tmp = strrchr(path, '/');
+
+        sstrncpy(argp, tmp ? tmp : path, sizeof(arg));
 
       } else {
         sstrncpy(argp, "", sizeof(arg));
@@ -613,7 +623,8 @@ static char *get_next_meta(pool *p, cmd_rec *cmd, unsigned char **f) {
           strcmp(cmd->argv[0], C_XCUP) == 0 ||
           strcmp(cmd->argv[0], C_XMKD) == 0 ||
           strcmp(cmd->argv[0], C_XRMD) == 0) {
-        sstrncpy(argp, dir_abs_path(p, cmd->arg, TRUE), sizeof(arg));
+        sstrncpy(argp, dir_abs_path(p, pr_fs_decode_path(p, cmd->arg), TRUE),
+          sizeof(arg));
 
       } else if (strcmp(cmd->argv[0], C_CWD) == 0 ||
                  strcmp(cmd->argv[0], C_XCWD) == 0) {
@@ -647,12 +658,26 @@ static char *get_next_meta(pool *p, cmd_rec *cmd, unsigned char **f) {
       argp = arg;
 
       if (strcmp(cmd->argv[0], C_RNTO) == 0) {
-        sstrncpy(argp, dir_abs_path(p, cmd->arg, TRUE), sizeof(arg));
+        sstrncpy(argp, dir_abs_path(p, pr_fs_decode_path(p, cmd->arg), TRUE),
+          sizeof(arg));
 
       } else if (session.xfer.p &&
                  session.xfer.path) {
         sstrncpy(argp, dir_abs_path(p, session.xfer.path, TRUE), sizeof(arg));
 
+      } else if (strcmp(cmd->argv[0], C_SITE) == 0 &&
+                 (strcasecmp(cmd->argv[1], "CHGRP") == 0 ||
+                  strcasecmp(cmd->argv[1], "CHMOD") == 0)) {
+        register unsigned int i;
+        char *tmp = "";
+
+        for (i = 3; i <= cmd->argc-1; i++) {
+          tmp = pstrcat(cmd->tmp_pool, tmp, *tmp ? " " : "",
+            pr_fs_decode_path(cmd->tmp_pool, cmd->argv[i]), NULL);
+        }
+
+        sstrncpy(argp, dir_abs_path(p, tmp, TRUE), sizeof(arg));
+
       } else {
         /* Some commands (i.e. DELE, MKD, RMD, XMKD, and XRMD) have associated
          * filenames that are not stored in the session.xfer structure; these
@@ -663,7 +688,8 @@ static char *get_next_meta(pool *p, cmd_rec *cmd, unsigned char **f) {
             strcmp(cmd->argv[0], C_RMD) == 0 ||
             strcmp(cmd->argv[0], C_XMKD) == 0 ||
             strcmp(cmd->argv[0], C_XRMD) == 0)
-          sstrncpy(arg, dir_abs_path(p, cmd->arg, TRUE), sizeof(arg));
+          sstrncpy(arg, dir_abs_path(p, pr_fs_decode_path(p, cmd->arg), TRUE),
+            sizeof(arg));
 
         else
           /* All other situations get a "-".  */
@@ -675,8 +701,16 @@ static char *get_next_meta(pool *p, cmd_rec *cmd, unsigned char **f) {
 
     case META_XFER_PATH:
       argp = arg;
-      if (session.xfer.p &&
-          session.xfer.path) {
+
+      if (strcmp(cmd->argv[0], C_RNTO) == 0) {
+        char *path;
+
+        path = dir_best_path(cmd->tmp_pool,
+          pr_fs_decode_path(cmd->tmp_pool, cmd->arg));
+        sstrncpy(arg, path, sizeof(arg));
+
+      } else if (session.xfer.p &&
+                 session.xfer.path) {
         sstrncpy(argp, session.xfer.path, sizeof(arg));
 
       } else {
@@ -684,10 +718,14 @@ static char *get_next_meta(pool *p, cmd_rec *cmd, unsigned char **f) {
          * stored in the session.xfer structure; these should be expanded
          * properly as well.
          */
-        if (strcmp(cmd->argv[0], C_DELE) == 0)
-          sstrncpy(arg, cmd->arg, sizeof(arg));
+        if (strcmp(cmd->argv[0], C_DELE) == 0) {
+          char *path;
 
-        else
+          path = dir_best_path(cmd->tmp_pool,
+            pr_fs_decode_path(cmd->tmp_pool, cmd->arg));
+          sstrncpy(arg, path, sizeof(arg));
+
+        } else
           sstrncpy(argp, "-", sizeof(arg));
       }
 
@@ -724,15 +762,40 @@ static char *get_next_meta(pool *p, cmd_rec *cmd, unsigned char **f) {
       m++;
       break;
 
-    case META_IDENT_USER:
+    case META_IDENT_USER: {
+      char *rfc1413_ident;
+
       argp = arg;
-      sstrncpy(argp, session.ident_user, sizeof(arg));
+      rfc1413_ident = pr_table_get(session.notes, "mod_ident.rfc1413-ident",
+        NULL);
+      if (rfc1413_ident == NULL)
+        rfc1413_ident = "UNKNOWN";
+
+      sstrncpy(argp, rfc1413_ident, sizeof(arg));
       m++;
       break;
+    }
 
     case META_METHOD:
       argp = arg;
-      sstrncpy(argp, cmd->argv[0], sizeof(arg));
+
+      if (strcmp(cmd->argv[0], C_SITE) != 0) {
+        sstrncpy(argp, cmd->argv[0], sizeof(arg));
+
+      } else {
+        char *ptr;
+
+        /* Make sure that the SITE command used is all in uppercase,
+         * for logging purposes.
+         */
+
+        for (ptr = cmd->argv[1]; *ptr; ptr++) {
+          *ptr = toupper((int) *ptr);
+        }
+
+        snprintf(argp, sizeof(arg), "%s %s", cmd->argv[0], cmd->argv[1]);
+      }
+
       m++;
       break;
 
@@ -853,7 +916,7 @@ static char *get_next_meta(pool *p, cmd_rec *cmd, unsigned char **f) {
         sstrncpy(argp, "(hidden)", sizeof(arg));
 
       } else {
-        sstrncpy(argp, cmd->arg, sizeof(arg));
+        sstrncpy(argp, pr_fs_decode_path(p, cmd->arg), sizeof(arg));
       }
 
       m++;
@@ -869,7 +932,7 @@ static char *get_next_meta(pool *p, cmd_rec *cmd, unsigned char **f) {
       argp = arg;
 
       if (!session.user) {
-        char *u = get_param_ptr(cmd->server->conf,"UserName",FALSE);
+        char *u = get_param_ptr(cmd->server->conf, "UserName", FALSE);
         if (!u)
           u = "root";
 
@@ -1117,6 +1180,27 @@ loop_extendedlogs:
   }
 }
 
+MODRET log_pre_dele(cmd_rec *cmd) {
+  char *path;
+
+  log_dele_filesz = 0;
+
+  path = dir_canonical_path(cmd->tmp_pool,
+    pr_fs_decode_path(cmd->tmp_pool, cmd->arg));
+  if (path) {
+    struct stat st;
+
+    /* Briefly cache the size of the file being deleted, so that it can be
+     * logged properly using %b.
+     */
+    pr_fs_clear_cache();
+    if (pr_fsio_stat(path, &st) == 0)
+      log_dele_filesz = st.st_size;
+  }
+
+  return PR_DECLINED(cmd);
+}
+
 MODRET log_post_pass(cmd_rec *cmd) {
   logfile_t *lf;
 
@@ -1263,6 +1347,7 @@ static conftable log_conftab[] = {
 };
 
 static cmdtable log_cmdtab[] = {
+  { PRE_CMD,		C_DELE,	G_NONE,	log_pre_dele,	FALSE, FALSE },
   { LOG_CMD,		C_ANY,	G_NONE,	log_any,		FALSE, FALSE },
   { LOG_CMD_ERR,	C_ANY,	G_NONE,	log_any,		FALSE, FALSE },
   { POST_CMD,		C_PASS,	G_NONE,	log_post_pass,		FALSE, FALSE },
