@@ -2,7 +2,7 @@
  * ProFTPD - FTP server daemon
  * Copyright (c) 1997, 1998 Public Flood Software
  * Copyright (c) 1999, 2000 MacGyver aka Habeeb J. Dihu <macgyver@tos.net>
- * Copyright (c) 2001-2004 The ProFTPD Project team
+ * Copyright (c) 2001-2009 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -25,7 +25,7 @@
  */
 
 /* Directory listing module for ProFTPD.
- * $Id: mod_ls.c,v 1.128 2005/08/24 16:10:30 castaglia Exp $
+ * $Id: mod_ls.c,v 1.152.2.1 2009/04/28 18:35:03 castaglia Exp $
  */
 
 #include "conf.h"
@@ -46,6 +46,14 @@ static int outputfiles(cmd_rec *);
 static int listfile(cmd_rec *, pool *, const char *);
 static int listdir(cmd_rec *, pool *, const char *);
 
+static int sendline(int flags, char *fmt, ...)
+#ifdef __GNUC__
+       __attribute__ ((format (printf, 2, 3)));
+#else
+       ;
+#endif
+#define LS_SENDLINE_FL_FLUSH	0x0001
+
 static unsigned char list_strict_opts = FALSE;
 static char *list_options = NULL;
 static unsigned char list_show_symlinks = TRUE, list_times_gmt = TRUE;
@@ -186,7 +194,8 @@ static int ls_perms_full(pool *p, cmd_rec *cmd, const char *path, int *hidden) {
       list_show_symlinks = *tmp;
   }
 
-  if ((fake_mode = get_param_ptr(CURRENT_CONF, "DirFakeMode", FALSE))) {
+  fake_mode = get_param_ptr(CURRENT_CONF, "DirFakeMode", FALSE);
+  if (fake_mode) {
     fakemode = *fake_mode;
     have_fake_mode = TRUE;
 
@@ -224,7 +233,8 @@ static int ls_perms(pool *p, cmd_rec *cmd, const char *path,int *hidden) {
       list_show_symlinks = *tmp;
   }
 
-  if ((fake_mode = get_param_ptr(CURRENT_CONF, "DirFakeMode", FALSE))) {
+  fake_mode = get_param_ptr(CURRENT_CONF, "DirFakeMode", FALSE);
+  if (fake_mode) {
     fakemode = *fake_mode;
     have_fake_mode = TRUE;
 
@@ -235,19 +245,26 @@ static int ls_perms(pool *p, cmd_rec *cmd, const char *path,int *hidden) {
 }
 
 /* sendline() now has an internal buffer, to help speed up LIST output. */
-static int sendline(char *fmt, ...) {
+static int sendline(int flags, char *fmt, ...) {
   static char listbuf[PR_TUNABLE_BUFFER_SIZE] = {'\0'};
   va_list msg;
   char buf[PR_TUNABLE_BUFFER_SIZE+1] = {'\0'};
   int res = 0;
 
-  /* A NULL fmt argument is the signal to flush the buffer */
-  if (!fmt) {
-    if ((res = pr_data_xfer(listbuf, strlen(listbuf))) < 0)
-      pr_log_debug(DEBUG3, "pr_data_xfer returned %d, error = %s.", res,
-        strerror(PR_NETIO_ERRNO(session.d->outstrm)));
+  if (flags & LS_SENDLINE_FL_FLUSH) {
+    size_t listbuflen = strlen(listbuf);
+
+    if (listbuflen > 0) {
+      res = pr_data_xfer(listbuf, listbuflen);
+      if (res < 0 &&
+          errno != 0) {
+        pr_log_debug(DEBUG3, "pr_data_xfer returned %d, error = %s.", res,
+          strerror(PR_NETIO_ERRNO(session.d->outstrm)));
+      }
+
+      memset(listbuf, '\0', sizeof(listbuf));
+    }
 
-    memset(listbuf, '\0', sizeof(listbuf));
     return res;
   }
 
@@ -259,9 +276,12 @@ static int sendline(char *fmt, ...) {
 
   /* If buf won't fit completely into listbuf, flush listbuf */
   if (strlen(buf) >= (sizeof(listbuf) - strlen(listbuf))) {
-    if ((res = pr_data_xfer(listbuf, strlen(listbuf))) < 0)
+    res = pr_data_xfer(listbuf, strlen(listbuf));
+    if (res < 0 &&
+        errno != 0) {
       pr_log_debug(DEBUG3, "pr_data_xfer returned %d, error = %s.", res,
         strerror(PR_NETIO_ERRNO(session.d->outstrm)));
+    }
 
     memset(listbuf, '\0', sizeof(listbuf));
   }
@@ -271,7 +291,12 @@ static int sendline(char *fmt, ...) {
 }
 
 static void ls_done(cmd_rec *cmd) {
-  pr_data_close(FALSE);
+  int quiet = FALSE;
+
+  if (session.sf_flags & SF_ABORT)
+    quiet = TRUE;
+
+  pr_data_close(quiet);
 }
 
 static char units[6][2] = 
@@ -335,10 +360,13 @@ static int listfile(cmd_rec *cmd, pool *p, const char *name) {
       if (pr_fsio_stat(name, &l_st) != -1) {
         memcpy(&st, &l_st, sizeof(struct stat));
 
-        len = pr_fsio_readlink(name, m, sizeof(m));
+        len = pr_fsio_readlink(name, m, sizeof(m) - 1);
         if (len < 0)
           return 0;
 
+        if (len >= sizeof(m))
+          return 0;
+
         m[len] = '\0';
 
         if (!ls_perms_full(p, cmd, m, NULL))
@@ -348,10 +376,13 @@ static int listfile(cmd_rec *cmd, pool *p, const char *name) {
         return 0;
 
     } else if (S_ISLNK(st.st_mode)) {
-      len = pr_fsio_readlink(name, l, sizeof(l));
+      len = pr_fsio_readlink(name, l, sizeof(l) - 1);
       if (len < 0)
         return 0;
 
+      if (len >= sizeof(l))
+        return 0;
+
       l[len] = '\0';
 
       if (!ls_perms_full(p, cmd, l, &hidden))
@@ -375,11 +406,6 @@ static int listfile(cmd_rec *cmd, pool *p, const char *name) {
     else
       t = pr_localtime(p, (time_t *) &mtime);
 
-    if (!t) {
-      pr_response_add_err(R_421, "Fatal error (localtime() returned NULL?!?)");
-      return -1;
-    }
-
     if (opt_F) {
       if (S_ISLNK(st.st_mode))
         suffix[0] = '@';
@@ -473,7 +499,8 @@ static int listfile(cmd_rec *cmd, pool *p, const char *name) {
           snprintf(nameline, sizeof(nameline)-1,
             "%s %3d %-8s %-8s %s %s %2d %s %s", m, (int) st.st_nlink,
             MAP_UID(st.st_uid), MAP_GID(st.st_gid), s,
-            months[t->tm_mon], t->tm_mday, timeline, name);
+            months[t->tm_mon], t->tm_mday, timeline,
+            pr_fs_encode_path(cmd->tmp_pool, name));
 
         } else {
 
@@ -481,7 +508,8 @@ static int listfile(cmd_rec *cmd, pool *p, const char *name) {
           snprintf(nameline, sizeof(nameline)-1,
             "%s %3d %-8u %-8u %s %s %2d %s %s", m, (int) st.st_nlink,
             (unsigned) st.st_uid, (unsigned) st.st_gid, s,
-            months[t->tm_mon], t->tm_mday, timeline, name);
+            months[t->tm_mon], t->tm_mday, timeline,
+            pr_fs_encode_path(cmd->tmp_pool, name));
         }
 
         nameline[sizeof(nameline)-1] = '\0';
@@ -523,7 +551,8 @@ static int listfile(cmd_rec *cmd, pool *p, const char *name) {
       if (S_ISREG(st.st_mode) ||
           S_ISDIR(st.st_mode) ||
           S_ISLNK(st.st_mode))
-           addfile(cmd, name, suffix, mtime, st.st_size);
+           addfile(cmd, pr_fs_encode_path(cmd->tmp_pool, name), suffix,
+             mtime, st.st_size);
     }
   }
 
@@ -726,12 +755,6 @@ static int outputfiles(cmd_rec *cmd) {
   if (p && p->down)
     p->down = NULL;
 
-#if 0
-  if (opt_l)
-    if (sendline("total 0\n") < 0)
-      return -1;
-#endif
-
   p = head;
   while (p) {
     q = p;
@@ -747,7 +770,7 @@ static int outputfiles(cmd_rec *cmd) {
         sstrncpy(pad, "\n", sizeof(pad));
       }
 
-      if (sendline("%s%s", q->line, pad) < 0)
+      if (sendline(0, "%s%s", q->line, pad) < 0)
         return -1;
 
       q = q->right;
@@ -761,8 +784,7 @@ static int outputfiles(cmd_rec *cmd) {
   colwidth = 0;
   filenames = 0;
 
-  /* flush the buffer */
-  if (sendline(NULL) < 0)
+  if (sendline(LS_SENDLINE_FL_FLUSH, " ") < 0)
     return -1;
 
   return 0;
@@ -779,7 +801,11 @@ static void discard_output(void) {
 }
 
 static int dircmp(const void *a, const void *b) {
+#if defined(PR_USE_NLS) && defined(HAVE_STRCOLL)
+  return strcoll(*(const char **)a, *(const char **)b);
+#else
   return strcmp(*(const char **)a, *(const char **)b);
+#endif /* !PR_USE_NLS or !HAVE_STRCOLL */
 }
 
 static char **sreaddir(const char *dirname, const int sort) {
@@ -799,7 +825,8 @@ static char **sreaddir(const char *dirname, const int sort) {
     return NULL;
   }
 
-  if ((d = pr_fsio_opendir(dirname)) == NULL)
+  d = pr_fsio_opendir(dirname);
+  if (d == NULL)
     return NULL;
 
   /* It doesn't matter if the following guesses are wrong, but it slows
@@ -811,12 +838,13 @@ static char **sreaddir(const char *dirname, const int sort) {
    */
   dsize = (st.st_size / 4) + 10;	 /* Guess number of entries in dir */
 
-  /*
-  ** The directory has been opened already, but portably accessing the file
-  ** descriptor inside the DIR struct isn't easy.  Some systems use "dd_fd" or
-  ** "__dd_fd" rather than "d_fd".  Still others work really hard at opacity.
-  */
-#if defined(HAVE_STRUCT_DIR_D_FD)
+  /* The directory has been opened already, but portably accessing the file
+   * descriptor inside the DIR struct isn't easy.  Some systems use "dd_fd" or
+   * "__dd_fd" rather than "d_fd".  Still others work really hard at opacity.
+   */
+#if defined(HAVE_DIRFD) 
+  dir_fd = dirfd(d);
+#elif defined(HAVE_STRUCT_DIR_D_FD)
   dir_fd = d->d_fd;
 #elif defined(HAVE_STRUCT_DIR_DD_FD)
   dir_fd = d->dd_fd;
@@ -825,7 +853,9 @@ static char **sreaddir(const char *dirname, const int sort) {
 #else
   dir_fd = 0;
 #endif
-  if ((ssize = get_name_max((char *) dirname, dir_fd)) < 1 ) {
+
+  ssize = get_name_max((char *) dirname, dir_fd);
+  if (ssize < 1) {
     pr_log_debug(DEBUG1, "get_name_max(%s, %d) = %d, using %d", dirname,
       dir_fd, ssize, NAME_MAX_GUESS);
     ssize = NAME_MAX_GUESS;
@@ -847,6 +877,7 @@ static char **sreaddir(const char *dirname, const int sort) {
   i = 0;
 
   while ((de = pr_fsio_readdir(d)) != NULL) {
+    pr_signals_handle();
 
     if (i >= dsize - 1) {
       char **newp;
@@ -940,7 +971,8 @@ static int listdir(cmd_rec *cmd, pool *workp, const char *name) {
   /* Search for relevant <Limit>'s to this LIST command.  If found,
    * check to see whether hidden files should be ignored.
    */
-  if ((c = _find_ls_limit(cmd->argv[0])) != NULL) {
+  c = _find_ls_limit(cmd->argv[0]);
+  if (c != NULL) {
     unsigned char *ignore = get_param_ptr(c->subset, "IgnoreHidden", FALSE);
 
     if (ignore && *ignore == TRUE)
@@ -953,15 +985,6 @@ static int listdir(cmd_rec *cmd, pool *workp, const char *name) {
 
     int d = 0;
 
-#if 0
-    if (opt_l) {
-      if (opt_STAT)
-        pr_response_add(R_211, "total 0");
-      else if (sendline("total 0\n") < 0)
-        return -1;
-    }
-#endif
-
     s = dir;
     while (*s) {
       if (**s == '.') {
@@ -1064,10 +1087,12 @@ static int listdir(cmd_rec *cmd, pool *workp, const char *name) {
 
         if (opt_STAT) {
           pr_response_add(R_211, "%s", "");
-          pr_response_add(R_211, "%s:", subdir);
+          pr_response_add(R_211, "%s:",
+            pr_fs_encode_path(cmd->tmp_pool, subdir));
 
-        } else if (sendline("\n%s:\n", subdir) < 0 ||
-            sendline(NULL) < 0) {
+        } else if (sendline(0, "\n%s:\n",
+                     pr_fs_encode_path(cmd->tmp_pool, subdir)) < 0 ||
+            sendline(LS_SENDLINE_FL_FLUSH, " ") < 0) {
           pop_cwd(cwd_buf, &symhold);
 
           if (dest_workp)
@@ -1134,17 +1159,21 @@ static void ls_terminate(void) {
     ls_errno = 0;
 
   } else if (ls_errno) {
-    pr_response_add(R_211, "ERROR: %s", strerror(ls_errno));
+    pr_response_add(R_211, _("ERROR: %s"), strerror(ls_errno));
     ls_errno = 0;
   }
 }
 
 static void parse_list_opts(char **opt, int *glob_flags, int handle_plus_opts) {
-  while (isspace((int) **opt))
+  while (isspace((int) **opt)) {
+    pr_signals_handle();
     (*opt)++;
+  }
 
   /* Check for standard /bin/ls options */
   while (*opt && **opt == '-') {
+    pr_signals_handle();
+
     while ((*opt)++ && isalnum((int) **opt)) {
       switch (**opt) {
         case '1':
@@ -1219,8 +1248,10 @@ static void parse_list_opts(char **opt, int *glob_flags, int handle_plus_opts) {
       }
     }
 
-    while (isspace((int) **opt))
+    while (isspace((int) **opt)) {
+      pr_signals_handle();
       (*opt)++;
+    }
   }
 
   if (!handle_plus_opts)
@@ -1228,6 +1259,8 @@ static void parse_list_opts(char **opt, int *glob_flags, int handle_plus_opts) {
 
   /* Check for non-standard options */
   while (*opt && **opt == '+') {
+    pr_signals_handle();
+
     while ((*opt)++ && isalnum((int) **opt)) {
       switch (**opt) {
         case '1':
@@ -1290,8 +1323,10 @@ static void parse_list_opts(char **opt, int *glob_flags, int handle_plus_opts) {
       }
     }
 
-    while (isspace((int) **opt))
+    while (isspace((int) **opt)) {
+      pr_signals_handle();
       (*opt)++;
+    }
   }
 }
 
@@ -1306,7 +1341,7 @@ static int dolist(cmd_rec *cmd, const char *opt, int clearflags) {
   ls_curtime = time(NULL);
 
   if (clearflags)
-    opt_a = opt_C = opt_d = opt_F = opt_h = opt_n = opt_r = opt_R =
+    opt_A = opt_a = opt_C = opt_d = opt_F = opt_h = opt_n = opt_r = opt_R =
       opt_S = opt_t = opt_STAT = opt_L = 0;
 
   if (!list_strict_opts) {
@@ -1317,8 +1352,10 @@ static int dolist(cmd_rec *cmd, const char *opt, int clearflags) {
     /* Even if the user-given options are ignored, they still need to
      * "processed" (ie skip past options) in order to get to the paths.
      */
-    while (*arg && isspace((int) *arg))
+    while (*arg && isspace((int) *arg)) {
+      pr_signals_handle();
       arg++;
+    }
 
     while (arg && *arg == '-') {
 
@@ -1342,7 +1379,7 @@ static int dolist(cmd_rec *cmd, const char *opt, int clearflags) {
     glob_t g;
     int globbed = FALSE;
     int a;
-    char   pbuffer[PR_TUNABLE_PATH_MAX + 1] = "";
+    char pbuffer[PR_TUNABLE_PATH_MAX + 1] = "";
     char *target;
 
     /* Make sure the glob_t is initialized. */
@@ -1369,6 +1406,13 @@ static int dolist(cmd_rec *cmd, const char *opt, int clearflags) {
 
     target = *pbuffer ? pbuffer : arg;
 
+    /* Open data connection */
+    if (!opt_STAT) {
+      if (pr_data_open(NULL, "file list", PR_NETIO_IO_WR, 0) < 0)
+        return -1;
+      session.sf_flags |= SF_ASCII_OVERRIDE;
+    }
+
     /* If there are no globbing characters in the given target,
      * we can check to see if it even exists.
      */
@@ -1377,18 +1421,12 @@ static int dolist(cmd_rec *cmd, const char *opt, int clearflags) {
 
       pr_fs_clear_cache();
       if (pr_fsio_stat(target, &st) < 0) {
-        pr_response_add_err(R_450, "%s: %s", target, strerror(errno));
+        pr_response_add_err(R_450, "%s: %s",
+          pr_fs_encode_path(cmd->tmp_pool, target), strerror(errno));
         return -1;
       }
     }
 
-    /* Open data connection */
-    if (!opt_STAT) {
-      session.sf_flags |= SF_ASCII_OVERRIDE;
-      if (pr_data_open(NULL, "file list", PR_NETIO_IO_WR, 0) < 0)
-        return -1;
-    }
-
     /* Check perms on the directory/file we are about to scan. */
     if (!ls_perms_full(cmd->tmp_pool, cmd, target, NULL)) {
       a = -1;
@@ -1480,11 +1518,13 @@ static int dolist(cmd_rec *cmd, const char *opt, int clearflags) {
           if (!justone) {
             if (opt_STAT) {
               pr_response_add(R_211, "%s", "");
-              pr_response_add(R_211, "%s:", *path);
+              pr_response_add(R_211, "%s:",
+                pr_fs_encode_path(cmd->tmp_pool, *path));
 
             } else {
-              sendline("\n%s:\n", *path);
-              sendline(NULL);
+              sendline(0, "\n%s:\n",
+                pr_fs_encode_path(cmd->tmp_pool, *path));
+              sendline(LS_SENDLINE_FL_FLUSH, " ");
             }
           }
 
@@ -1523,13 +1563,16 @@ static int dolist(cmd_rec *cmd, const char *opt, int clearflags) {
 
     } else if (!skiparg) {
       if (a == GLOB_NOSPACE)
-        pr_response_add(R_226, "Out of memory during globbing of %s", arg);
+        pr_response_add(R_226, _("Out of memory during globbing of %s"),
+          pr_fs_encode_path(cmd->tmp_pool, arg));
 
       else if (a == GLOB_ABORTED)
-        pr_response_add(R_226, "Read error during globbing of %s", arg);
+        pr_response_add(R_226, _("Read error during globbing of %s"),
+          pr_fs_encode_path(cmd->tmp_pool, arg));
 
       else if (a != GLOB_NOMATCH)
-        pr_response_add(R_226, "Unknown error during globbing of %s", arg);
+        pr_response_add(R_226, _("Unknown error during globbing of %s"),
+          pr_fs_encode_path(cmd->tmp_pool, arg));
     }
 
     if (!skiparg && use_globbing && globbed)
@@ -1544,9 +1587,9 @@ static int dolist(cmd_rec *cmd, const char *opt, int clearflags) {
 
     /* Open data connection */
     if (!opt_STAT) {
-      session.sf_flags |= SF_ASCII_OVERRIDE;
       if (pr_data_open(NULL, "file list", PR_NETIO_IO_WR, 0) < 0)
         return -1;
+      session.sf_flags |= SF_ASCII_OVERRIDE;
     }
 
     if (ls_perms_full(cmd->tmp_pool, cmd, ".", NULL)) {
@@ -1584,22 +1627,22 @@ static int dolist(cmd_rec *cmd, const char *opt, int clearflags) {
 static int nlstfile(cmd_rec *cmd, const char *file) {
   int res = 0;
 
-  /* If the data connection isn't open, open it now. */
+  /* If the data connection isn't open, return an error */
   if ((session.sf_flags & SF_XFER) == 0) {
-    if (pr_data_open(NULL, "file list", PR_NETIO_IO_WR, 0) < 0) {
-      pr_data_reset();
-      return -1;
-    }
-
-    session.sf_flags |= SF_ASCII_OVERRIDE;
+    errno = EPERM;
+    return -1;
   }
 
   if (dir_hide_file(file))
     return 1;
 
   /* Be sure to flush the output */
-  if ((res = sendline("%s\n", file)) < 0 ||
-      (res = sendline(NULL)) < 0)
+  res = sendline(0, "%s\n", pr_fs_encode_path(cmd->tmp_pool, file));
+  if (res < 0)
+    return res;
+
+  res = sendline(LS_SENDLINE_FL_FLUSH, " ");
+  if (res < 0)
     return res;
 
   return 1;
@@ -1670,7 +1713,8 @@ static int nlstdir(cmd_rec *cmd, const char *dir) {
   /* Search for relevant <Limit>'s to this NLST command.  If found,
    * check to see whether hidden files should be ignored.
    */
-  if ((c = _find_ls_limit(cmd->argv[0])) != NULL) {
+  c = _find_ls_limit(cmd->argv[0]);
+  if (c != NULL) {
     unsigned char *ignore = get_param_ptr(c->subset, "IgnoreHidden", FALSE);
 
     if (ignore && *ignore == TRUE)
@@ -1690,7 +1734,11 @@ static int nlstdir(cmd_rec *cmd, const char *dir) {
         continue;
     }
 
-    if ((i = pr_fsio_readlink(p, file, sizeof(file))) > 0) {
+    i = pr_fsio_readlink(p, file, sizeof(file) - 1);
+    if (i > 0) {
+      if (i >= sizeof(file))
+        continue;
+
       file[i] = '\0';
       f = file;
 
@@ -1698,26 +1746,20 @@ static int nlstdir(cmd_rec *cmd, const char *dir) {
       f = p;
     }
 
-    if (ls_perms(workp, cmd, f, &hidden)) {
+    if (ls_perms(workp, cmd, dir_best_path(cmd->tmp_pool, f), &hidden)) {
       if (hidden)
         continue;
 
-      /* If the data connection isn't open, open it now. */
-      if ((session.sf_flags & SF_XFER) == 0) {
-        if (pr_data_open(NULL, "file list", PR_NETIO_IO_WR, 0) < 0) {
-          pr_data_reset();
-          count = -1;
-          continue;
-        }
-
-        session.sf_flags |= SF_ASCII_OVERRIDE;
-      }
-
-      if ((mode = file_mode(f)) == 0)
+      mode = file_mode(f);
+      if (mode == 0)
         continue;
 
       if (!curdir) {
-        if (sendline("%s/%s\n", dir, p) < 0 || sendline(NULL) < 0)
+        char *str = pr_fs_encode_path(cmd->tmp_pool,
+          pdircat(cmd->tmp_pool, dir, p, NULL));
+
+        if (sendline(0, "%s\n", str) < 0 ||
+            sendline(LS_SENDLINE_FL_FLUSH, " ") < 0)
           count = -1;
 
         else {
@@ -1738,7 +1780,8 @@ static int nlstdir(cmd_rec *cmd, const char *dir) {
         }
 
       } else {
-        if (sendline("%s\n", p) < 0 || sendline(NULL) < 0)
+        if (sendline(0, "%s\n", pr_fs_encode_path(cmd->tmp_pool, p)) < 0 ||
+            sendline(LS_SENDLINE_FL_FLUSH, " ") < 0)
           count = -1;
 
         else {
@@ -1824,7 +1867,8 @@ MODRET genericlist(cmd_rec *cmd) {
   if (fakegroup && strcmp(fakegroup, "~") == 0)
     fakegroup = session.group;
 
-  if ((fake_mode = get_param_ptr(CURRENT_CONF, "DirFakeMode", FALSE))) {
+  fake_mode = get_param_ptr(CURRENT_CONF, "DirFakeMode", FALSE);
+  if (fake_mode) {
     fakemode = *fake_mode;
     have_fake_mode = TRUE;
 
@@ -1835,7 +1879,7 @@ MODRET genericlist(cmd_rec *cmd) {
   if (tmp != NULL)
     list_times_gmt = *tmp;
 
-  res = dolist(cmd, cmd->arg, TRUE);
+  res = dolist(cmd, pr_fs_decode_path(cmd->tmp_pool, cmd->arg), TRUE);
 
   if (XFER_ABORTED) {
     pr_data_abort(0, 0);
@@ -1846,17 +1890,17 @@ MODRET genericlist(cmd_rec *cmd) {
 
   opt_l = 0;
 
-  return (res == -1 ? ERROR(cmd) : HANDLED(cmd));
+  return (res == -1 ? PR_ERROR(cmd) : PR_HANDLED(cmd));
 }
 
 MODRET ls_log_nlst(cmd_rec *cmd) {
   pr_data_cleanup();
-  return DECLINED(cmd);
+  return PR_DECLINED(cmd);
 }
 
 MODRET ls_err_nlst(cmd_rec *cmd) {
   pr_data_cleanup();
-  return DECLINED(cmd);
+  return PR_DECLINED(cmd);
 }
 
 MODRET ls_stat(cmd_rec *cmd) {
@@ -1875,28 +1919,27 @@ MODRET ls_stat(cmd_rec *cmd) {
     if (!dir_check(cmd->tmp_pool, cmd->argv[0], cmd->group, session.cwd,
         NULL)) {
       pr_response_add_err(R_500, "%s: %s", cmd->argv[0], strerror(EPERM));
-      return ERROR(cmd);
+      return PR_ERROR(cmd);
     }
 
-    pr_response_add(R_211, "Status of '%s'", main_server->ServerName);
-    pr_response_add(R_DUP, "Connected from %s (%s)", session.c->remote_name,
+    pr_response_add(R_211, _("Status of '%s'"), main_server->ServerName);
+    pr_response_add(R_DUP, _("Connected from %s (%s)"), session.c->remote_name,
       pr_netaddr_get_ipstr(session.c->remote_addr));
-    pr_response_add(R_DUP, "Logged in as %s", session.user);
-    pr_response_add(R_DUP, "TYPE: %s, STRUcture: File, Mode: Stream",
+    pr_response_add(R_DUP, _("Logged in as %s"), session.user);
+    pr_response_add(R_DUP, _("TYPE: %s, STRUcture: File, Mode: Stream"),
       (session.sf_flags & SF_ASCII) ? "ASCII" : "BINARY");
 
     if (session.total_bytes)
-      pr_response_add(R_DUP, "Total bytes transferred for session: %" PR_LU,
+      pr_response_add(R_DUP, _("Total bytes transferred for session: %" PR_LU),
         (pr_off_t) session.total_bytes);
 
     if (session.sf_flags & SF_XFER) {
-
       /* Report on the data transfer attributes.
        */
 
-      pr_response_add(R_DUP, "%s from %s port %u",
+      pr_response_add(R_DUP, _("%s from %s port %u"),
         (session.sf_flags & SF_PASSIVE) ?
-          "Passive data transfer from" : "Active data transfer to",
+          _("Passive data transfer from") : _("Active data transfer to"),
         pr_netaddr_get_ipstr(session.d->remote_addr), session.d->remote_port);
 
       if (session.xfer.file_size)
@@ -1911,23 +1954,28 @@ MODRET ls_stat(cmd_rec *cmd) {
           session.xfer.path, (pr_off_t) session.xfer.total_bytes);
 
     } else
-      pr_response_add(R_DUP, "No data connection");
+      pr_response_add(R_DUP, _("No data connection"));
 
-    pr_response_add(R_DUP, "End of status");
+    pr_response_add(R_DUP, _("End of status"));
 
-    return HANDLED(cmd);
+    return PR_HANDLED(cmd);
   }
 
   list_nfiles.curr = list_ndirs.curr = list_ndepth.curr = 0;
   list_nfiles.logged = list_ndirs.logged = list_ndepth.logged = FALSE;
 
+  arg = pr_fs_decode_path(cmd->tmp_pool, arg);
+
   /* Get to the actual argument. */
   if (*arg == '-')
-    while (arg && *arg && !isspace((int) *arg)) arg++;
+    while (arg && *arg && !isspace((int) *arg))
+      arg++;
 
-  while (arg && *arg && isspace((int) *arg)) arg++;
+  while (arg && *arg && isspace((int) *arg))
+    arg++;
 
-  if ((tmp = get_param_ptr(TOPLEVEL_CONF, "ShowSymlinks", FALSE)) != NULL)
+  tmp = get_param_ptr(TOPLEVEL_CONF, "ShowSymlinks", FALSE);
+  if (tmp != NULL)
     list_show_symlinks = *tmp;
 
   list_strict_opts = FALSE;
@@ -1965,23 +2013,27 @@ MODRET ls_stat(cmd_rec *cmd) {
   if (fakegroup && strcmp(fakegroup, "~") == 0)
     fakegroup = session.group;
 
-  if ((fake_mode = get_param_ptr(CURRENT_CONF, "DirFakeMode", FALSE))) {
+  fake_mode = get_param_ptr(CURRENT_CONF, "DirFakeMode", FALSE);
+  if (fake_mode) {
     fakemode = *fake_mode;
     have_fake_mode = TRUE;
 
   } else
     have_fake_mode = FALSE;
 
-  if ((tmp = get_param_ptr(TOPLEVEL_CONF, "TimesGMT", FALSE)) != NULL)
+  tmp = get_param_ptr(TOPLEVEL_CONF, "TimesGMT", FALSE);
+  if (tmp != NULL)
     list_times_gmt = *tmp;
 
   opt_C = opt_d = opt_F = opt_R = 0;
   opt_a = opt_l = opt_STAT = 1;
 
-  pr_response_add(R_211, "Status of %s:", arg && *arg ? arg : ".");
+  pr_response_add(R_211, _("Status of %s:"), arg && *arg ?
+    pr_fs_encode_path(cmd->tmp_pool, arg) :
+    pr_fs_encode_path(cmd->tmp_pool, "."));
   res = dolist(cmd, arg && *arg ? arg : ".", FALSE);
-  pr_response_add(R_211, "End of Status");
-  return (res == -1 ? ERROR(cmd) : HANDLED(cmd));
+  pr_response_add(R_211, _("End of status"));
+  return (res == -1 ? PR_ERROR(cmd) : PR_HANDLED(cmd));
 }
 
 MODRET ls_list(cmd_rec *cmd) {
@@ -1999,6 +2051,7 @@ MODRET ls_list(cmd_rec *cmd) {
 
 MODRET ls_nlst(cmd_rec *cmd) {
   char *target, buf[PR_TUNABLE_PATH_MAX + 1] = {'\0'};
+  size_t targetlen = 0;
   config_rec *c = NULL;
   int count = 0, res = 0, hidden = 0;
   int glob_flags = GLOB_PERIOD;
@@ -2011,7 +2064,8 @@ MODRET ls_nlst(cmd_rec *cmd) {
   if (tmp != NULL)
     list_show_symlinks = *tmp;
 
-  target = cmd->argc == 1 ? "." : cmd->arg;
+  target = cmd->argc == 1 ? "." :
+    pr_fs_decode_path(cmd->tmp_pool, cmd->arg);
 
   c = find_config(CURRENT_CONF, CONF_PARAM, "ListOptions", FALSE);
   if (c != NULL) {
@@ -2034,8 +2088,8 @@ MODRET ls_nlst(cmd_rec *cmd) {
   }
 
   /* Clear the listing option flags. */
-  opt_a = opt_C = opt_d = opt_F = opt_n = opt_r = opt_R = opt_S = opt_t =
-    opt_STAT = opt_L = 0;
+  opt_A = opt_a = opt_C = opt_d = opt_F = opt_n = opt_r = opt_R = opt_S =
+    opt_t = opt_STAT = opt_L = 0;
 
   if (!list_strict_opts) {
     parse_list_opts(&target, &glob_flags, FALSE);
@@ -2054,11 +2108,11 @@ MODRET ls_nlst(cmd_rec *cmd) {
       while (*target != '\0' && !isspace((int) *target))
         target++;
 
-      while (isspace((int) *target))
+      while (*target && isspace((int) *target))
         target++;
     }
 
-    while (isspace((int) *target))
+    while (*target && isspace((int) *target))
       target++;
   }
 
@@ -2114,11 +2168,21 @@ MODRET ls_nlst(cmd_rec *cmd) {
   }
 
   /* Remove any trailing separators. */
-  while (target[strlen(target)-1] == '/')
-    target[strlen(target)-1] = '\0';
+  targetlen = strlen(target);
+  while (targetlen >= 1 &&
+         target[targetlen-1] == '/') {
+
+    if (strcmp(target, "/") == 0) {
+      break;
+    }
+
+    target[targetlen-1] = '\0';
+    targetlen = strlen(target);
+  }
 
   /* If the target is a glob, get the listing of files/dirs to send. */
-  if (use_globbing && strpbrk(target, "{[*?") != NULL) {
+  if (use_globbing &&
+      strpbrk(target, "{[*?") != NULL) {
     glob_t g;
     char **path,*p;
 
@@ -2126,10 +2190,14 @@ MODRET ls_nlst(cmd_rec *cmd) {
     memset(&g, '\0', sizeof(glob_t));
 
     if (pr_fs_glob(target, GLOB_PERIOD, NULL, &g) != 0) {
-      pr_response_add_err(R_450, "No files found");
-      return ERROR(cmd);
+      pr_response_add_err(R_450, _("No files found"));
+      return PR_ERROR(cmd);
     }
 
+    if (pr_data_open(NULL, "file list", PR_NETIO_IO_WR, 0) < 0)
+      return PR_ERROR(cmd);
+    session.sf_flags |= SF_ASCII_OVERRIDE;
+
     /* Iterate through each matching entry */
     path = g.gl_pathv;
     while (path && *path && res >= 0) {
@@ -2143,10 +2211,10 @@ MODRET ls_nlst(cmd_rec *cmd) {
 
       if (pr_fsio_stat(p, &st) == 0) {
         /* If it's a directory, hand off to nlstdir */
-        if (S_ISDIR(st.st_mode))
+        if (S_ISDIR(st.st_mode)) {
           res = nlstdir(cmd, p);
 
-        else if (S_ISREG(st.st_mode) &&
+        } else if (S_ISREG(st.st_mode) &&
             ls_perms(cmd->tmp_pool, cmd, p, &hidden)) {
           /* Don't display hidden files */
           if (hidden)
@@ -2170,9 +2238,9 @@ MODRET ls_nlst(cmd_rec *cmd) {
     struct stat st;
 
     if (!ls_perms_full(cmd->tmp_pool, cmd, target, &hidden)) {
-      pr_response_add_err(R_450, "%s: %s", *cmd->arg ? cmd->arg : session.vwd,
-        strerror(errno));
-      return ERROR(cmd);
+      pr_response_add_err(R_450, "%s: %s", *cmd->arg ? cmd->arg :
+        pr_fs_encode_path(cmd->tmp_pool, session.vwd), strerror(errno));
+      return PR_ERROR(cmd);
     }
 
     /* Don't display hidden files */
@@ -2183,12 +2251,16 @@ MODRET ls_nlst(cmd_rec *cmd) {
         unsigned char *ignore_hidden = get_param_ptr(c->subset,
           "IgnoreHidden", FALSE);
 
-        if (ignore_hidden && *ignore_hidden == TRUE)
-          pr_response_add_err(R_450, "%s: %s", target, strerror(ENOENT));
-        else
-          pr_response_add_err(R_450, "%s: %s", target, strerror(EACCES));
+        if (ignore_hidden && *ignore_hidden == TRUE) {
+          pr_response_add_err(R_450, "%s: %s",
+            pr_fs_encode_path(cmd->tmp_pool, target), strerror(ENOENT));
 
-        return ERROR(cmd);
+        } else {
+          pr_response_add_err(R_450, "%s: %s",
+            pr_fs_encode_path(cmd->tmp_pool, target), strerror(EACCES));
+        }
+
+        return PR_ERROR(cmd);
       }
     }
 
@@ -2198,24 +2270,26 @@ MODRET ls_nlst(cmd_rec *cmd) {
     pr_fs_clear_cache();
     if (pr_fsio_stat(target, &st) < 0) {
       pr_response_add_err(R_450, "%s: %s", cmd->arg, strerror(errno));
-      return ERROR(cmd);
+      return PR_ERROR(cmd);
     }
 
-    if (S_ISREG(st.st_mode))
+    if (S_ISREG(st.st_mode)) {
+      if (pr_data_open(NULL, "file list", PR_NETIO_IO_WR, 0) < 0)
+        return PR_ERROR(cmd);
+      session.sf_flags |= SF_ASCII_OVERRIDE;
+
       res = nlstfile(cmd, target);
 
-    else if (S_ISDIR(st.st_mode)) {
-      if (pr_fsio_access(target, R_OK, session.uid, session.gid,
-          session.gids) != 0) {
-        pr_response_add_err(R_450, "%s: %s", cmd->arg, strerror(errno));
-        return ERROR(cmd);
-      }
+    } else if (S_ISDIR(st.st_mode)) {
+      if (pr_data_open(NULL, "file list", PR_NETIO_IO_WR, 0) < 0)
+        return PR_ERROR(cmd);
+      session.sf_flags |= SF_ASCII_OVERRIDE;
 
       res = nlstdir(cmd, target);
 
     } else {
-      pr_response_add_err(R_450, "%s: Not a regular file", cmd->arg);
-      return ERROR(cmd);
+      pr_response_add_err(R_450, _("%s: Not a regular file"), cmd->arg);
+      return PR_ERROR(cmd);
     }
 
     if (res > 0)
@@ -2233,7 +2307,7 @@ MODRET ls_nlst(cmd_rec *cmd) {
     ls_done(cmd);
   }
 
-  return (res < 0 ? ERROR(cmd) : HANDLED(cmd));
+  return (res < 0 ? PR_ERROR(cmd) : PR_HANDLED(cmd));
 }
 
 /* Check for the UseGlobbing setting, if any, after the PASS command has
@@ -2242,13 +2316,14 @@ MODRET ls_nlst(cmd_rec *cmd) {
 MODRET ls_post_pass(cmd_rec *cmd) {
   unsigned char *globbing = NULL;
 
-  if ((globbing = get_param_ptr(TOPLEVEL_CONF, "UseGlobbing",
-      FALSE)) != NULL && *globbing == FALSE) {
+  globbing = get_param_ptr(TOPLEVEL_CONF, "UseGlobbing", FALSE);
+  if (globbing != NULL &&
+      *globbing == FALSE) {
     pr_log_debug(DEBUG3, "UseGlobbing: disabling globbing functionality");
     use_globbing = FALSE;
   }
 
-  return DECLINED(cmd);
+  return PR_DECLINED(cmd);
 }
 
 /* Configuration handlers
@@ -2283,7 +2358,7 @@ MODRET set_dirfakeusergroup(cmd_rec *cmd) {
 
   c->flags |= CF_MERGEDOWN;
 
-  return HANDLED(cmd);
+  return PR_HANDLED(cmd);
 }
 
 MODRET set_dirfakemode(cmd_rec *cmd) {
@@ -2305,7 +2380,7 @@ MODRET set_dirfakemode(cmd_rec *cmd) {
   *((mode_t *) c->argv[0]) = fake_mode;
   c->flags |= CF_MERGEDOWN;
 
-  return HANDLED(cmd);
+  return PR_HANDLED(cmd);
 }
 
 MODRET set_listoptions(cmd_rec *cmd) {
@@ -2384,7 +2459,7 @@ MODRET set_listoptions(cmd_rec *cmd) {
     }
   }
 
-  return HANDLED(cmd);
+  return PR_HANDLED(cmd);
 }
 
 MODRET set_showsymlinks(cmd_rec *cmd) {
@@ -2402,7 +2477,7 @@ MODRET set_showsymlinks(cmd_rec *cmd) {
   *((unsigned char *) c->argv[0]) = bool;
   c->flags |= CF_MERGEDOWN;
 
-  return HANDLED(cmd);
+  return PR_HANDLED(cmd);
 }
 
 MODRET set_useglobbing(cmd_rec *cmd) {
@@ -2420,7 +2495,7 @@ MODRET set_useglobbing(cmd_rec *cmd) {
   *((unsigned char *) c->argv[0]) = bool;
   c->flags |= CF_MERGEDOWN;
 
-  return HANDLED(cmd);
+  return PR_HANDLED(cmd);
 }
 
 /* Initialization routines
