@@ -2,7 +2,7 @@
  * ProFTPD - FTP server daemon
  * Copyright (c) 1997, 1998 Public Flood Software
  * Copyright (c) 1999, 2000 MacGyver aka Habeeb J. Dihu <macgyver@tos.net>
- * Copyright (c) 2001-2007 The ProFTPD Project team
+ * Copyright (c) 2001-2009 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -25,7 +25,7 @@
  */
 
 /* Read configuration file(s), and manage server/configuration structures.
- * $Id: dirtree.c,v 1.185 2007/07/24 22:09:39 castaglia Exp $
+ * $Id: dirtree.c,v 1.205.2.3 2009/04/28 22:33:57 castaglia Exp $
  */
 
 #include "conf.h"
@@ -45,11 +45,13 @@ int SocketBindTight = FALSE;
 char ServerType = SERVER_STANDALONE;
 int ServerMaxInstances = 0;
 int ServerUseReverseDNS = TRUE;
-int TimeoutIdle = PR_TUNABLE_TIMEOUTIDLE;
-int TimeoutNoXfer = PR_TUNABLE_TIMEOUTNOXFER;
-int TimeoutStalled = PR_TUNABLE_TIMEOUTSTALLED;
 char MultilineRFC2228 = 0;
 
+/* Default TCP send/receive buffer sizes. */
+static int tcp_rcvbufsz = 0;
+static int tcp_sndbufsz = 0;
+static int xfer_bufsz = 0;
+
 /* From src/pool.c */
 extern pool *global_config_pool;
 
@@ -61,9 +63,16 @@ static void merge_down(xaset_t *, int);
 static config_rec *_last_param_ptr = NULL;
 static unsigned char _kludge_disable_umask = 0;
 
+/* We have two different lists for Defines.  The 'perm' pool/list are
+ * for "permanent" defines, i.e. those set on the command-line via the
+ * -D/--define options.
+ */
 static pool *defines_pool = NULL;
 static array_header *defines_list = NULL;
 
+static pool *defines_perm_pool = NULL;
+static array_header *defines_perm_list = NULL;
+
 static pool *config_tab_pool = NULL;
 static pr_table_t *config_tab = NULL;
 static unsigned int config_id = 0;
@@ -84,7 +93,7 @@ static int allow_dyn_config(void) {
     if (c->argc == 3) {
       if (!strcmp(c->argv[2], "user")) {
 
-        if (pr_expr_eval_user_or((char **) &c->argv[3])) {
+        if (pr_expr_eval_user_or((char **) &c->argv[3]) == TRUE) {
           if (*((unsigned int *) c->argv[1]) > ctxt_precedence) {
 
             /* Set the context precedence. */
@@ -99,7 +108,7 @@ static int allow_dyn_config(void) {
 
       } else if (!strcmp(c->argv[2], "group")) {
 
-        if (pr_expr_eval_group_and((char **) &c->argv[3])) {
+        if (pr_expr_eval_group_and((char **) &c->argv[3]) == TRUE) {
           if (*((unsigned int *) c->argv[1]) > ctxt_precedence) {
 
             /* Set the context precedence. */
@@ -114,7 +123,7 @@ static int allow_dyn_config(void) {
 
       } else if (!strcmp(c->argv[2], "class")) {
 
-        if (pr_expr_eval_class_or((char **) &c->argv[3])) {
+        if (pr_expr_eval_class_or((char **) &c->argv[3]) == TRUE) {
           if (*((unsigned int *) c->argv[1]) > ctxt_precedence) {
 
             /* Set the context precedence. */
@@ -368,6 +377,10 @@ unsigned char dir_hide_file(const char *path) {
   unsigned char have_user_regex, have_group_regex, have_class_regex,
     have_all_regex, inverted = FALSE;
 
+  if (path == NULL) {
+    return FALSE;
+  }
+
   pr_pool_tag(tmp_pool, "dir_hide_file() tmp pool");
 
   have_user_regex = have_group_regex = have_class_regex = have_all_regex =
@@ -376,8 +389,9 @@ unsigned char dir_hide_file(const char *path) {
   /* Separate the given path into directory and file components. */
   dir_name = pstrdup(tmp_pool, path);
 
-  if ((file_name = strrchr(dir_name, '/')) != NULL) {
-    file_name = '\0';
+  file_name = strrchr(dir_name, '/');
+  if (file_name != NULL) {
+    *file_name = '\0';
     file_name++;
 
   } else
@@ -392,7 +406,7 @@ unsigned char dir_hide_file(const char *path) {
 
       /* check for a specified "user" classifier first... */
       if (strcmp(c->argv[3], "user") == 0) {
-        if (pr_expr_eval_user_or((char **) &c->argv[4])) {
+        if (pr_expr_eval_user_or((char **) &c->argv[4]) == TRUE) {
 
           if (*((unsigned int *) c->argv[2]) > ctxt_precedence) {
             ctxt_precedence = *((unsigned int *) c->argv[2]);
@@ -407,7 +421,7 @@ unsigned char dir_hide_file(const char *path) {
 
       /* ...then for a "group" classifier... */
       } else if (strcmp(c->argv[3], "group") == 0) {
-        if (pr_expr_eval_group_and((char **) &c->argv[4])) {
+        if (pr_expr_eval_group_and((char **) &c->argv[4]) == TRUE) {
           if (*((unsigned int *) c->argv[2]) > ctxt_precedence) {
             ctxt_precedence = *((unsigned int *) c->argv[2]);
 
@@ -425,7 +439,7 @@ unsigned char dir_hide_file(const char *path) {
        * need to be updated to process class-expressions.
        */
       } else if (strcmp(c->argv[3], "class") == 0) {
-        if (pr_expr_eval_class_or((char **) &c->argv[4])) {
+        if (pr_expr_eval_class_or((char **) &c->argv[4]) == TRUE) {
           if (*((unsigned int *) c->argv[2]) > ctxt_precedence) {
             ctxt_precedence = *((unsigned int *) c->argv[2]);
 
@@ -459,7 +473,6 @@ unsigned char dir_hide_file(const char *path) {
     c = find_config_next(c, c->next, CONF_PARAM, "HideFiles", FALSE);
   }
 
-
   if (have_user_regex || have_group_regex ||
       have_class_regex || have_all_regex) {
 
@@ -504,17 +517,48 @@ static void define_restart_ev(const void *event_data, void *user_data) {
   pr_event_unregister(NULL, "core.restart", define_restart_ev);
 }
 
-int pr_define_add(const char *definition) {
+/* The 'survive_restarts' boolean indicates whether this Define is to be
+ * permanent for the lifetime of the daemon (i.e. survives across restarts)
+ * or whether it should be cleared when restarted.
+ *
+ * Right now, defines from the command-line will surive restarts, but
+ * defines from the config (via the Define directive) will not.
+ */
+int pr_define_add(const char *definition, int survive_restarts) {
 
-  if (!defines_list) {
-    defines_pool = make_sub_pool(permanent_pool);
-    pr_pool_tag(defines_pool, "Defines Pool");
-    defines_list = make_array(defines_pool, 0, sizeof(char *));
-
-    pr_event_register(NULL, "core.restart", define_restart_ev, NULL);
+  if (definition == NULL ||
+      (survive_restarts != FALSE && survive_restarts != TRUE)) {
+    errno = EINVAL;
+    return -1;
   }
 
-  *((char **) push_array(defines_list)) = pstrdup(defines_pool, definition);
+  if (survive_restarts == FALSE) {
+    if (defines_pool == NULL) {
+      defines_pool = make_sub_pool(permanent_pool);
+      pr_pool_tag(defines_pool, "Defines Pool");
+      pr_event_register(NULL, "core.restart", define_restart_ev, NULL);
+    }
+
+    if (!defines_list) {
+      defines_list = make_array(defines_pool, 0, sizeof(char *));
+
+    }
+
+    *((char **) push_array(defines_list)) = pstrdup(defines_pool, definition);
+    return 0;
+  }
+
+  if (defines_perm_pool == NULL) {
+    defines_perm_pool = make_sub_pool(permanent_pool);
+    pr_pool_tag(defines_perm_pool, "Permanent Defines Pool");
+  }
+
+  if (!defines_perm_list) {
+    defines_perm_list = make_array(defines_perm_pool, 0, sizeof(char *));
+  }
+
+  *((char **) push_array(defines_perm_list)) =
+    pstrdup(defines_perm_pool, definition);
   return 0;
 }
 
@@ -543,83 +587,41 @@ void kludge_enable_umask(void) {
   _kludge_disable_umask = FALSE;
 }
 
-char *pr_str_get_word(char **cp, int flags) {
-  char *ret, *dst;
-  char quote_mode = 0;
-
-  if (!cp || !*cp || !**cp)
-    return NULL;
-
-  if (!(flags & PR_STR_FL_PRESERVE_WHITESPACE)) {
-    while (**cp && isspace((int) **cp))
-      (*cp)++;
-  }
-
-  if (!**cp)
-    return NULL;
-
-  ret = dst = *cp;
-
-  if (!(flags & PR_STR_FL_PRESERVE_COMMENTS)) {
-    /* Stop processing at start of an inline comment. */
-    if (**cp == '#')
-      return NULL;
-  }
-
-  if (**cp == '\"') {
-    quote_mode++;
-    (*cp)++;
-  }
-
-  while (**cp && (quote_mode ? (**cp != '\"') : !isspace((int) **cp))) {
-    if (**cp == '\\' && quote_mode) {
-
-      /* Escaped char */
-      if (*((*cp)+1))
-        *dst = *(++(*cp));
-    }
-
-    *dst++ = **cp;
-    ++(*cp);
-  }
-
-  if (**cp)
-    (*cp)++;
-  *dst = '\0';
-
-  return ret;
-}
-
 cmd_rec *pr_cmd_alloc(pool *p, int argc, ...) {
   pool *newpool = NULL;
-  cmd_rec *c = NULL;
+  cmd_rec *cmd = NULL;
   va_list args;
 
   newpool = make_sub_pool(p);
   pr_pool_tag(newpool, "pr_cmd_alloc() subpool");
 
-  c = pcalloc(newpool, sizeof(cmd_rec));
-  c->argc = argc;
-  c->stash_index = -1;
-  c->pool = newpool;
-  c->tmp_pool = make_sub_pool(c->pool);
-  pr_pool_tag(c->tmp_pool, "pr_cmd_alloc() tmp pool");
+  cmd = pcalloc(newpool, sizeof(cmd_rec));
+  cmd->argc = argc;
+  cmd->stash_index = -1;
+  cmd->pool = newpool;
+  cmd->tmp_pool = make_sub_pool(cmd->pool);
+  pr_pool_tag(cmd->tmp_pool, "pr_cmd_alloc() tmp pool");
 
   if (argc) {
     register unsigned int i = 0;
 
-    c->argv = pcalloc(newpool, sizeof(void *) * (argc + 1));
+    cmd->argv = pcalloc(newpool, sizeof(void *) * (argc + 1));
     va_start(args, argc);
 
     for (i = 0; i < argc; i++)
-      c->argv[i] = (void *) va_arg(args, char *);
+      cmd->argv[i] = (void *) va_arg(args, char *);
 
     va_end(args);
 
-    c->argv[argc] = NULL;
+    cmd->argv[argc] = NULL;
   }
 
-  return c;
+  /* This table will not contain that many entries, so a low number
+   * of chains should suffice.
+   */
+  cmd->notes = pr_table_nalloc(cmd->pool, 0, 8);
+
+  return cmd;
 }
 
 /* Adds a config_rec to the specified set */
@@ -636,8 +638,11 @@ config_rec *add_config_set(xaset_t **set, const char *name) {
     *set = xaset_create(set_pool, NULL);
     (*set)->pool = set_pool;
 
-    /* Now, make a subpool for the config_rec to be allocated. */
-    conf_pool = make_sub_pool(set_pool);
+    /* Now, make a subpool for the config_rec to be allocated.  The default
+     * pool size (PR_TUNABLE_NEW_POOL_SIZE, 512 by default) is a bit large
+     * for config_rec pools; use a smaller size.
+     */
+    conf_pool = pr_pool_create_sz(set_pool, 128);
 
   } else {
 
@@ -645,11 +650,15 @@ config_rec *add_config_set(xaset_t **set, const char *name) {
     if ((*set)->xas_list)
       parent = ((config_rec*)((*set)->xas_list))->parent;
 
-    /* Allocate a subpool for the config_rec from the parent's pool. */
-    conf_pool = make_sub_pool((*set)->pool);
+    /* Now, make a subpool for the config_rec to be allocated.  The default
+     * pool size (PR_TUNABLE_NEW_POOL_SIZE, 512 by default) is a bit large
+     * for config_rec pools; use a smaller size.  Allocate the subpool
+     * from the parent's pool.
+     */
+    conf_pool = pr_pool_create_sz((*set)->pool, 128);
   }
 
-  pr_pool_tag(conf_pool, "config_rec subpool");
+  pr_pool_tag(conf_pool, "config_rec pool");
 
   c = (config_rec *) pcalloc(conf_pool, sizeof(config_rec));
 
@@ -707,211 +716,6 @@ config_rec *add_config(server_rec *s, const char *name) {
   return c;
 }
 
-array_header *pr_expr_create(pool *p, int *argc, char **argv) {
-  array_header *acl = NULL;
-  int cnt = *argc;
-  char *s, *ent;
-
-  if (cnt) {
-    acl = make_array(p, cnt, sizeof(char *));
-
-    while (cnt-- && *(++argv)) {
-      s = pstrdup(p, *argv);
-
-      while ((ent = get_token(&s, ",")) != NULL)
-        if (*ent)
-          *((char **) push_array(acl)) = ent;
-    }
-
-    *argc = acl->nelts;
-
-  } else
-    *argc = 0;
-
-  return acl;
-}
-
-/* Boolean "class-expression" AND matching, returns TRUE if the expression
- * evaluates to TRUE.
- */
-unsigned char pr_expr_eval_class_and(char **expr) {
-  unsigned char found;
-  char *class;
-
-  for (; *expr; expr++) {
-    class = *expr;
-    found = FALSE;
-
-    if (*class == '!') {
-      found = !found;
-      class++;
-    }
-
-    if (!session.class && !found)
-      return FALSE;
-
-    if (session.class && strcmp(session.class->cls_name, class) == 0)
-      found = !found;
-
-    if (!found)
-      return FALSE;
-  }
-
-  return TRUE;
-}
-
-/* Boolean "class-expression" OR matching, returns TRUE if the expression
- * evaluates to TRUE.
- */
-unsigned char pr_expr_eval_class_or(char **expr) {
-  unsigned char found;
-  char *class;
-
-  for (; *expr; expr++) {
-    class = *expr;
-    found = FALSE;
-
-    if (*class == '!') {
-      found = !found;
-      class++;
-    }
-
-    if (!session.class)
-      return found;
-
-    if (strcmp(session.class->cls_name, class) == 0)
-      found = !found;
-
-    if (found)
-      return TRUE;
-  }
-
-  return FALSE;
-}
-
-/* Boolean "group-expression" AND matching, returns TRUE if the expression
- * evaluates to TRUE.
- */
-unsigned char pr_expr_eval_group_and(char **expr) {
-  unsigned char found;
-  char *grp;
-
-  for (; *expr; expr++) {
-    grp = *expr;
-    found = FALSE;
-
-    if (*grp == '!') {
-      found = !found;
-      grp++;
-    }
-
-    if (session.group && strcmp(session.group, grp) == 0)
-      found = !found;
-
-    else if (session.groups) {
-      register int i = 0;
-
-      for (i = session.groups->nelts-1; i >= 0; i--)
-        if (strcmp(*(((char **) session.groups->elts) + i), grp) == 0) {
-          found = !found;
-          break;
-        }
-    }
-
-    if (!found)
-      return FALSE;
-  }
-
-  return TRUE;
-}
-
-/* Boolean "group-expression" OR matching, returns TRUE if the expression
- * evaluates to TRUE.
- */
-unsigned char pr_expr_eval_group_or(char **expr) {
-  unsigned char found;
-  char *grp;
-
-  for (; *expr; expr++) {
-    grp = *expr;
-    found = FALSE;
-
-    if (*grp == '!') {
-      found = !found;
-      grp++;
-    }
-
-    if (session.group && strcmp(session.group, grp) == 0)
-      found = !found;
-
-    else if (session.groups) {
-      register int i = 0;
-
-      for (i = session.groups->nelts-1; i >= 0; i--)
-        if (strcmp(*(((char **) session.groups->elts) + i), grp) == 0) {
-          found = !found;
-          break;
-        }
-    }
-
-    if (found)
-      return TRUE;
-  }
-
-  return FALSE;
-}
-
-/* Boolean "user-expression" AND matching, returns TRUE if the expression
- * evaluates to TRUE.
- */
-unsigned char pr_expr_eval_user_and(char **expr) {
-  unsigned char found;
-  char *user;
-
-  for (; *expr; expr++) {
-    user = *expr;
-    found = FALSE;
-
-    if (*user == '!') {
-      found = !found;
-      user++;
-    }
-
-    if (strcmp(session.user, user) == 0)
-      found = !found;
-
-    if (!found) 
-      return FALSE;
-  }
-
-  return TRUE;
-}
-
-/* Boolean "user-expression" OR matching, returns TRUE if the expression
- * evaluates to TRUE.
- */
-unsigned char pr_expr_eval_user_or(char **expr) {
-  unsigned char found;
-  char *user;
-
-  for (; *expr; expr++) {
-    user = *expr;
-    found = FALSE;
-
-    if (*user == '!') {
-      found = !found;
-      user++;
-    }
-
-    if (strcmp(session.user, user) == 0)
-      found = !found;
-
-    if (found)
-      return TRUE;
-  }
-
-  return FALSE;
-}
 /* Per-directory configuration */
 
 static size_t _strmatch(register char *s1, register char *s2) {
@@ -927,11 +731,15 @@ static config_rec *recur_match_path(pool *p, xaset_t *s, char *path) {
   char *tmp_path = NULL;
   config_rec *c = NULL, *res = NULL;
 
-  if (!s)
+  if (!s) {
+    errno = EINVAL;
     return NULL;
+  }
 
-  for (c = (config_rec *) s->xas_list; c; c = c->next)
+  for (c = (config_rec *) s->xas_list; c; c = c->next) {
     if (c->config_type == CONF_DIR) {
+      size_t path_len;
+
       tmp_path = c->name;
 
       if (c->argv[1]) {
@@ -942,33 +750,74 @@ static config_rec *recur_match_path(pool *p, xaset_t *s, char *path) {
       }
 
       /* Exact path match */
-      if (strcmp(tmp_path, path) == 0)
+      if (strcmp(tmp_path, path) == 0) {
+        pr_trace_msg("directory", 8,
+          "<Directory %s> is an exact path match for '%s'", c->name, path);
         return c;
+      }
 
-      if (!strstr(tmp_path, "/*")) {
-        size_t tmplen = strlen(tmp_path);
+      /* Bug#3146 occurred because using strstr(3) works well for paths
+       * which DO NOT contain the glob sequence, i.e. we used to do:
+       *
+       *  if (strstr(tmp_path, slash_star) == NULL) {
+       *
+       * But what if they do, just not at the end of the path?
+       *
+       * The fix is to explicitly check the last two characters of the path
+       * for '/' and '*', rather than using strstr(3).  (Again, I wish there
+       * was a strrstr(3) libc function.)
+       */
+      path_len = strlen(tmp_path);
+      if (path_len > 2 &&
+          tmp_path[path_len-2] != '/' &&
+          tmp_path[path_len-1] != '*') {
 
         /* Trim a trailing path separator, if present. */
-        if (*tmp_path && *(tmp_path + tmplen - 1) == '/' && tmplen > 1) {
-          *(tmp_path + tmplen - 1) = '\0';
+        if (*tmp_path && *(tmp_path + path_len - 1) == '/' && path_len > 1) {
+          *(tmp_path + path_len - 1) = '\0';
 
-          if (strcmp(tmp_path, path) == 0)
+          if (strcmp(tmp_path, path) == 0) {
+            pr_trace_msg("directory", 8,
+              "<Directory %s> is an exact path match for '%s'", c->name, path);
             return c;
+          }
         }
 
         tmp_path = pdircat(p, tmp_path, "*", NULL);
+
+      } else if (path_len == 1) {
+        /* We still need to append the "*" if the path is just '/'. */
+        tmp_path = pstrcat(p, tmp_path, "*", NULL);
       }
 
+      pr_trace_msg("directory", 9,
+        "checking if <Directory %s> is a glob match for %s", tmp_path, path);
+
       if (pr_fnmatch(tmp_path, path, 0) == 0) {
+        pr_trace_msg("directory", 8,
+          "<Directory %s> is a glob match for '%s'", tmp_path, path);
+
         if (c->subset) {
-          if ((res = recur_match_path(p, c->subset, path)))
+          /* If there's a subset config, check to see if there's a closer
+           * match there.
+           */
+          res = recur_match_path(p, c->subset, path);
+          if (res) {
+            pr_trace_msg("directory", 8,
+              "found closer matching <Directory %s> for '%s' in <Directory %s> "
+              "sub-config", res->name, path, tmp_path);
             return res;
+          }
         }
 
+        pr_trace_msg("directory", 8, "found <Directory %s> for '%s'",
+          c->name, path);
         return c;
       }
     }
+  }
 
+  errno = ENOENT;
   return NULL;
 }
 
@@ -977,8 +826,9 @@ config_rec *dir_match_path(pool *p, char *path) {
   char *tmp = NULL;
   size_t tmplen;
 
-  if (!p || !path || !*path)
+  if (!p || !path || !*path) {
     return NULL;
+  }
 
   tmp = pstrdup(p, path);
   tmplen = strlen(tmp);
@@ -1001,8 +851,18 @@ config_rec *dir_match_path(pool *p, char *path) {
     }
   }
 
-  if (!res)
+  if (!res) {
     res = recur_match_path(p, main_server->conf, tmp);
+  }
+
+  if (res) {
+    pr_trace_msg("directory", 3, "matched <Directory %s> for path '%s'",
+      res->name, tmp);
+
+  } else {
+    pr_trace_msg("directory", 3, "no matching <Directory> found for '%s': %s",
+      tmp, strerror(errno));
+  }
 
   return res;
 }
@@ -1109,13 +969,13 @@ static int _check_user_access(xaset_t *set, char *name) {
     if (*((unsigned char *) c->argv[0]) == PR_EXPR_EVAL_OR) {
       res = pr_expr_eval_user_or((char **) &c->argv[1]);
 
-      if (res)
+      if (res == TRUE)
         break;
 
     } else if (*((unsigned char *) c->argv[0]) == PR_EXPR_EVAL_AND) {
       res = pr_expr_eval_user_and((char **) &c->argv[1]);
 
-      if (res)
+      if (res == TRUE)
         break;
     }
 
@@ -1155,13 +1015,13 @@ static int _check_group_access(xaset_t *set, char *name) {
     if (*((unsigned char *) c->argv[0]) == PR_EXPR_EVAL_OR) {
       res = pr_expr_eval_group_or((char **) &c->argv[1]);
 
-      if (res)
+      if (res == TRUE)
         break;
 
     } else if (*((unsigned char *) c->argv[0]) == PR_EXPR_EVAL_AND) {
       res = pr_expr_eval_group_and((char **) &c->argv[1]);
 
-      if (res)
+      if (res == TRUE)
         break;
     }
 
@@ -1192,13 +1052,13 @@ static int _check_class_access(xaset_t *set, char *name) {
     if (*((unsigned char *) c->argv[0]) == PR_EXPR_EVAL_OR) {
       res = pr_expr_eval_class_or((char **) &c->argv[1]);
 
-      if (res)
+      if (res == TRUE)
         break;
 
     } else if (*((unsigned char *) c->argv[0]) == PR_EXPR_EVAL_AND) {
       res = pr_expr_eval_class_and((char **) &c->argv[1]);
 
-      if (res)
+      if (res == TRUE)
         break;
     }
 
@@ -1438,11 +1298,13 @@ int login_check_limits(xaset_t *set, int recurse, int and, int *found) {
     return TRUE;			/* default is to allow */
 
   /* First check top level */
-  for (c = (config_rec *) set->xas_list; c; c = c->next)
+  for (c = (config_rec *) set->xas_list; c; c = c->next) {
     if (c->config_type == CONF_LIMIT) {
-      for (argc = c->argc, argv = (char **) c->argv; argc; argc--, argv++)
-        if (strcasecmp("LOGIN", *argv) == 0)
+      for (argc = c->argc, argv = (char **) c->argv; argc; argc--, argv++) {
+        if (strcasecmp("LOGIN", *argv) == 0) {
           break;
+        }
+      }
 
       if (argc) {
         if (and) {
@@ -1475,10 +1337,13 @@ int login_check_limits(xaset_t *set, int recurse, int and, int *found) {
         }
       }
     }
+  }
 
-  if ( ((res && and) || (!res && !and && *found)) && recurse ) {
-    for (c = (config_rec *) set->xas_list; c; c = c->next)
-      if (c->config_type == CONF_ANON && c->subset && c->subset->xas_list) {
+  if (((res && and) || (!res && !and && *found)) && recurse) {
+    for (c = (config_rec *) set->xas_list; c; c = c->next) {
+      if (c->config_type == CONF_ANON &&
+          c->subset &&
+          c->subset->xas_list) {
        if (and) {
          res = (res && login_check_limits(c->subset, recurse, and, &rfound));
          (*found) += rfound;
@@ -1491,11 +1356,13 @@ int login_check_limits(xaset_t *set, int recurse, int and, int *found) {
          rres = login_check_limits(c->subset, recurse, and, &rfound);
          if (rfound)
            res = (res || rres);
+
          (*found) += rfound;
          if (res)
            break;
        }
      }
+    }
   }
 
   if (!*found && !and)
@@ -1516,13 +1383,15 @@ static int _check_limits(xaset_t *set, char *cmd, int hidden) {
     return res;
 
   for (lc = (config_rec *) set->xas_list; lc && (res == 1); lc = lc->next) {
+    pr_signals_handle();
 
     if (lc->config_type == CONF_LIMIT) {
       register unsigned int i = 0;
 
       for (i = 0; i < lc->argc; i++) {
-        if (strcasecmp(cmd, (char *) lc->argv[i]) == 0)
+        if (strcasecmp(cmd, (char *) lc->argv[i]) == 0) {
           break;
+        }
       }
 	
       if (i == lc->argc)
@@ -1574,14 +1443,16 @@ static int _check_limits(xaset_t *set, char *cmd, int hidden) {
 int dir_check_limits(config_rec *c, char *cmd, int hidden) {
   int res = 1;
 
-  for (; c && (res == 1); c = c->parent)
+  for (; c && (res == 1); c = c->parent) {
     res = _check_limits(c->subset, cmd, hidden);
+  }
 
-  if (!c && (res == 1))
+  if (!c && (res == 1)) {
     /* vhost or main server has been reached without an explicit permit or deny,
      * so try the current server.
      */
     res = _check_limits(main_server->conf, cmd, hidden);
+  }
 
   return res;
 }
@@ -1731,10 +1602,32 @@ void build_dyn_config(pool *p, char *_path, struct stat *stp,
       break;
 
     cp = strrchr(path, '/');
+    if (cp) {
+      if (strcmp(path, "/") != 0) {
 
-    if (cp &&
-        strcmp(path, "/") != 0) {
-      *cp = '\0';
+        /* We need to handle the case where path might be "/path".  We
+         * can't just set *cp to '\0', as that would result in the empty
+         * string.  Thus check if cp is the same value as path, i.e.
+         * that cp points to the start of the string.  If so, by definition
+         * we know that we are dealing with the "/path" case.
+         */
+        if (cp != path) {
+          *cp = '\0';
+
+        } else {
+          /* Set the recurse flag to 'false', so that we go one more pass
+           * through the loop, but stop after that.
+           */
+          *(cp+1) = '\0';
+          recurse = FALSE;        
+        }
+
+      } else {
+        /* Set the recurse flag to 'false', so that we go one more pass
+         * through the loop, but stop after that.
+         */
+        recurse = FALSE;        
+      }
 
     } else {
       path = NULL;
@@ -1763,7 +1656,7 @@ void build_dyn_config(pool *p, char *_path, struct stat *stp,
  */
 
 int dir_check_full(pool *pp, char *cmd, char *group, char *path, int *hidden) {
-  char *fullpath, *owner, *tmp = NULL;
+  char *fullpath, *owner;
   config_rec *c;
   struct stat st;
   pool *p;
@@ -1796,13 +1689,16 @@ int dir_check_full(pool *pp, char *cmd, char *group, char *path, int *hidden) {
   build_dyn_config(p, path, &st, TRUE);
 
   /* Check to see if this path is hidden by HideFiles. */
-  tmp = strrchr(fullpath, '/');
-  regex_hidden = tmp ? dir_hide_file(++tmp) : dir_hide_file(fullpath);
+  regex_hidden = dir_hide_file(path);
 
   /* Cache a pointer to the set of configuration data for this directory in
    * session.dir_config.
    */
   session.dir_config = c = dir_match_path(p, fullpath);
+  if (session.dir_config) {
+    pr_trace_msg("directory", 2, "matched <Directory %s> for '%s'",
+      session.dir_config->name, fullpath);
+  }
 
   if (!c && session.anon_config)
     c = session.anon_config;
@@ -1900,7 +1796,7 @@ int dir_check_full(pool *pp, char *cmd, char *group, char *path, int *hidden) {
  */
 
 int dir_check(pool *pp, char *cmd, char *group, char *path, int *hidden) {
-  char *fullpath, *owner, *tmp = NULL;
+  char *fullpath, *owner;
   config_rec *c;
   struct stat st;
   pool *p;
@@ -1938,13 +1834,16 @@ int dir_check(pool *pp, char *cmd, char *group, char *path, int *hidden) {
   build_dyn_config(p, path, &st, FALSE);
 
   /* Check to see if this path is hidden by HideFiles. */
-  tmp = strrchr(fullpath, '/');
-  regex_hidden = tmp ? dir_hide_file(++tmp) : dir_hide_file(fullpath);
+  regex_hidden = dir_hide_file(path);
 
   /* Cache a pointer to the set of configuration data for this directory in
    * session.dir_config.
    */
   session.dir_config = c = dir_match_path(p, fullpath);
+  if (session.dir_config) {
+    pr_trace_msg("directory", 2, "matched <Directory %s> for '%s'",
+      session.dir_config->name, fullpath);
+  }
 
   if (!c && session.anon_config)
     c = session.anon_config;
@@ -2491,6 +2390,8 @@ config_rec *find_config_next(config_rec *prev, config_rec *c, int type,
     do {
       config_rec *res = NULL;
 
+      pr_signals_handle();
+
       for (c = top; c; c = c->next) {
         if (c->subset &&
             c->subset->xas_list) {
@@ -2904,13 +2805,13 @@ int fixup_servers(xaset_t *list) {
 
 #ifdef PR_USE_IPV6
           if (pr_netaddr_use_ipv6()) {
-            char *ipbuf = pcalloc(s->pool, INET6_ADDRSTRLEN);
+            char *ipbuf = pcalloc(s->pool, INET6_ADDRSTRLEN + 1);
             if (pr_netaddr_get_family(elts[i]) == AF_INET) {
 
               /* Create the bind record using the IPv4-mapped IPv6 version of
                * this address.
                */
-              snprintf(ipbuf, sizeof(ipbuf), "::ffff:%s", ipstr);
+              snprintf(ipbuf, INET6_ADDRSTRLEN, "::ffff:%s", ipstr);
               ipstr = pstrdup(s->pool, ipbuf);
             }
           }
@@ -2946,13 +2847,13 @@ int fixup_servers(xaset_t *list) {
     }
 
     if (!s->tcp_rcvbuf_len)
-      s->tcp_rcvbuf_len = PR_TUNABLE_RCVBUFSZ;
+      s->tcp_rcvbuf_len = tcp_rcvbufsz;
 
     if (!s->tcp_sndbuf_len)
-      s->tcp_sndbuf_len = PR_TUNABLE_SNDBUFSZ;
+      s->tcp_sndbuf_len = tcp_sndbufsz;
 
-    if ((c = find_config(s->conf, CONF_PARAM, "MasqueradeAddress",
-        FALSE)) != NULL) {
+    c = find_config(s->conf, CONF_PARAM, "MasqueradeAddress", FALSE);
+    if (c != NULL) {
       pr_log_pri(PR_LOG_INFO, "%s:%d masquerading as %s",
         pr_netaddr_get_ipstr(s->addr), s->ServerPort,
         pr_netaddr_get_ipstr((pr_netaddr_t *) c->argv[0]));
@@ -2987,6 +2888,105 @@ int fixup_servers(xaset_t *list) {
   return 0;
 }
 
+static void set_tcp_bufsz(void) {
+  int sockfd;
+  socklen_t optlen = 0;
+  struct protoent *p = NULL;
+
+#ifdef HAVE_SETPROTOENT
+  setprotoent(FALSE);
+#endif
+
+  p = getprotobyname("tcp");
+  if (!p) {
+#ifndef PR_TUNABLE_RCVBUFSZ
+    tcp_rcvbufsz = PR_TUNABLE_DEFAULT_RCVBUFSZ;
+#else
+    tcp_rcvbufsz = PR_TUNABLE_RCVBUFSZ;
+#endif /* PR_TUNABLE_RCVBUFSZ */
+
+#ifndef PR_TUNABLE_SNDBUFSZ
+    tcp_sndbufsz = PR_TUNABLE_DEFAULT_SNDBUFSZ;
+#else
+    tcp_sndbufsz = PR_TUNABLE_SNDBUFSZ;
+#endif /* PR_TUNABLE_SNDBUFSZ */
+
+    pr_log_debug(DEBUG3, "getprotobyname error for 'tcp': %s", strerror(errno));
+    pr_log_debug(DEBUG4, "using default TCP receive/send buffer sizes");
+
+#ifndef PR_TUNABLE_XFER_BUFFER_SIZE
+  /* Choose the smaller of the two TCP buffer sizes as the overall transfer
+   * size (for use by the data transfer layer).
+   */
+   xfer_bufsz = tcp_sndbufsz < tcp_rcvbufsz ? tcp_sndbufsz : tcp_rcvbufsz;
+#else
+  xfer_bufsz = PR_TUNABLE_XFER_BUFFER_SIZE;
+#endif /* PR_TUNABLE_XFER_BUFFER_SIZE */
+
+    return;
+  }
+
+  sockfd = socket(AF_INET, SOCK_STREAM, p->p_proto);
+  if (sockfd < 0) {
+    tcp_rcvbufsz = PR_TUNABLE_DEFAULT_RCVBUFSZ;
+    tcp_sndbufsz = PR_TUNABLE_DEFAULT_SNDBUFSZ;
+
+    pr_log_debug(DEBUG3, "socket error: %s", strerror(errno));
+    pr_log_debug(DEBUG4, "using default TCP receive/send buffer sizes");
+  }
+
+#ifndef PR_TUNABLE_RCVBUFSZ
+  /* Determine the optimal size of the TCP receive buffer. */
+  optlen = sizeof(tcp_rcvbufsz);
+  if (getsockopt(sockfd, SOL_SOCKET, SO_RCVBUF, (void *) &tcp_rcvbufsz,
+      &optlen) < 0) {
+    tcp_rcvbufsz = PR_TUNABLE_DEFAULT_RCVBUFSZ;
+
+    pr_log_debug(DEBUG3, "getsockopt error for SO_RCVBUF: %s", strerror(errno));
+    pr_log_debug(DEBUG4, "using default TCP receive buffer size of %d bytes",
+      tcp_rcvbufsz);
+
+  } else {
+    pr_log_debug(DEBUG5, "using TCP receive buffer size of %d bytes",
+      tcp_rcvbufsz);
+  }
+#else
+  optlen = -1;
+  tcp_sndbufsz = PR_TUNABLE_RCVBUFSZ;
+  pr_log_debug(DEBUG5, "using preset TCP receive buffer size of %d bytes",
+    tcp_rcvbufsz);
+#endif /* PR_TUNABLE_RCVBUFSZ */
+
+#ifndef PR_TUNABLE_SNDBUFSZ
+  /* Determine the optimal size of the TCP send buffer. */
+  optlen = sizeof(tcp_sndbufsz);
+  if (getsockopt(sockfd, SOL_SOCKET, SO_SNDBUF, (void *) &tcp_sndbufsz,
+      &optlen) < 0) {
+    tcp_sndbufsz = PR_TUNABLE_DEFAULT_SNDBUFSZ;
+    
+    pr_log_debug(DEBUG3, "getsockopt error for SO_SNDBUF: %s", strerror(errno));
+    pr_log_debug(DEBUG4, "using default TCP send buffer size of %d bytes",
+      tcp_sndbufsz);
+  
+  } else {
+    pr_log_debug(DEBUG5, "using TCP send buffer size of %d bytes",
+      tcp_sndbufsz);
+  }
+#else
+  optlen = -1;
+  tcp_sndbufsz = PR_TUNABLE_SNDBUFSZ;
+  pr_log_debug(DEBUG5, "using preset TCP send buffer size of %d bytes",
+    tcp_sndbufsz);
+#endif /* PR_TUNABLE_SNDBUFSZ */
+
+  /* Choose the smaller of the two TCP buffer sizes as the overall transfer
+   * size (for use by the data transfer layer).
+   */
+   xfer_bufsz = tcp_sndbufsz < tcp_rcvbufsz ? tcp_sndbufsz : tcp_rcvbufsz;
+
+  (void) close(sockfd);
+}
+
 void init_config(void) {
   pool *conf_pool = make_sub_pool(permanent_pool);
   pr_pool_tag(conf_pool, "Config Pool");
@@ -3060,6 +3060,7 @@ void init_config(void) {
   main_server->ServerPort = pr_inet_getservport(main_server->pool,
     "ftp", "tcp");
 
+  set_tcp_bufsz();
   return;
 }
 
@@ -3120,42 +3121,10 @@ char *get_context_name(cmd_rec *cmd) {
   }
 }
 
-/* Boolean string can be "on", "off", "yes", "no", "true", "false",
- * "1" or "0."
- */
-int pr_is_boolean(const char *str) {
-  if (strcasecmp(str, "on") == 0)
-    return 1;
-
-  if (strcasecmp(str, "off") == 0)
-    return 0;
-
-  if (strcasecmp(str, "yes") == 0)
-    return 1;
-
-  if (strcasecmp(str, "no") == 0)
-    return 0;
-
-  if (strcasecmp(str, "true") == 0)
-    return 1;
-
-  if (strcasecmp(str, "false") == 0)
-    return 0;
-
-  if (strcasecmp(str, "1") == 0)
-    return 1;
-
-  if (strcasecmp(str, "0") == 0)
-    return 0;
-
-  errno = EINVAL;
-  return -1;
-}
-
 int get_boolean(cmd_rec *cmd, int av) {
   char *cp = cmd->argv[av];
 
-  return pr_is_boolean(cp);
+  return pr_str_is_boolean(cp);
 }
 
 char *get_full_cmd(cmd_rec *cmd) {
@@ -3163,14 +3132,14 @@ char *get_full_cmd(cmd_rec *cmd) {
   char *res = "";
 
   if (cmd->arg && *cmd->arg)
-    res = pstrcat(p, cmd->argv[0], " ", cmd->arg, NULL);
+    res = pstrcat(p, cmd->argv[0], " ", pr_fs_decode_path(p, cmd->arg), NULL);
 
   else if (cmd->argc > 1) {
     register unsigned int i = 0;
     res = cmd->argv[0];
 
     for (i = 1; i < cmd->argc; i++)
-      res = pstrcat(p, res, cmd->argv[i], " ", NULL);
+      res = pstrcat(p, res, pr_fs_decode_path(p, cmd->argv[i]), " ", NULL);
 
     while (res[strlen(res)-1] == ' ' && *res)
       res[strlen(res)-1] = '\0';
@@ -3232,3 +3201,7 @@ unsigned int pr_config_set_id(const char *name) {
 
   return *ptr;
 }
+
+int pr_config_get_xfer_bufsz(void) {
+  return xfer_bufsz;
+}
