@@ -2,7 +2,7 @@
  * ProFTPD - FTP server daemon
  * Copyright (c) 1997, 1998 Public Flood Software
  * Copyright (c) 1999, 2000 MacGyver aka Habeeb J. Dihu <macgyver@tos.net>
- * Copyright (c) 2001-2005 The ProFTPD Project team
+ * Copyright (c) 2001-2007 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -25,7 +25,7 @@
  */
 
 /* Inet support functions, many wrappers for netdb functions
- * $Id: inet.c,v 1.94 2005/11/29 02:42:19 castaglia Exp $
+ * $Id: inet.c,v 1.112 2008/12/04 21:35:30 castaglia Exp $
  */
 
 #include "conf.h"
@@ -49,6 +49,8 @@ static int inet_errno = 0;		/* Holds errno */
  */
 static int inet_family = 0;
 
+static const char *trace_channel = "inet";
+
 /* Called by others after running a number of pr_inet_* functions in order
  * to free up memory.
  */
@@ -77,50 +79,37 @@ int pr_inet_getservport(pool *p, const char *serv, const char *proto) {
   return (servent ? ntohs(servent->s_port) : -1);
 }
 
-/* Validate anything returned from the 'outside', since it's untrusted
- * information.
- */
-char *pr_inet_validate(char *buf) {
-  char *p;
-
-  /* Validate anything returned from a DNS.
-   */
-  for (p = buf; p && *p; p++) {
-
-    /* Per RFC requirements, these are all that are valid from a DNS. */
-    if (!isalnum((int) *p) && *p != '.' && *p != '-'
-#ifdef PR_USE_IPV6
-        && *p != ':'
-#endif /* PR_USE_IPV6 */
-        ) {
-
-      /* We set it to _ because we know that's an invalid, yet safe, option
-       * for a DNS entry.
-       */
-      *p = '_';
-    }
-  }
-
-  return buf;
-}
-
 static void conn_cleanup_cb(void *cv) {
   conn_t *c = (conn_t *) cv;
 
-  if (c->instrm)
+  /* XXX These closes' return values should be checked, ideally. Do
+   * we really care if they fail, though?
+   */
+
+  if (c->instrm) {
     pr_netio_close(c->instrm);
+    c->instrm = NULL;
+  }
 
-  if (c->outstrm && c->outstrm != c->instrm)
+  if (c->outstrm && c->outstrm != c->instrm) {
     pr_netio_close(c->outstrm);
+    c->outstrm = NULL;
+  }
 
-  if (c->listen_fd != -1)
+  if (c->listen_fd != -1) {
     close(c->listen_fd);
+    c->listen_fd = -1;
+  }
 
-  if (c->rfd != -1)
+  if (c->rfd != -1) {
     close(c->rfd);
+    c->rfd = -1;
+  }
 
-  if (c->wfd != -1)
+  if (c->wfd != -1) {
     close(c->wfd);
+    c->wfd = -1;
+  }
 }
 
 /* Copy a connection structure, also creates a sub pool for the new
@@ -133,7 +122,7 @@ conn_t *pr_inet_copy_connection(pool *p, conn_t *c) {
   sub_pool = make_sub_pool(p);
   pr_pool_tag(sub_pool, "pr_inet_copy_connection() subpool");
 
-  res = (conn_t *) pcalloc(sub_pool,sizeof(conn_t));
+  res = (conn_t *) pcalloc(sub_pool, sizeof(conn_t));
 
   memcpy(res, c, sizeof(conn_t));
   res->pool = sub_pool;
@@ -209,7 +198,11 @@ static conn_t *inet_initialize_connection(pool *p, xaset_t *servers, int fd,
      * support is enabled), otherwise use IPv4.
      */
 #ifdef PR_USE_IPV6
-    addr_family = AF_INET6;
+    if (pr_netaddr_use_ipv6())
+      addr_family = AF_INET6;
+
+    else
+      addr_family = AF_INET;
 #else
     addr_family = AF_INET;
 #endif /* PR_USE_IPV6 */
@@ -232,11 +225,11 @@ static conn_t *inet_initialize_connection(pool *p, xaset_t *servers, int fd,
      */
 
 #if defined(SOLARIS2) || defined(FREEBSD2) || defined(FREEBSD3) || \
-    defined(FREEBSD4) || defined(FREEBSD5) || defined(__OpenBSD__) || \
-    defined(__NetBSD__) || defined(DARWIN6) || defined(DARWIN7) || \
-    defined(DARWIN8) || \
+    defined(FREEBSD4) || defined(FREEBSD5) || defined(FREEBSD6) || \
+    defined(FREEBSD7) || defined(__OpenBSD__) || defined(__NetBSD__) || \
+    defined(DARWIN6) || defined(DARWIN7) || defined(DARWIN8) || defined(DARWIN9) || \
     defined(SCO3) || defined(CYGWIN) || defined(SYSV4_2MP) || \
-    defined(SYSV5UNIXWARE7)
+    defined(SYSV5SCO_SV6) || defined(SYSV5UNIXWARE7)
 # ifdef SOLARIS2
     if (port != INPORT_ANY && port < 1024) {
 # endif
@@ -250,11 +243,11 @@ static conn_t *inet_initialize_connection(pool *p, xaset_t *servers, int fd,
     fd = socket(addr_family, SOCK_STREAM, tcp_proto);
 
 #if defined(SOLARIS2) || defined(FREEBSD2) || defined(FREEBSD3) || \
-    defined(FREEBSD4) || defined(FREEBSD5) || defined(__OpenBSD__) || \
-    defined(__NetBSD__) || defined(DARWIN6) || defined(DARWIN7) || \
-    defined(DARWIN8) || \
+    defined(FREEBSD4) || defined(FREEBSD5) || defined(FREEBSD6) || \
+    defined(FREEBSD7) || defined(__OpenBSD__) || defined(__NetBSD__) || \
+    defined(DARWIN6) || defined(DARWIN7) || defined(DARWIN8) || defined(DARWIN9) || \
     defined(SCO3) || defined(CYGWIN) || defined(SYSV4_2MP) || \
-    defined(SYSV5UNIXWARE7)
+    defined(SYSV5SCO_SV6) || defined(SYSV5UNIXWARE7)
 # ifdef SOLARIS2
     if (port != INPORT_ANY && port < 1024) {
 # endif
@@ -282,6 +275,12 @@ static conn_t *inet_initialize_connection(pool *p, xaset_t *servers, int fd,
       pr_log_pri(PR_LOG_NOTICE, "error setting SO_REUSEADDR: %s",
         strerror(errno));
 
+    /* Allow socket keep-alive messages. */
+    if (setsockopt(fd, SOL_SOCKET, SO_KEEPALIVE, (void *) &one,
+        sizeof(one)) < 0)
+      pr_log_pri(PR_LOG_NOTICE, "error setting SO_KEEPALIVE: %s",
+        strerror(errno));
+
     memset(&na, 0, sizeof(na));
     pr_netaddr_set_family(&na, addr_family);
 
@@ -292,14 +291,11 @@ static conn_t *inet_initialize_connection(pool *p, xaset_t *servers, int fd,
       pr_netaddr_set_sockaddr_any(&na);
 
 #if defined(PR_USE_IPV6) && defined(IPV6_V6ONLY)
-    if (addr_family == AF_INET6) {
-      int on = 0;
-
-# ifdef SOL_IP
-      int level = SOL_IP;
-# else
+    if (pr_netaddr_use_ipv6() &&
+        addr_family == AF_INET6) {
       int level = ipv6_proto;
-# endif /* SOL_IP */
+      int off;
+      socklen_t len = sizeof(off);
 
       /* If creating a wildcard socket IPv6 socket, make sure that it
        * will accept IPv4 connections as well.  This is the default on
@@ -310,15 +306,32 @@ static conn_t *inet_initialize_connection(pool *p, xaset_t *servers, int fd,
        * SocketOptions directive.
        */
 
-      if (setsockopt(fd, level, IPV6_V6ONLY, (void *) &on, sizeof(on)) < 0)
-        pr_log_pri(PR_LOG_NOTICE, "error setting IPV6_V6ONLY: %s",
+      if (getsockopt(fd, level, IPV6_V6ONLY, (void *) &off, &len) >= 0) {
+        if (off != 0) {
+          off = 0;
+
+          pr_trace_msg(trace_channel, 5,
+            "disabling IPV6_V6ONLY on server socket %d", fd);
+
+          if (setsockopt(fd, level, IPV6_V6ONLY, (void *) &off,
+              len) < 0) {
+            pr_log_pri(PR_LOG_NOTICE, "error setting IPV6_V6ONLY: %s",
+              strerror(errno));
+          }
+        }
+
+      } else {
+        pr_trace_msg(trace_channel, 3,
+          "error getting IPV6_V6ONLY setting on socket %d: %s", fd,
           strerror(errno));
+      }
     }
 #endif /* PR_USE_IPV6 and IPV6_V6ONLY */
 
     pr_netaddr_set_port(&na, htons(port));
 
-    if (port != INPORT_ANY && port < 1024) {
+    if (port != INPORT_ANY &&
+        port < 1024) {
       pr_signals_block();
       PRIVS_ROOT
     }
@@ -334,7 +347,8 @@ static conn_t *inet_initialize_connection(pool *p, xaset_t *servers, int fd,
         pr_netaddr_get_sockaddr_len(&na));
       hold_errno = errno;
 
-      if (res == -1 && errno == EINTR) {
+      if (res == -1 &&
+          errno == EINTR) {
         pr_signals_handle();
 	i++;
 	continue;
@@ -345,21 +359,24 @@ static conn_t *inet_initialize_connection(pool *p, xaset_t *servers, int fd,
 	  (port != INPORT_ANY && !retry_bind))
         break;
 
-      if (port != INPORT_ANY && port < 1024) {
+      if (port != INPORT_ANY &&
+          port < 1024) {
         PRIVS_RELINQUISH
         pr_signals_unblock();
       }
 
       pr_timer_sleep(1);
 
-      if (port != INPORT_ANY && port < 1024) {
+      if (port != INPORT_ANY &&
+          port < 1024) {
         pr_signals_block();
         PRIVS_ROOT
       }
     }
 
     if (res == -1) {
-      if (port != INPORT_ANY && port < 1024) {
+      if (port != INPORT_ANY &&
+          port < 1024) {
         PRIVS_RELINQUISH
         pr_signals_unblock();
       }
@@ -377,7 +394,8 @@ static conn_t *inet_initialize_connection(pool *p, xaset_t *servers, int fd,
       return NULL;
     }
 
-    if (port != INPORT_ANY && port < 1024) {
+    if (port != INPORT_ANY &&
+        port < 1024) {
       PRIVS_RELINQUISH
       pr_signals_unblock();
     }
@@ -395,11 +413,19 @@ static conn_t *inet_initialize_connection(pool *p, xaset_t *servers, int fd,
       pr_netaddr_set_sockaddr(c->local_addr, pr_netaddr_get_sockaddr(&na));
       c->local_port = ntohs(pr_netaddr_get_port(&na));
     }
+
+  } else {
+    /* Make sure the netaddr has its address family set. */
+    if (pr_netaddr_get_family(&na) == 0)
+      pr_netaddr_set_family(&na, addr_family);
   }
 
   c->listen_fd = fd;
   register_cleanup(c->pool, (void *) c, conn_cleanup_cb, conn_cleanup_cb);
 
+  pr_trace_msg("binding", 4, "bound address %s, port %d to socket fd %d",
+    pr_netaddr_get_ipstr(&na), port, fd);
+
   return c;
 }
 
@@ -506,7 +532,7 @@ void pr_inet_lingering_close(pool *p, conn_t *c, long linger) {
    * the output stream.
    */
   if (c->instrm != c->outstrm)
-    pr_netio_lingering_close(c->instrm, linger);
+    pr_netio_close(c->instrm);
 
   c->outstrm = NULL;
   c->instrm = NULL;
@@ -529,7 +555,7 @@ void pr_inet_lingering_abort(pool *p, conn_t *c, long linger) {
    * want and need one.
    */
   if (c->outstrm != c->instrm)
-    pr_netio_lingering_close(c->outstrm, linger);
+    pr_netio_close(c->outstrm);
 
   c->instrm = NULL;
   c->outstrm = NULL;
@@ -567,31 +593,27 @@ int pr_inet_set_proto_opts(pool *p, conn_t *c, int mss, int nodelay,
   unsigned char *no_delay = get_param_ptr(main_server->conf, "tcpNoDelay",
     FALSE);
 
-  if (!no_delay || *no_delay == TRUE) {
-    if (c->wfd != -1)
-      if (setsockopt(c->wfd, tcp_level, TCP_NODELAY, (void *) &nodelay,
-          sizeof(nodelay)) < 0)
-        pr_log_pri(PR_LOG_NOTICE, "error setting write fd TCP_NODELAY: %s",
-          strerror(errno));
-
-    if (c->rfd != -1)
-      if (setsockopt(c->rfd, IPPROTO_TCP, TCP_NODELAY, (void *) &nodelay,
-          sizeof(nodelay)) < 0)
-        pr_log_pri(PR_LOG_NOTICE, "error setting read fd TCP_NODELAY: %s",
+  if (!no_delay ||
+      *no_delay == TRUE) {
+    if (c->listen_fd != -1) {
+      if (setsockopt(c->listen_fd, tcp_level, TCP_NODELAY, (void *) &nodelay,
+          sizeof(nodelay)) < 0) {
+        pr_log_pri(PR_LOG_NOTICE, "error setting listen fd TCP_NODELAY: %s",
           strerror(errno));
+      }
+    }
   }
 #endif /* TCP_NODELAY */
 
 #ifdef TCP_MAXSEG
-  if (c->wfd != -1 && mss)
-    if (setsockopt(c->wfd, tcp_level, TCP_MAXSEG, &mss, sizeof(mss)) < 0)
-      pr_log_pri(PR_LOG_NOTICE, "error setting write fd TCP_MAXSEG(%d): %s",
-        mss, strerror(errno));
-
-  if (c->rfd != -1 && mss)
-    if (setsockopt(c->wfd, tcp_level, TCP_MAXSEG, &mss, sizeof(mss)) < 0)
-      pr_log_pri(PR_LOG_NOTICE, "error setting read fd TCP_MAXSEG(%d): %s",
+  if (c->listen_fd != -1 &&
+      mss) {
+    if (setsockopt(c->listen_fd, tcp_level, TCP_MAXSEG, &mss,
+        sizeof(mss)) < 0) {
+      pr_log_pri(PR_LOG_NOTICE, "error setting listen fd TCP_MAXSEG(%d): %s",
         mss, strerror(errno));
+    }
+  }
 #endif /* TCP_MAXSEG */
 
 #ifdef IPTOS_LOWDELAY
@@ -610,31 +632,25 @@ int pr_inet_set_proto_opts(pool *p, conn_t *c, int mss, int nodelay,
    * them.
    */
   if (pr_netaddr_get_family(c->local_addr) == AF_INET) {
-    if (c->wfd != -1)
-      if (setsockopt(c->wfd, ip_level, IP_TOS, (void *) &tos, sizeof(tos)) < 0)
-        pr_log_pri(PR_LOG_NOTICE, "error setting write fd IP_TOS: %s",
-          strerror(errno));
-
-    if (c->rfd != -1)
-      if (setsockopt(c->rfd, ip_level, IP_TOS, (void *) &tos, sizeof(tos)) < 0)
-        pr_log_pri(PR_LOG_NOTICE, "error setting read fd IP_TOS: %s",
+    if (c->listen_fd != -1) {
+      if (setsockopt(c->listen_fd, ip_level, IP_TOS, (void *) &tos,
+          sizeof(tos)) < 0) {
+        pr_log_pri(PR_LOG_NOTICE, "error setting listen fd IP_TOS: %s",
           strerror(errno));
+      }
+    }
   }
 #endif /* IP_TOS */
 
 #ifdef TCP_NOPUSH
   /* NOTE: TCP_NOPUSH is a BSDism. */
-  if (c->wfd != -1)
-    if (setsockopt(c->wfd, tcp_level, TCP_NOPUSH, (void *) &nopush,
-        sizeof(nopush)) < 0)
-      pr_log_pri(PR_LOG_NOTICE, "error setting write fd TCP_NOPUSH: %s",
-        strerror(errno));
-
-  if (c->rfd != -1)
-    if (setsockopt(c->rfd, tcp_level, TCP_NOPUSH, (void *) &nopush,
-        sizeof(nopush)) < 0)
-      pr_log_pri(PR_LOG_NOTICE, "error setting read fd TCP_NOPUSH: %s",
+  if (c->listen_fd != -1) {
+    if (setsockopt(c->listen_fd, tcp_level, TCP_NOPUSH, (void *) &nopush,
+        sizeof(nopush)) < 0) {
+      pr_log_pri(PR_LOG_NOTICE, "error setting listen fd TCP_NOPUSH: %s",
         strerror(errno));
+    }
+  }
 #endif /* TCP_NOPUSH */
 
   return 0;
@@ -642,9 +658,6 @@ int pr_inet_set_proto_opts(pool *p, conn_t *c, int mss, int nodelay,
 
 /* Set socket options on a connection.  */
 int pr_inet_set_socket_opts(pool *p, conn_t *c, int rcvbuf, int sndbuf) {
-  int no_keep_alive = 0;
-  int crcvbuf, csndbuf;
-  socklen_t len;
 
   /* Linux and "most" newer networking OSes probably use a highly adaptive
    * window size system, which generally wouldn't require user-space
@@ -653,38 +666,39 @@ int pr_inet_set_socket_opts(pool *p, conn_t *c, int rcvbuf, int sndbuf) {
    * than their current size.
    */
 
-  if (c->wfd != -1) {
-    if (setsockopt(c->wfd, SOL_SOCKET, SO_KEEPALIVE, (void *) &no_keep_alive,
-        sizeof(int)) < 0)
-      pr_log_pri(PR_LOG_NOTICE, "error setting write fd SO_KEEPALIVE: %s",
+  if (c->listen_fd != -1) {
+    int keepalive = 0;
+    int crcvbuf, csndbuf;
+    socklen_t len;
+
+    if (setsockopt(c->listen_fd, SOL_SOCKET, SO_KEEPALIVE, (void *)
+        &keepalive, sizeof(int)) < 0)
+      pr_log_pri(PR_LOG_NOTICE, "error setting listen fd SO_KEEPALIVE: %s",
         strerror(errno));
 
     len = sizeof(csndbuf);
-    getsockopt(c->wfd, SOL_SOCKET, SO_SNDBUF, (void *) &csndbuf, &len);
+    getsockopt(c->listen_fd, SOL_SOCKET, SO_SNDBUF, (void *) &csndbuf, &len);
 
-    if (sndbuf && sndbuf > csndbuf)
-      if (setsockopt(c->wfd, SOL_SOCKET, SO_SNDBUF, (void *) &sndbuf,
+    if (sndbuf &&
+        sndbuf > csndbuf) {
+      if (setsockopt(c->listen_fd, SOL_SOCKET, SO_SNDBUF, (void *) &sndbuf,
           sizeof(sndbuf)) < 0)
-        pr_log_pri(PR_LOG_NOTICE, "error setting SO_SNDBUF: %s",
+        pr_log_pri(PR_LOG_NOTICE, "error setting listen fd SO_SNDBUF: %s",
           strerror(errno));
+    }
 
     c->sndbuf = (sndbuf ? sndbuf : csndbuf);
-  }
-
-  if (c->rfd != -1) {
-    if (setsockopt(c->rfd, SOL_SOCKET, SO_KEEPALIVE, (void *) &no_keep_alive,
-        sizeof(int)) < 0)
-      pr_log_pri(PR_LOG_NOTICE, "error setting read fd SO_KEEPALIVE: %s",
-        strerror(errno));
 
     len = sizeof(crcvbuf);
-    getsockopt(c->rfd, SOL_SOCKET, SO_RCVBUF, (void *) &crcvbuf, &len);
+    getsockopt(c->listen_fd, SOL_SOCKET, SO_RCVBUF, (void *) &crcvbuf, &len);
 
-    if (rcvbuf && rcvbuf > crcvbuf)
-      if (setsockopt(c->rfd, SOL_SOCKET, SO_RCVBUF, (void *) &rcvbuf,
+    if (rcvbuf &&
+        rcvbuf > crcvbuf) {
+      if (setsockopt(c->listen_fd, SOL_SOCKET, SO_RCVBUF, (void *) &rcvbuf,
           sizeof(rcvbuf)) < 0)
-        pr_log_pri(PR_LOG_NOTICE, "error setting SO_RCVFBUF: %s",
+        pr_log_pri(PR_LOG_NOTICE, "error setting listen fd SO_RCVFBUF: %s",
           strerror(errno));
+    }
 
     c->rcvbuf = (rcvbuf ? rcvbuf : crcvbuf);
   }
@@ -705,7 +719,7 @@ static void set_oobinline(int fd) {
 #ifdef F_SETOWN
 static void set_owner(int fd) {
   if (fd != -1)
-    fcntl(fd, F_SETOWN, getpid());
+    fcntl(fd, F_SETOWN, session.pid ? session.pid : getpid());
 }
 #endif
 
@@ -714,8 +728,16 @@ static void set_owner(int fd) {
 int pr_inet_set_async(pool *p, conn_t *c) {
 
 #ifdef SO_OOBINLINE
+  pr_trace_msg(trace_channel, 7,
+    "setting SO_OOBINLINE for listening socket %d",  c->listen_fd);
   set_oobinline(c->listen_fd);
+
+  pr_trace_msg(trace_channel, 7,
+    "setting SO_OOBINLINE for reading socket %d",  c->rfd);
   set_oobinline(c->rfd);
+
+  pr_trace_msg(trace_channel, 7,
+    "setting SO_OOBINLINE for writing socket %d",  c->wfd);
   set_oobinline(c->wfd);
 #endif
 
@@ -1019,7 +1041,11 @@ int pr_inet_get_conn_info(conn_t *c, int fd) {
   pr_netaddr_clear(&na);
 
 #ifdef PR_USE_IPV6
-  pr_netaddr_set_family(&na, AF_INET6);
+  if (pr_netaddr_use_ipv6())
+    pr_netaddr_set_family(&na, AF_INET6);
+
+  else
+    pr_netaddr_set_family(&na, AF_INET);
 #else
   pr_netaddr_set_family(&na, AF_INET);
 #endif /* PR_USE_IPV6 */
@@ -1044,7 +1070,11 @@ int pr_inet_get_conn_info(conn_t *c, int fd) {
 
   /* "Reset" the pr_netaddr_t struct for the getpeername(2) call. */
 #ifdef PR_USE_IPV6
-  pr_netaddr_set_family(&na, AF_INET6);
+  if (pr_netaddr_use_ipv6())
+    pr_netaddr_set_family(&na, AF_INET6);
+
+  else
+    pr_netaddr_set_family(&na, AF_INET);
 #else
   pr_netaddr_set_family(&na, AF_INET);
 #endif /* PR_USE_IPV6 */
@@ -1064,59 +1094,6 @@ int pr_inet_get_conn_info(conn_t *c, int fd) {
   return 0;
 }
 
-/* Associate already open streams with the connection, returns NULL if either
- * stream points to a non-socket descriptor.  If addr is non-NULL, remote
- * address discovery is attempted. If resolve is non-zero, the remote address
- * is reverse resolved.
- */
-conn_t *pr_inet_associate(pool *p, conn_t *c, pr_netaddr_t *addr,
-    pr_netio_stream_t *in, pr_netio_stream_t *out, int resolve) {
-  int rfd, wfd;
-  int socktype;
-  socklen_t socktype_len = sizeof(socktype);
-  conn_t *res;
-
-  rfd = PR_NETIO_FD(in);
-  wfd = PR_NETIO_FD(out);
-
-  if (getsockopt(rfd, SOL_SOCKET, SO_TYPE, (void *) &socktype,
-      &socktype_len) == -1 || socktype != SOCK_STREAM)
-    return NULL;
-
-  if (getsockopt(wfd, SOL_SOCKET, SO_TYPE, (void *) &socktype,
-      &socktype_len) == -1 || socktype != SOCK_STREAM)
-    return NULL;
-
-  res = pr_inet_copy_connection(p, c);
-
-  res->rfd = rfd;
-  res->wfd = wfd;
-  res->instrm = in;
-  res->outstrm = out;
-  res->mode = CM_OPEN;
-
-  if (pr_inet_get_conn_info(res, wfd) < 0)
-    return NULL;
-
-  /* Get the remote address */
-
-  if (addr) {
-    if (!res->remote_addr)
-      res->remote_addr = pr_netaddr_alloc(res->pool);
-
-    memcpy(res->remote_addr, addr, sizeof(pr_netaddr_t));
-  }
-
-  if (resolve && res->remote_addr)
-    res->remote_name = pr_netaddr_get_dnsstr(res->remote_addr);
-
-  if (!res->remote_name)
-    res->remote_name = pr_netaddr_get_ipstr(res->remote_addr);
-
-  pr_inet_set_socket_opts(res->pool, res, 0, 0);
-  return res;
-}
-
 /* Open streams for a new socket. If rfd and wfd != -1, two new fds are duped
  * to the respective read/write fds. If the fds specified correspond to the
  * normal stdin and stdout, the streams opened will be assigned to stdin and
