@@ -2,7 +2,7 @@
  * ProFTPD - FTP server daemon
  * Copyright (c) 1997, 1998 Public Flood Software
  * Copyright (c) 1999, 2000 MacGyver aka Habeeb J. Dihu <macgyver@tos.net>
- * Copyright (c) 2001-2007 The ProFTPD Project team
+ * Copyright (c) 2001-2008 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -25,7 +25,7 @@
  */
 
 /* ProFTPD logging support.
- * $Id: log.c,v 1.82 2007/09/11 00:49:44 castaglia Exp $
+ * $Id: log.c,v 1.89 2008/12/16 23:43:55 castaglia Exp $
  */
 
 #include "conf.h"
@@ -94,7 +94,7 @@ int log_wtmp(char *line, const char *name, const char *host,
     sstrncpy(utx.ut_id, "ftp", sizeof(utx.ut_user));
     sstrncpy(utx.ut_line, line, sizeof(utx.ut_line));
     sstrncpy(utx.ut_host, host, sizeof(utx.ut_host));
-    utx.ut_pid = getpid();
+    utx.ut_pid = session.pid ? session.pid : getpid();
 #if defined(__NetBSD__) && defined(HAVE_UTMPX_H)
     memcpy(&utx.ut_ss, pr_netaddr_get_inaddr(ip), sizeof(utx.ut_ss));
     gettimeofday(&utx.ut_tv, NULL);
@@ -152,7 +152,7 @@ int log_wtmp(char *line, const char *name, const char *host,
     sstrncpy(ut.ut_line, line, sizeof(ut.ut_line));
     if (name && *name)
       sstrncpy(ut.ut_user, name, sizeof(ut.ut_user));
-    ut.ut_pid = getpid();
+    ut.ut_pid = session.pid ? session.pid : getpid();
     if (name && *name)
       ut.ut_type = USER_PROCESS;
     else
@@ -168,12 +168,16 @@ int log_wtmp(char *line, const char *name, const char *host,
       sstrncpy(ut.ut_host, host, sizeof(ut.ut_host));
 #endif /* HAVE_UT_UT_HOST */
 
-    time(&ut.ut_time);
-    if (write(fd, (char *)&ut, sizeof(ut)) != sizeof(ut))
-      ftruncate(fd, buf.st_size);
+    ut.ut_time = time(NULL);
+    if (write(fd, (char *)&ut, sizeof(ut)) != sizeof(ut)) {
+      if (ftruncate(fd, buf.st_size) < 0) {
+        pr_log_debug(DEBUG0, "error truncating '%s': %s", WTMP_FILE,
+          strerror(errno));
+      }
+    }
 
   } else {
-    pr_log_debug(DEBUG0, "%s fstat(): %s",WTMP_FILE,strerror(errno));
+    pr_log_debug(DEBUG0, "%s fstat(): %s", WTMP_FILE, strerror(errno));
     res = -1;
   }
 #endif /* SVR4 */
@@ -367,7 +371,7 @@ int pr_log_writefile(int logfd, const char *ident, const char *fmt, ...) {
 
   /* Prepend a small header */
   snprintf(buf + strlen(buf), sizeof(buf) - strlen(buf), "%s[%u]: ",
-    ident, (unsigned int) getpid());
+    ident, (unsigned int) (session.pid ? session.pid : getpid()));
   buf[sizeof(buf)-1] = '\0';
 
   /* Affix the message */
@@ -406,9 +410,10 @@ int log_opensyslog(const char *fn) {
     /* The child may have inherited a valid socket from the parent. */
     pr_closelog(syslog_sockfd);
 
-    if ((syslog_sockfd = pr_openlog("proftpd", LOG_NDELAY|LOG_PID,
-        facility)) < 0)
+    syslog_sockfd = pr_openlog("proftpd", LOG_NDELAY|LOG_PID, facility);
+    if (syslog_sockfd < 0)
       return -1;
+
     systemlog_fd = -1;
 
   } else if ((res = pr_log_openfile(systemlog_fn, &systemlog_fd,
@@ -422,14 +427,11 @@ int log_opensyslog(const char *fn) {
 }
 
 void log_closesyslog(void) {
-  if (systemlog_fd != -1) {
-    close(systemlog_fd);
-    systemlog_fd = -1;
+  (void) close(systemlog_fd);
+  systemlog_fd = -1;
 
-  } else {
-    pr_closelog(syslog_sockfd);
-    syslog_sockfd = -1;
-  }
+  (void) pr_closelog(syslog_sockfd);
+  syslog_sockfd = -1;
 
   syslog_open = FALSE;
 }
@@ -489,11 +491,12 @@ static void log_write(int priority, int f, char *s) {
     if (*serverinfo) {
       snprintf(buf + strlen(buf), sizeof(buf) - strlen(buf),
 	       "%s proftpd[%u] %s: %s\n", systemlog_host,
-	       (unsigned int) getpid(), serverinfo, s);
+	       (unsigned int) (session.pid ? session.pid : getpid()),
+               serverinfo, s);
     } else {
       snprintf(buf + strlen(buf), sizeof(buf) - strlen(buf),
 	       "%s proftpd[%u]: %s\n", systemlog_host,
-	       (unsigned int) getpid(), s);
+	       (unsigned int) (session.pid ? session.pid : getpid()), s);
     }
 
     buf[sizeof(buf) - 1] = '\0';
@@ -513,6 +516,7 @@ static void log_write(int priority, int f, char *s) {
 
   if (!syslog_open) {
     syslog_sockfd = pr_openlog("proftpd", LOG_NDELAY|LOG_PID, f);
+    syslog_open = TRUE;
 
   } else if (f != facility) {
     (void) pr_setlogfacility(f);
@@ -527,14 +531,6 @@ static void log_write(int priority, int f, char *s) {
     pr_syslog(syslog_sockfd, priority, "%s - %s\n", serverinfo, s);
   else
     pr_syslog(syslog_sockfd, priority, "%s\n", s);
-
-  if (!syslog_open) {
-    pr_closelog(syslog_sockfd);
-    syslog_sockfd = -1;
-
-  } else if (f != facility) {
-    (void) pr_setlogfacility(f);
-  }
 }
 
 void pr_log_pri(int priority, const char *fmt, ...) {
@@ -615,6 +611,7 @@ int pr_log_str2sysloglevel(const char *name) {
   else if (strcasecmp(name, "debug") == 0)
     return PR_LOG_DEBUG;
 
+  errno = ENOENT;
   return -1;
 }
 
@@ -625,6 +622,9 @@ void pr_log_debug(int level, const char *fmt, ...) {
   if (debug_level < level)
     return;
 
+  if (fmt == NULL)
+    return;
+
   memset(buf, '\0', sizeof(buf));
   va_start(msg, fmt);
   vsnprintf(buf, sizeof(buf), fmt, msg);
@@ -637,12 +637,13 @@ void pr_log_debug(int level, const char *fmt, ...) {
 }
 
 void init_log(void) {
-  char buf[256] = {'\0'};
+  char buf[256];
 
+  memset(buf, '\0', sizeof(buf));
   if (gethostname(buf, sizeof(buf)) == -1)
     sstrncpy(buf, "localhost", sizeof(buf));
 
-  sstrncpy(systemlog_host, (char *) pr_inet_validate(buf),
+  sstrncpy(systemlog_host, (char *) pr_netaddr_validate_dns_str(buf),
     sizeof(systemlog_host));
   memset(systemlog_fn, '\0', sizeof(systemlog_fn));
   log_closesyslog();
