@@ -1,7 +1,7 @@
 /*
  * ProFTPD - FTP server daemon
  * Copyright (c) 1997, 1998 Public Flood Software
- * Copyright (c) 2001, 2002, 2003 The ProFTPD Project team
+ * Copyright (c) 2001-2008 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -25,7 +25,7 @@
 
 /*
  * Module handling routines
- * $Id: modules.c,v 1.51 2007/01/19 21:59:45 castaglia Exp $
+ * $Id: modules.c,v 1.56 2008/10/04 04:53:59 castaglia Exp $
  */
 
 #include "conf.h"
@@ -69,38 +69,62 @@ typedef struct mod_cb {
 
 /* Symbol stash lookup code and management */
 
-/* This wrapper will be used in the future to track when to rehash through
- * the symbol memory, to prevent symbol_pool from growing too large.
- */
 static struct stash *sym_alloc(void) {
-  static unsigned int count = 0;
+  pool *sub_pool;
+  struct stash *sym;
 
-  pool *sub_pool = make_sub_pool(symbol_pool);
-  struct stash *sym = pcalloc(sub_pool, sizeof(struct stash));
+  /* XXX Use a smaller pool size, since there are lots of sub-pools allocated
+   * for Stash symbols.  The default pool size (PR_TUNABLE_POOL_SIZE, 512
+   * by default) is a bit large for symbols.
+   */
+  sub_pool = pr_pool_create_sz(symbol_pool, 128);
+
+  sym = pcalloc(sub_pool, sizeof(struct stash));
   sym->sym_pool = sub_pool; 
-  pr_pool_tag(sub_pool, "symbol subpool");
-  count++;
+  pr_pool_tag(sub_pool, "symbol");
 
   return sym;
 }
 
 static int sym_cmp(struct stash *s1, struct stash *s2) {
-  int ret;
+  int res;
 
-  ret = strcmp(s1->sym_name, s2->sym_name);
+  res = strcmp(s1->sym_name, s2->sym_name);
 
   /* Higher priority modules must go BEFORE lower priority in the
    * hash tables.
    */
 
-  if (!ret) {
-    if (s1->sym_module->priority > s2->sym_module->priority)
-      ret = -1;
-    else if (s1->sym_module->priority < s2->sym_module->priority)
-      ret = 1;
+  if (res == 0) {
+    if (s1->sym_module != NULL &&
+        s2->sym_module != NULL) {
+
+      if (s1->sym_module->priority > s2->sym_module->priority) {
+        return -1;
+      }
+    
+      if (s1->sym_module->priority < s2->sym_module->priority) {
+        return 1;
+      }
+
+      return res;
+    }
+
+    if (s1->sym_module != NULL &&
+        s2->sym_module == NULL) {
+      return -1;
+    }
+
+    if (s1->sym_module == NULL &&
+        s2->sym_module != NULL) {
+      return 1;
+    }
+
+    /* Both sym_module fields are null. */
+    return 0;
   }
 
-  return ret;
+  return res;
 }
 
 static int symtab_hash(const char *name) {
@@ -160,15 +184,22 @@ int pr_stash_add_symbol(pr_stash_type_t sym_type, void *data) {
       break;
 
     default:
-      errno = EINVAL;
+      errno = ENOENT;
       return -1;
   }
 
+  /* XXX Should we check for null sym->sym_module as well? */
+  if (sym->sym_name == NULL) {
+    destroy_pool(sym->sym_pool);
+    errno = EPERM;
+    return -1;
+  }
+
   /* XXX Ugly hack to support mixed cases of directives in config files. */
-  if (sym_type != PR_SYM_CONF)
+  if (sym_type != PR_SYM_CONF) {
     idx = symtab_hash(sym->sym_name);
 
-  else {
+  } else {
     register unsigned int i;
     char buf[1024];
 
@@ -181,11 +212,12 @@ int pr_stash_add_symbol(pr_stash_type_t sym_type, void *data) {
     idx = symtab_hash(buf);
   }
 
-  if (!symbol_table[idx])
+  if (!symbol_table[idx]) {
     symbol_table[idx] = xaset_create(symbol_pool, (XASET_COMPARE) sym_cmp);
+  }
 
   xaset_insert_sort(symbol_table[idx], (xasetmember_t *) sym, TRUE);
-  return idx;
+  return 0;
 }
 
 static struct stash *stash_lookup(pr_stash_type_t sym_type,
@@ -234,10 +266,10 @@ void *pr_stash_get_symbol(pr_stash_type_t sym_type, const char *name,
   } else {
 
     /* XXX Ugly hack to support mixed cases of directives in config files. */
-    if (sym_type != PR_SYM_CONF)
+    if (sym_type != PR_SYM_CONF) {
       idx = symtab_hash(name);
 
-    else {
+    } else {
       register unsigned int i;
       char buf[1024];
 
@@ -269,19 +301,38 @@ void *pr_stash_get_symbol(pr_stash_type_t sym_type, const char *name,
 
   switch (sym_type) {
     case PR_SYM_CONF:
-      return sym ? sym->ptr.sym_conf : NULL;
+      if (sym) {
+        return sym->ptr.sym_conf;
+      }
+
+      errno = ENOENT;
+      return NULL;
 
     case PR_SYM_CMD:
-      return sym ? sym->ptr.sym_cmd : NULL;
+      if (sym) {
+        return sym->ptr.sym_cmd;
+      }
+
+      errno = ENOENT;
+      return NULL;
 
     case PR_SYM_AUTH:
-      return sym ? sym->ptr.sym_auth : NULL;
+      if (sym) {
+        return sym->ptr.sym_auth;
+      }
+
+      errno = ENOENT;
+      return NULL;
 
     case PR_SYM_HOOK:
-      return sym ? sym->ptr.sym_hook : NULL;
+      if (sym) {
+        return sym->ptr.sym_hook;
+      }
+
+      errno = ENOENT;
+      return NULL;
   }
 
-  /* In case the compiler complains */
   errno = EINVAL;
   return NULL;
 }
@@ -296,10 +347,10 @@ int pr_stash_remove_symbol(pr_stash_type_t sym_type, const char *sym_name,
   }
 
   /* XXX Ugly hack to support mixed cases of directives in config files. */
-  if (sym_type != PR_SYM_CONF)
+  if (sym_type != PR_SYM_CONF) {
     symtab_idx = symtab_hash(sym_name);
 
-  else {
+  } else {
     register unsigned int i;
     char buf[1024];
 
@@ -334,6 +385,7 @@ int pr_stash_remove_symbol(pr_stash_type_t sym_type, const char *sym_name,
           destroy_pool(curr_sym->sym_pool);
           curr_sym = NULL;
           tab = NULL;
+          count++;
         }
 
         tab = pr_stash_get_symbol(PR_SYM_CONF, sym_name, tab, &idx);
@@ -361,6 +413,7 @@ int pr_stash_remove_symbol(pr_stash_type_t sym_type, const char *sym_name,
           xaset_remove(symbol_table[symtab_idx], (xasetmember_t *) curr_sym);
           destroy_pool(curr_sym->sym_pool);
           tab = NULL;
+          count++;
         }
 
         tab = pr_stash_get_symbol(PR_SYM_CMD, sym_name, tab, &idx);
@@ -388,6 +441,7 @@ int pr_stash_remove_symbol(pr_stash_type_t sym_type, const char *sym_name,
           xaset_remove(symbol_table[symtab_idx], (xasetmember_t *) curr_sym);
           destroy_pool(curr_sym->sym_pool);
           tab = NULL;
+          count++;
         }
 
         tab = pr_stash_get_symbol(PR_SYM_AUTH, sym_name, tab, &idx);
@@ -410,6 +464,7 @@ int pr_stash_remove_symbol(pr_stash_type_t sym_type, const char *sym_name,
           xaset_remove(symbol_table[symtab_idx], (xasetmember_t *) curr_sym);
           destroy_pool(curr_sym->sym_pool);
           tab = NULL;
+          count++;
         }
 
         tab = pr_stash_get_symbol(PR_SYM_HOOK, sym_name, tab, &idx);
@@ -426,13 +481,21 @@ int pr_stash_remove_symbol(pr_stash_type_t sym_type, const char *sym_name,
   return count;
 }
 
-modret_t *call_module(module *m, modret_t *(*func)(cmd_rec *), cmd_rec *cmd) {
+modret_t *pr_module_call(module *m, modret_t *(*func)(cmd_rec *),
+    cmd_rec *cmd) {
   modret_t *res;
   module *prev_module = curr_module;
 
+  if (m == NULL ||
+      func == NULL ||
+      cmd == NULL) {
+    errno = EINVAL;
+    return NULL;
+  }
+
   if (!cmd->tmp_pool) {
     cmd->tmp_pool = make_sub_pool(cmd->pool);
-    pr_pool_tag(cmd->tmp_pool, "call_module() cmd tmp_pool");
+    pr_pool_tag(cmd->tmp_pool, "Module call tmp_pool");
   }
 
   curr_module = m;
@@ -445,37 +508,41 @@ modret_t *call_module(module *m, modret_t *(*func)(cmd_rec *), cmd_rec *cmd) {
   return res;
 }
 
-modret_t *mod_create_data(cmd_rec *cmd,void *d) {
-  modret_t *ret;
+modret_t *mod_create_data(cmd_rec *cmd, void *d) {
+  modret_t *res;
 
-  ret = pcalloc(cmd->tmp_pool, sizeof(modret_t));
-  ret->data = d;
+  res = pcalloc(cmd->tmp_pool, sizeof(modret_t));
+  res->data = d;
 
-  return ret;
+  return res;
 }
 
 modret_t *mod_create_ret(cmd_rec *cmd, unsigned char err, char *n, char *m) {
-  modret_t *ret;
+  modret_t *res;
 
-  ret = pcalloc(cmd->tmp_pool, sizeof(modret_t));
-  ret->mr_handler_module = curr_module;
-  ret->mr_error = err;
-  if (n)
-    ret->mr_numeric = pstrdup(cmd->tmp_pool, n);
-  if (m)
-    ret->mr_message = pstrdup(cmd->tmp_pool, m);
+  res = pcalloc(cmd->tmp_pool, sizeof(modret_t));
+  res->mr_handler_module = curr_module;
+  res->mr_error = err;
 
-  return ret;
+  if (n) {
+    res->mr_numeric = pstrdup(cmd->tmp_pool, n);
+  }
+
+  if (m) {
+    res->mr_message = pstrdup(cmd->tmp_pool, m);
+  }
+
+  return res;
 }
 
 modret_t *mod_create_error(cmd_rec *cmd, int mr_errno) {
-  modret_t *ret;
+  modret_t *res;
 
-  ret = pcalloc(cmd->tmp_pool, sizeof(modret_t));
-  ret->mr_handler_module = curr_module;
-  ret->mr_error = mr_errno;
+  res = pcalloc(cmd->tmp_pool, sizeof(modret_t));
+  res->mr_handler_module = curr_module;
+  res->mr_error = mr_errno;
 
-  return ret;
+  return res;
 }
 
 /* Called after forking in order to inform/initialize modules
@@ -538,20 +605,56 @@ module *pr_module_get(const char *name) {
   return NULL;
 }
 
-void modules_list(void) {
-  register unsigned int i = 0;
+void modules_list(int flags) {
 
-  printf("Compiled-in modules:\n");
-  for (i = 0; static_modules[i]; i++) {
-    module *m = static_modules[i];
-    printf("  mod_%s.c\n", m->name);
+  if (flags & PR_MODULES_LIST_FL_SHOW_STATIC) {
+    register unsigned int i = 0;
+
+    printf("Compiled-in modules:\n");
+    for (i = 0; static_modules[i]; i++) {
+      module *m = static_modules[i];
+
+      if (flags & PR_MODULES_LIST_FL_SHOW_VERSION) {
+        char *version = m->module_version;
+        if (version) {
+          printf("  %s\n", version);
+
+        } else {
+          printf("  mod_%s.c\n", m->name);
+        }
+
+      } else {
+        printf("  mod_%s.c\n", m->name);
+      }
+    }
+
+  } else {
+    module *m;
+
+    printf("Loaded modules:\n");
+    for (m = loaded_modules; m; m = m->next) {
+
+      if (flags & PR_MODULES_LIST_FL_SHOW_VERSION) {
+        char *version = m->module_version;
+        if (version) {
+          printf("  %s\n", version);
+
+        } else {  
+          printf("  mod_%s.c\n", m->name);
+        }
+
+      } else {
+        printf("  mod_%s.c\n", m->name);
+      }
+    }
   }
 }
 
 int pr_module_load(module *m) {
   char buf[256];
 
-  if (!m) {
+  if (m == NULL ||
+      m->name == NULL) {
     errno = EINVAL;
     return -1;
   }
@@ -585,7 +688,10 @@ int pr_module_load(module *m) {
 
       for (conftab = m->conftable; conftab->directive; conftab++) {
         conftab->m = m;
-        pr_stash_add_symbol(PR_SYM_CONF, conftab);
+
+        if (pr_stash_add_symbol(PR_SYM_CONF, conftab) < 0) {
+          return -1;
+        }
       }
     }
 
@@ -595,12 +701,17 @@ int pr_module_load(module *m) {
       for (cmdtab = m->cmdtable; cmdtab->command; cmdtab++) {
         cmdtab->m = m;
 
-        if (cmdtab->cmd_type == HOOK)
-          pr_stash_add_symbol(PR_SYM_HOOK, cmdtab);
+        if (cmdtab->cmd_type == HOOK) {
+          if (pr_stash_add_symbol(PR_SYM_HOOK, cmdtab) < 0) {
+            return -1;
+          }
 
-        else
+        } else {
           /* All other cmd_types are for CMDs: PRE_CMD, CMD, POST_CMD, etc. */
-          pr_stash_add_symbol(PR_SYM_CMD, cmdtab);
+          if (pr_stash_add_symbol(PR_SYM_CMD, cmdtab) < 0) {
+            return -1;
+          }
+        }
       }
     }
 
@@ -609,7 +720,10 @@ int pr_module_load(module *m) {
 
       for (authtab = m->authtable; authtab->name; authtab++) {
         authtab->m = m;
-        pr_stash_add_symbol(PR_SYM_AUTH, authtab);
+
+        if (pr_stash_add_symbol(PR_SYM_AUTH, authtab) < 0) {
+          return -1;
+        }
       }
     }
 
@@ -634,7 +748,8 @@ int pr_module_load(module *m) {
 int pr_module_unload(module *m) {
   char buf[256];
 
-  if (!m) {
+  if (m == NULL ||
+      m->name == NULL) {
     errno = EINVAL;
     return -1;
   }
@@ -656,14 +771,15 @@ int pr_module_unload(module *m) {
   pr_event_generate("core.module-unload", buf);
 
   /* Remove the module from the loaded_modules list. */
-  if (m->prev)
+  if (m->prev) {
     m->prev->next = m->next;
 
-  else
+  } else {
     /* This module is the start of the loaded_modules list (prev is NULL),
      * so we need to update that pointer, too.
      */
     loaded_modules = m->next;
+  }
 
   if (m->next)
     m->next->prev = m->prev;
@@ -674,28 +790,31 @@ int pr_module_unload(module *m) {
   if (m->conftable) {
     conftable *conftab;
 
-    for (conftab = m->conftable; conftab->directive; conftab++)
+    for (conftab = m->conftable; conftab->directive; conftab++) {
       pr_stash_remove_symbol(PR_SYM_CONF, conftab->directive, conftab->m);
+    }
   }
 
   if (m->cmdtable) {
     cmdtable *cmdtab;
 
     for (cmdtab = m->cmdtable; cmdtab->command; cmdtab++) {
-      if (cmdtab->cmd_type == HOOK)
+      if (cmdtab->cmd_type == HOOK) {
         pr_stash_remove_symbol(PR_SYM_HOOK, cmdtab->command, cmdtab->m);
 
-      else
+      } else {
         /* All other cmd_types are for CMDs: PRE_CMD, CMD, POST_CMD, etc. */
         pr_stash_remove_symbol(PR_SYM_CMD, cmdtab->command, cmdtab->m);
+      }
     }
   }
 
   if (m->authtable) {
     authtable *authtab;
 
-    for (authtab = m->authtable; authtab->name; authtab++)
+    for (authtab = m->authtable; authtab->name; authtab++) {
       pr_stash_remove_symbol(PR_SYM_AUTH, authtab->name, authtab->m);
+    }
   }
 
   return 0;
