@@ -1,6 +1,6 @@
 /*
  * ProFTPD - FTP server daemon
- * Copyright (c) 2003-2005 The ProFTPD Project team
+ * Copyright (c) 2003-2008 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -23,7 +23,7 @@
  */
 
 /* Network address routines
- * $Id: netaddr.c,v 1.51 2005/09/19 21:35:38 castaglia Exp $
+ * $Id: netaddr.c,v 1.65 2008/03/20 21:56:30 castaglia Exp $
  */
 
 #include "conf.h"
@@ -31,7 +31,7 @@
 /* Define an IPv4 equivalent of the IN6_IS_ADDR_LOOPBACK macro. */
 #undef IN_IS_ADDR_LOOPBACK
 #define IN_IS_ADDR_LOOPBACK(a) \
-  ((((long int) (a)->s_addr) & 0xff000000) == 0x7f000000)
+  ((((unsigned long int) ntohl((a)->s_addr)) & 0xff000000) == 0x7f000000)
 
 static pr_netaddr_t sess_local_addr;
 static int have_sess_local_addr = FALSE;
@@ -43,6 +43,83 @@ static int have_sess_remote_addr = FALSE;
 /* Do reverse DNS lookups? */
 static int reverse_dns = 1;
 
+/* Use IPv6? */
+#ifdef PR_USE_IPV6
+static int use_ipv6 = TRUE;
+#else
+static int use_ipv6 = FALSE;
+#endif /* PR_USE_IPV6 */
+
+static pool *netaddr_pool = NULL;
+static pr_table_t *netaddr_tab = NULL;
+
+static const char *trace_channel = "dns";
+
+/* Netaddr cache management */
+static pr_netaddr_t *netaddr_cache_get(pool *p, const char *name) {
+  pr_netaddr_t *res = NULL;
+
+  if (netaddr_tab) {
+    void *v = pr_table_get(netaddr_tab, name, NULL);
+    if (v) {
+      res = v;
+      pr_trace_msg(trace_channel, 4,
+        "using IP address '%s' from netaddr cache for name '%s'",
+        pr_netaddr_get_ipstr(res), name);
+
+      /* We return a copy of the cache's netaddr_t, if the caller provided
+       * a pool for duplication.
+       */
+      if (p) {
+        return pr_netaddr_dup(p, res);
+      }
+
+      return res;
+
+    } else {
+      pr_trace_msg(trace_channel, 2,
+        "no IP address found in netaddr cache for name '%s'", name);
+    }
+  }
+
+  return NULL;
+}
+
+static void netaddr_cache_set(const char *name, pr_netaddr_t *na) {
+  if (netaddr_tab) {
+    int count = 0;
+    void *v = NULL;
+
+    /* We store an internal copy of the netaddr_t in the cache. */
+    v = pr_netaddr_dup(netaddr_pool, na);
+
+    count = pr_table_exists(netaddr_tab, name);
+    if (count <= 0) {
+      if (pr_table_add(netaddr_tab, pstrdup(netaddr_pool, name), v,
+          sizeof(pr_netaddr_t *)) < 0) {
+        pr_trace_msg(trace_channel, 3,
+          "error adding IP address '%s' for name '%s' to the netaddr cache: %s",
+          pr_netaddr_get_ipstr(na), name, strerror(errno));
+
+      } else {
+        pr_trace_msg(trace_channel, 5,
+          "stashed IP address '%s' for name '%s' in the netaddr cache",
+          pr_netaddr_get_ipstr(v), name);
+      }
+
+    } else {
+      if (pr_table_set(netaddr_tab, pstrdup(netaddr_pool, name), v,
+          sizeof(pr_netaddr_t *)) < 0) {
+        pr_trace_msg(trace_channel, 3,
+          "error setting IP address '%s' for name '%s' to the netaddr "
+          "cache: %s", pr_netaddr_get_ipstr(na), name, strerror(errno));
+      }
+    }
+  }
+
+  return;
+}
+
 /* Provide replacements for needed functions. */
 
 #if !defined(HAVE_GETNAMEINFO) || defined(PR_USE_GETNAMEINFO)
@@ -219,8 +296,7 @@ int pr_inet_pton(int af, const char *src, void *dst) {
 }
 #endif /* !HAVE_INET_PTON */
 
-#ifdef HAVE_GETHOSTBYNAME2
-static void *get_v4inaddr(pr_netaddr_t *na) {
+static void *get_v4inaddr(const pr_netaddr_t *na) {
 
   /* This function is specifically for IPv4 clients (when gethostbyname2(2) is
    * present) that have an IPv4-mapped IPv6 address, when performing reverse
@@ -234,7 +310,39 @@ static void *get_v4inaddr(pr_netaddr_t *na) {
 
   return (((char *) pr_netaddr_get_inaddr(na)) + 12);
 }
-#endif /* HAVE_GETHOSTBYNAME2 */
+
+/* Validate anything returned from the 'outside', since it's untrusted
+ * information.
+ */
+char *pr_netaddr_validate_dns_str(char *buf) {
+  char *p;
+
+  if (buf == NULL) {
+    errno = EINVAL;
+    return NULL;
+  }
+
+  /* Validate anything returned from a DNS. */
+  for (p = buf; p && *p; p++) {
+
+    /* Per RFC requirements, these are all that are valid from a DNS. */
+    if (!isalnum((int) *p) &&
+        *p != '.' &&
+        *p != '-'
+#ifdef PR_USE_IPV6
+        && *p != ':'
+#endif /* PR_USE_IPV6 */
+        ) {
+
+      /* We set it to _ because we know that's an invalid, yet safe, option
+       * for a DNS entry.
+       */
+      *p = '_';
+    }
+  }
+
+  return buf;
+}
 
 int pr_netaddr_set_reverse_dns(int enable) {
   int old_enable = reverse_dns;
@@ -271,6 +379,16 @@ pr_netaddr_t *pr_netaddr_dup(pool *p, pr_netaddr_t *na) {
   pr_netaddr_set_family(dup_na, pr_netaddr_get_family(na));
   pr_netaddr_set_sockaddr(dup_na, pr_netaddr_get_sockaddr(na));  
 
+  if (na->na_have_ipstr) {
+    sstrncpy(dup_na->na_ipstr, na->na_ipstr, sizeof(dup_na->na_ipstr));
+    dup_na->na_have_ipstr = 1;
+  }
+
+  if (na->na_have_dnsstr) {
+    sstrncpy(dup_na->na_dnsstr, na->na_dnsstr, sizeof(dup_na->na_dnsstr));
+    dup_na->na_have_dnsstr = 1;
+  }
+
   return dup_na;
 }
 
@@ -278,17 +396,31 @@ pr_netaddr_t *pr_netaddr_get_addr(pool *p, const char *name,
     array_header **addrs) {
 
   struct sockaddr_in v4;
-#ifdef PR_USE_IPV6
-  struct sockaddr_in6 v6;
-#endif /* PR_USE_IPV6 */
   pr_netaddr_t *na = NULL;
   int res;
 
-  if (p == NULL || name == NULL) {
+  if (p == NULL ||
+      name == NULL) {
     errno = EINVAL;
     return NULL;
   }
 
+  pr_trace_msg(trace_channel, 10, "resolving name '%s' to IP address",
+    name);
+
+  /* First, check our cache to see if this name has already been
+   * resolved.  We only want to use the cache, though, if the caller did not
+   * provide the `addrs' pointer, indicating that the caller wants to know
+   * about any additional addresses for the given name.  The netaddr cache
+   * is a simple cache, hidden from callers, and thus is unable to populate
+   * that `addrs' pointer if the name is in the cache.
+   */
+  if (addrs == NULL) {
+    na = netaddr_cache_get(p, name);
+    if (na)
+      return na;
+  }
+
   /* Attempt to translate the given name into a pr_netaddr_t using
    * pr_inet_pton() first.
    *
@@ -302,23 +434,29 @@ pr_netaddr_t *pr_netaddr_get_addr(pool *p, const char *name,
   na = (pr_netaddr_t *) pcalloc(p, sizeof(pr_netaddr_t));
 
 #ifdef PR_USE_IPV6
-  memset(&v6, 0, sizeof(v6));
-  v6.sin6_family = AF_INET6;
+  if (use_ipv6) {
+    struct sockaddr_in6 v6;
+    memset(&v6, 0, sizeof(v6));
+    v6.sin6_family = AF_INET6;
 
 # ifdef SIN6_LEN
-  v6.sin6_len = sizeof(struct sockaddr_in6);
+    v6.sin6_len = sizeof(struct sockaddr_in6);
 # endif /* SIN6_LEN */
 
-  res = pr_inet_pton(AF_INET6, name, &v6.sin6_addr);
-  if (res > 0) {
-    pr_netaddr_set_family(na, AF_INET6);
-    pr_netaddr_set_sockaddr(na, (struct sockaddr *) &v6);
-    if (addrs)
-      *addrs = NULL;
+    res = pr_inet_pton(AF_INET6, name, &v6.sin6_addr);
+    if (res > 0) {
+      pr_netaddr_set_family(na, AF_INET6);
+      pr_netaddr_set_sockaddr(na, (struct sockaddr *) &v6);
+      if (addrs)
+        *addrs = NULL;
 
-    pr_log_debug(DEBUG10, "'%s' resolved to IPv6 address %s", name,
-      pr_netaddr_get_ipstr(na));
-    return na;
+      pr_trace_msg(trace_channel, 7, "'%s' resolved to IPv6 address %s", name,
+        pr_netaddr_get_ipstr(na));
+
+      netaddr_cache_set(name, na);
+      netaddr_cache_set(pr_netaddr_get_ipstr(na), na);
+      return na;
+    }
   }
 #endif /* PR_USE_IPV6 */
 
@@ -336,8 +474,11 @@ pr_netaddr_t *pr_netaddr_get_addr(pool *p, const char *name,
     if (addrs)
       *addrs = NULL;
 
-    pr_log_debug(DEBUG10, "'%s' resolved to IPv4 address %s", name,
+    pr_trace_msg(trace_channel, 7, "'%s' resolved to IPv4 address %s", name,
       pr_netaddr_get_ipstr(na));
+
+    netaddr_cache_set(name, na);
+    netaddr_cache_set(pr_netaddr_get_ipstr(na), na);
     return na;
 
   } else if (res == 0) {
@@ -357,12 +498,20 @@ pr_netaddr_t *pr_netaddr_get_addr(pool *p, const char *name,
     hints.ai_family = AF_INET;
     hints.ai_socktype = SOCK_STREAM;
 
-    pr_log_debug(DEBUG10,
+    pr_trace_msg(trace_channel, 7,
       "attempting to resolve '%s' to IPv4 address via DNS", name);
     gai_res = pr_getaddrinfo(name, NULL, &hints, &info);
     if (gai_res != 0) {
-      pr_log_pri(PR_LOG_INFO, "IPv4 getaddrinfo '%s' error: %s", name,
-        gai_res != EAI_SYSTEM ? pr_gai_strerror(gai_res) : strerror(errno));
+      if (gai_res != EAI_SYSTEM) {
+        pr_trace_msg(trace_channel, 1, "IPv4 getaddrinfo '%s' error: %s",
+          name, pr_gai_strerror(gai_res));
+
+      } else {
+        pr_trace_msg(trace_channel, 1,
+          "IPv4 getaddrinfo '%s' system error: [%d] %s", name,
+          errno, strerror(errno));
+      }
+
       return NULL;
     }
 
@@ -371,15 +520,18 @@ pr_netaddr_t *pr_netaddr_get_addr(pool *p, const char *name,
       pr_netaddr_set_family(na, info->ai_family);
       pr_netaddr_set_sockaddr(na, info->ai_addr);
 
-      pr_log_debug(DEBUG10, "resolved '%s' to %s address %s", name,
+      pr_trace_msg(trace_channel, 7, "resolved '%s' to %s address %s", name,
         info->ai_family == AF_INET ? "IPv4" : "IPv6",
         pr_netaddr_get_ipstr(na));
 
+      netaddr_cache_set(name, na);
+      netaddr_cache_set(pr_netaddr_get_ipstr(na), na);
+
       pr_freeaddrinfo(info);
     }
 
 #ifdef PR_USE_IPV6
-    if (addrs) {
+    if (use_ipv6 && addrs) {
       /* Do the call again, this time for IPv6 addresses.
        *
        * We make two separate getaddrinfo(3) calls, rather than one
@@ -398,12 +550,20 @@ pr_netaddr_t *pr_netaddr_get_addr(pool *p, const char *name,
       hints.ai_family = AF_INET6;
       hints.ai_socktype = SOCK_STREAM;
 
-      pr_log_debug(DEBUG10,
+      pr_trace_msg(trace_channel, 7,
         "attempting to resolve '%s' to IPv6 address via DNS", name);
       gai_res = pr_getaddrinfo(name, NULL, &hints, &info);
       if (gai_res != 0) {
-        pr_log_pri(PR_LOG_INFO, "IPv6 getaddrinfo '%s' error: %s", name,
-          gai_res != EAI_SYSTEM ? pr_gai_strerror(gai_res) : strerror(errno));
+        if (gai_res != EAI_SYSTEM) {
+          pr_trace_msg(trace_channel, 1, "IPv6 getaddrinfo '%s' error: %s",
+            name, pr_gai_strerror(gai_res));
+
+        } else {
+          pr_trace_msg(trace_channel, 1, 
+            "IPv6 getaddrinfo '%s' system error: [%d] %s", name,
+            errno, strerror(errno));
+        }
+
         return na;
       }
 
@@ -417,7 +577,7 @@ pr_netaddr_t *pr_netaddr_get_addr(pool *p, const char *name,
         pr_netaddr_set_family(*elt, info->ai_family);
         pr_netaddr_set_sockaddr(*elt, info->ai_addr);
 
-        pr_log_debug(DEBUG10, "resolved '%s' to %s address %s", name,
+        pr_trace_msg(trace_channel, 7, "resolved '%s' to %s address %s", name,
           info->ai_family == AF_INET ? "IPv4" : "IPv6",
           pr_netaddr_get_ipstr(*elt));
 
@@ -429,7 +589,8 @@ pr_netaddr_t *pr_netaddr_get_addr(pool *p, const char *name,
     return na;
   }
 
-  pr_log_debug(DEBUG10, "failed to resolve '%s' to an IP address", name);
+  pr_trace_msg(trace_channel, 8, "failed to resolve '%s' to an IP address",
+    name);
   return NULL;
 }
 
@@ -456,8 +617,10 @@ int pr_netaddr_set_family(pr_netaddr_t *na, int family) {
 
 #ifdef PR_USE_IPV6
     case AF_INET6:
-      na->na_addr.v6.sin6_family = AF_INET6;
-      break;
+      if (use_ipv6) {
+        na->na_addr.v6.sin6_family = AF_INET6;
+        break;
+      }
 #endif /* PR_USE_IPV6 */
 
     default:
@@ -485,7 +648,8 @@ size_t pr_netaddr_get_sockaddr_len(const pr_netaddr_t *na) {
  
 #ifdef PR_USE_IPV6
     case AF_INET6:
-      return sizeof(struct sockaddr_in6);
+      if (use_ipv6)
+        return sizeof(struct sockaddr_in6);
 #endif /* PR_USE_IPV6 */   
   }
 
@@ -505,7 +669,8 @@ size_t pr_netaddr_get_inaddr_len(const pr_netaddr_t *na) {
 
 #ifdef PR_USE_IPV6
     case AF_INET6:
-      return sizeof(struct in6_addr);
+      if (use_ipv6)
+        return sizeof(struct in6_addr);
 #endif /* PR_USE_IPV6 */
   }
 
@@ -525,7 +690,8 @@ struct sockaddr *pr_netaddr_get_sockaddr(const pr_netaddr_t *na) {
 
 #ifdef PR_USE_IPV6
     case AF_INET6:
-      return (struct sockaddr *) &na->na_addr.v6;
+      if (use_ipv6)
+        return (struct sockaddr *) &na->na_addr.v6;
 #endif /* PR_USE_IPV6 */
   }
 
@@ -547,8 +713,10 @@ int pr_netaddr_set_sockaddr(pr_netaddr_t *na, struct sockaddr *addr) {
 
 #ifdef PR_USE_IPV6
     case AF_INET6:
-      memcpy(&(na->na_addr.v6), addr, sizeof(struct sockaddr_in6));
-      return 0;
+      if (use_ipv6) {
+        memcpy(&(na->na_addr.v6), addr, sizeof(struct sockaddr_in6));
+        return 0;
+      }
 #endif /* PR_USE_IPV6 */
   }
 
@@ -576,12 +744,14 @@ int pr_netaddr_set_sockaddr_any(pr_netaddr_t *na) {
 
 #ifdef PR_USE_IPV6
     case AF_INET6:
-      na->na_addr.v6.sin6_family = AF_INET6;
+      if (use_ipv6) {
+        na->na_addr.v6.sin6_family = AF_INET6;
 #ifdef SIN6_LEN
-      na->na_addr.v6.sin6_len = sizeof(struct sockaddr_in6);
+        na->na_addr.v6.sin6_len = sizeof(struct sockaddr_in6);
 #endif /* SIN6_LEN */
-      memcpy(&na->na_addr.v6.sin6_addr, &in6addr_any, sizeof(struct in6_addr));
-      return 0;
+        memcpy(&na->na_addr.v6.sin6_addr, &in6addr_any, sizeof(struct in6_addr));
+        return 0;
+      }
 #endif /* PR_USE_IPV6 */
   }
 
@@ -601,7 +771,8 @@ void *pr_netaddr_get_inaddr(const pr_netaddr_t *na) {
 
 #ifdef PR_USE_IPV6
     case AF_INET6:
-      return (void *) &na->na_addr.v6.sin6_addr;
+      if (use_ipv6)
+        return (void *) &na->na_addr.v6.sin6_addr;
 #endif /* PR_USE_IPV6 */
   }
 
@@ -621,7 +792,8 @@ unsigned int pr_netaddr_get_port(const pr_netaddr_t *na) {
 
 #ifdef PR_USE_IPV6
     case AF_INET6:
-      return na->na_addr.v6.sin6_port;
+      if (use_ipv6)
+        return na->na_addr.v6.sin6_port;
 #endif /* PR_USE_IPV6 */
   }
 
@@ -642,16 +814,22 @@ int pr_netaddr_set_port(pr_netaddr_t *na, unsigned int port) {
 
 #ifdef PR_USE_IPV6
     case AF_INET6:
-      na->na_addr.v6.sin6_port = port;
-      return 0;
+      if (use_ipv6) {
+        na->na_addr.v6.sin6_port = port;
+        return 0;
+      }
 #endif /* PR_USE_IPV6 */
   }
 
   errno = EPERM;
-  return 0;
+  return -1;
 }
 
 int pr_netaddr_cmp(const pr_netaddr_t *na1, const pr_netaddr_t *na2) {
+  pool *tmp_pool = NULL;
+  pr_netaddr_t *a, *b;
+  int res;
+
   if (na1 && !na2)
     return 1;
 
@@ -662,29 +840,119 @@ int pr_netaddr_cmp(const pr_netaddr_t *na1, const pr_netaddr_t *na2) {
     return 0;
 
   if (pr_netaddr_get_family(na1) != pr_netaddr_get_family(na2)) {
-    /* Cannot compare addresses from different families. */
-    errno = EINVAL;
-    return -1;
+
+    /* Cannot compare addresses from different families, unless one
+     * of the netaddrs has an AF_INET family, and the other has an
+     * AF_INET6 family AND is an IPv4-mapped IPv6 address.
+     */
+
+    if (pr_netaddr_is_v4mappedv6(na1) != TRUE &&
+        pr_netaddr_is_v4mappedv6(na2) != TRUE) {
+      errno = EINVAL;
+      return -1;
+    }
+
+    if (pr_netaddr_is_v4mappedv6(na1) == TRUE) {
+      tmp_pool = make_sub_pool(permanent_pool);
+
+      pr_trace_msg(trace_channel, 5, "addr '%s' is an IPv4-mapped IPv6 address",
+        pr_netaddr_get_ipstr((pr_netaddr_t *) na1));
+
+      /* This case means that na1 is an IPv4-mapped IPv6 address, and
+       * na2 is an IPv4 address.
+       */
+      a = pr_netaddr_alloc(tmp_pool);
+      pr_netaddr_set_family(a, AF_INET);
+      pr_netaddr_set_port(a, pr_netaddr_get_port(na1));
+      memcpy(&a->na_addr.v4.sin_addr, get_v4inaddr(na1),
+        sizeof(struct in_addr));
+
+      b = (pr_netaddr_t *) na2;
+
+      pr_trace_msg(trace_channel, 6, "comparing IPv4 address '%s' against "
+        "IPv4-mapped IPv6 address '%s'", pr_netaddr_get_ipstr(b),
+        pr_netaddr_get_ipstr(a));
+
+    } else if (pr_netaddr_is_v4mappedv6(na2) == TRUE) {
+      tmp_pool = make_sub_pool(permanent_pool);
+
+      pr_trace_msg(trace_channel, 5, "addr '%s' is an IPv4-mapped IPv6 address",
+        pr_netaddr_get_ipstr((pr_netaddr_t *) na2));
+
+      /* This case means that na is an IPv4 address, and na2 is an
+       * IPv4-mapped IPv6 address.
+       */
+      a = (pr_netaddr_t *) na1;
+
+      b = pr_netaddr_alloc(tmp_pool);
+      pr_netaddr_set_family(b, AF_INET);
+      pr_netaddr_set_port(b, pr_netaddr_get_port(na2));
+      memcpy(&b->na_addr.v4.sin_addr, get_v4inaddr(na2),
+        sizeof(struct in_addr));
+
+      pr_trace_msg(trace_channel, 6, "comparing IPv4 address '%s' against "
+        "IPv4-mapped IPv6 address '%s'", pr_netaddr_get_ipstr(a),
+        pr_netaddr_get_ipstr(b));
+
+    } else {
+      a = (pr_netaddr_t *) na1;
+      b = (pr_netaddr_t *) na2;
+    }
+
+  } else {
+    a = (pr_netaddr_t *) na1;
+    b = (pr_netaddr_t *) na2;
   }
 
-  switch (pr_netaddr_get_family(na1)) {
+  switch (pr_netaddr_get_family(a)) {
     case AF_INET:
-      return memcmp(&na1->na_addr.v4.sin_addr, &na2->na_addr.v4.sin_addr,
+      res = memcmp(&a->na_addr.v4.sin_addr, &b->na_addr.v4.sin_addr,
         sizeof(struct in_addr));
 
+      if (res != 0) {
+        pr_trace_msg(trace_channel, 4, "addr %s does not match addr %s",
+          pr_netaddr_get_ipstr(a), pr_netaddr_get_ipstr(b));
+      }
+
+      if (tmp_pool) {
+        destroy_pool(tmp_pool);
+        tmp_pool = NULL;
+      }
+
+      return res;
+
 #ifdef PR_USE_IPV6
     case AF_INET6:
-      return memcmp(&na1->na_addr.v6.sin6_addr, &na2->na_addr.v6.sin6_addr,
-        sizeof(struct in6_addr));
+      if (use_ipv6) {
+        res = memcmp(&a->na_addr.v6.sin6_addr, &b->na_addr.v6.sin6_addr,
+          sizeof(struct in6_addr));
+
+        if (res != 0) {
+          pr_trace_msg(trace_channel, 4, "addr %s does not match addr %s",
+            pr_netaddr_get_ipstr(a), pr_netaddr_get_ipstr(b));
+        }
+
+        if (tmp_pool) {
+          destroy_pool(tmp_pool);
+          tmp_pool = NULL;
+        }
+
+        return res;
+      }
 #endif /* PR_USE_IPV6 */
   }
 
+  if (tmp_pool)
+    destroy_pool(tmp_pool);
+
   errno = EPERM;
   return -1;
 }
 
 int pr_netaddr_ncmp(const pr_netaddr_t *na1, const pr_netaddr_t *na2,
     unsigned int bitlen) {
+  pool *tmp_pool = NULL;
+  pr_netaddr_t *a, *b;
   unsigned int nbytes, nbits;
   const unsigned char *in1, *in2;
 
@@ -698,12 +966,65 @@ int pr_netaddr_ncmp(const pr_netaddr_t *na1, const pr_netaddr_t *na2,
     return 0;
 
   if (pr_netaddr_get_family(na1) != pr_netaddr_get_family(na2)) {
-    /* Cannot compare addresses from different families. */
-    errno = EINVAL;
-    return -1;
+
+    /* Cannot compare addresses from different families, unless one
+     * of the netaddrs has an AF_INET family, and the other has an
+     * AF_INET6 family AND is an IPv4-mapped IPv6 address.
+     */
+
+    if (pr_netaddr_is_v4mappedv6(na1) != TRUE &&
+        pr_netaddr_is_v4mappedv6(na2) != TRUE) {
+      errno = EINVAL;
+      return -1;
+    }
+
+    if (pr_netaddr_is_v4mappedv6(na1) == TRUE) {
+      tmp_pool = make_sub_pool(permanent_pool);
+
+      /* This case means that na1 is an IPv4-mapped IPv6 address, and
+       * na2 is an IPv4 address.
+       */
+      a = pr_netaddr_alloc(tmp_pool);
+      pr_netaddr_set_family(a, AF_INET);
+      pr_netaddr_set_port(a, pr_netaddr_get_port(na1));
+      memcpy(&a->na_addr.v4.sin_addr, get_v4inaddr(na1),
+        sizeof(struct in_addr));
+
+      b = (pr_netaddr_t *) na2;
+
+      pr_trace_msg(trace_channel, 6, "comparing IPv4 address '%s' against "
+        "IPv4-mapped IPv6 address '%s'", pr_netaddr_get_ipstr(b),
+        pr_netaddr_get_ipstr(a));
+
+    } else if (pr_netaddr_is_v4mappedv6(na2) == TRUE) {
+      tmp_pool = make_sub_pool(permanent_pool);
+
+      /* This case means that na is an IPv4 address, and na2 is an
+       * IPv4-mapped IPv6 address.
+       */
+      a = (pr_netaddr_t *) na1;
+
+      b = pr_netaddr_alloc(tmp_pool);
+      pr_netaddr_set_family(b, AF_INET);
+      pr_netaddr_set_port(b, pr_netaddr_get_port(na2));
+      memcpy(&b->na_addr.v4.sin_addr, get_v4inaddr(na2),
+        sizeof(struct in_addr));
+
+      pr_trace_msg(trace_channel, 6, "comparing IPv4 address '%s' against "
+        "IPv4-mapped IPv6 address '%s'", pr_netaddr_get_ipstr(a),
+        pr_netaddr_get_ipstr(b));
+
+    } else {
+      a = (pr_netaddr_t *) na1;
+      b = (pr_netaddr_t *) na2;
+    }
+
+  } else {
+    a = (pr_netaddr_t *) na1;
+    b = (pr_netaddr_t *) na2;
   }
 
-  switch (pr_netaddr_get_family(na1)) {
+  switch (pr_netaddr_get_family(a)) {
     case AF_INET: {
       /* Make sure that the given number of bits is not more than supported
        * for IPv4 addresses (32).
@@ -718,15 +1039,17 @@ int pr_netaddr_ncmp(const pr_netaddr_t *na1, const pr_netaddr_t *na2,
 
 #ifdef PR_USE_IPV6
     case AF_INET6: {
-      /* Make sure that the given number of bits is not more than supported
-       * for IPv6 addresses (128).
-       */
-      if (bitlen > 128) {
-        errno = EINVAL;
-        return -1;
-      }
+      if (use_ipv6) {
+        /* Make sure that the given number of bits is not more than supported
+         * for IPv6 addresses (128).
+         */
+        if (bitlen > 128) {
+          errno = EINVAL;
+          return -1;
+        }
 
-      break;
+        break;
+      }
     }
 #endif /* PR_USE_IPV6 */
 
@@ -736,8 +1059,8 @@ int pr_netaddr_ncmp(const pr_netaddr_t *na1, const pr_netaddr_t *na2,
   }
 
   /* Retrieve pointers to the contained in_addrs. */
-  in1 = (const unsigned char *) pr_netaddr_get_inaddr(na1);
-  in2 = (const unsigned char *) pr_netaddr_get_inaddr(na2);
+  in1 = (const unsigned char *) pr_netaddr_get_inaddr(a);
+  in2 = (const unsigned char *) pr_netaddr_get_inaddr(b);
 
   /* Determine the number of bytes, and leftover bits, in the given
    * bit length.
@@ -750,8 +1073,12 @@ int pr_netaddr_ncmp(const pr_netaddr_t *na1, const pr_netaddr_t *na2,
     int res = memcmp(in1, in2, nbytes);
 
     /* No need to continue comparing the addresses if they differ already. */
-    if (res != 0)
+    if (res != 0) {
+      if (tmp_pool)
+        destroy_pool(tmp_pool);
+
       return res;
+    }
   }
 
   /* Next, compare the remaining bits in the addresses. */
@@ -765,13 +1092,24 @@ int pr_netaddr_ncmp(const pr_netaddr_t *na1, const pr_netaddr_t *na2,
     /* Build up a mask covering the bits left to be checked. */
     mask = (0xff << (8 - nbits)) & 0xff;
 
-    if ((in1byte & mask) > (in2byte & mask))
+    if ((in1byte & mask) > (in2byte & mask)) {
+      if (tmp_pool)
+        destroy_pool(tmp_pool);
+
       return 1;
+    }
+
+    if ((in1byte & mask) < (in2byte & mask)) {
+      if (tmp_pool)
+        destroy_pool(tmp_pool);
 
-    if ((in1byte & mask) < (in2byte & mask))
       return -1;
+    }
   }
 
+  if (tmp_pool)
+    destroy_pool(tmp_pool);
+
   /* If we've made it this far, the addresses match, for the given bit
    * length.
    */
@@ -797,21 +1135,55 @@ int pr_netaddr_fnmatch(pr_netaddr_t *na, const char *pattern, int flags) {
   if (flags & PR_NETADDR_MATCH_DNS) {
     const char *dnsstr = pr_netaddr_get_dnsstr(na);
 
-    pr_log_debug(DEBUG6, "comparing DNS name '%s' to pattern '%s'", dnsstr,
-      pattern);
-    if (pr_fnmatch(pattern, dnsstr, match_flags) == 0)
+    if (pr_fnmatch(pattern, dnsstr, match_flags) == 0) {
+      pr_trace_msg(trace_channel, 6, "DNS name '%s' matches pattern '%s'",
+        dnsstr, pattern);
       return TRUE;
+    }
   }
 
   if (flags & PR_NETADDR_MATCH_IP) {
     const char *ipstr = pr_netaddr_get_ipstr(na);
 
-    pr_log_debug(DEBUG6, "comparing IP address '%s' to pattern '%s'", ipstr,
-      pattern);
-    if (pr_fnmatch(pattern, ipstr, match_flags) == 0)
+    if (pr_fnmatch(pattern, ipstr, match_flags) == 0) {
+      pr_trace_msg(trace_channel, 6, "DNS name '%s' matches pattern '%s'",
+        ipstr, pattern);
       return TRUE;
+    }
+
+    /* If the address is an IPv4-mapped IPv6 address, get the IPv4 address
+     * and try to match that against the configured glob pattern.
+     */
+    if (pr_netaddr_is_v4mappedv6(na) == TRUE) {
+      pool *tmp_pool;
+      pr_netaddr_t *a;
+
+      pr_trace_msg(trace_channel, 5, "addr '%s' is an IPv4-mapped IPv6 address",
+        ipstr);
+
+      tmp_pool = make_sub_pool(permanent_pool);
+      a = pr_netaddr_alloc(tmp_pool);
+      pr_netaddr_set_family(a, AF_INET);
+      pr_netaddr_set_port(a, pr_netaddr_get_port(na));
+      memcpy(&a->na_addr.v4.sin_addr, get_v4inaddr(na),
+        sizeof(struct in_addr));
+
+      ipstr = pr_netaddr_get_ipstr(a);
+
+      if (pr_fnmatch(pattern, ipstr, match_flags) == 0) {
+        pr_trace_msg(trace_channel, 6, "DNS name '%s' matches pattern '%s'",
+          ipstr, pattern);
+
+        destroy_pool(tmp_pool);
+        return TRUE;
+      }
+
+      destroy_pool(tmp_pool);
+    }
   }
 
+  pr_trace_msg(trace_channel, 4, "addr %s does not match pattern '%s'",
+    pr_netaddr_get_ipstr(na), pattern);
   return FALSE;
 }
 
@@ -839,8 +1211,14 @@ const char *pr_netaddr_get_ipstr(pr_netaddr_t *na) {
     pr_netaddr_get_sockaddr_len(na), buf, sizeof(buf), NULL, 0, NI_NUMERICHOST);
 
   if (res != 0) {
-    pr_log_pri(PR_LOG_NOTICE, "getnameinfo error: %s",
-      res != EAI_SYSTEM ? pr_gai_strerror(res) : strerror(errno));
+    if (res != EAI_SYSTEM) {
+      pr_log_pri(PR_LOG_INFO, "getnameinfo error: %s", pr_gai_strerror(res));
+
+    } else {
+      pr_log_pri(PR_LOG_INFO, "getnameinfo system error: [%d] %s",
+        errno, strerror(errno));
+    }
+
     return NULL;
   }
 
@@ -861,12 +1239,23 @@ const char *pr_netaddr_get_ipstr(pr_netaddr_t *na) {
 const char *pr_netaddr_get_dnsstr(pr_netaddr_t *na) {
   char *name = NULL;
   char buf[256];
+  pr_netaddr_t *cache = NULL;
 
   if (!na) {
     errno = EINVAL;
     return NULL;
   }
 
+  cache = netaddr_cache_get(NULL, pr_netaddr_get_ipstr(na));
+  if (cache &&
+      cache->na_have_dnsstr) {
+    memset(na->na_dnsstr, '\0', sizeof(na->na_dnsstr));
+    sstrncpy(na->na_dnsstr, cache->na_dnsstr, sizeof(na->na_dnsstr));
+    na->na_have_dnsstr = TRUE;
+
+    return na->na_dnsstr;
+  }
+
   /* If this pr_netaddr_t has already been resolved to an DNS string, return the
    * cached string.
    */
@@ -876,9 +1265,14 @@ const char *pr_netaddr_get_dnsstr(pr_netaddr_t *na) {
   if (reverse_dns) {
     int res = 0;
 
+    pr_trace_msg(trace_channel, 3,
+      "verifying DNS name for IP address %s via reverse DNS lookup",
+      pr_netaddr_get_ipstr(na));
+
     memset(buf, '\0', sizeof(buf));
     res = pr_getnameinfo(pr_netaddr_get_sockaddr(na),
       pr_netaddr_get_sockaddr_len(na), buf, sizeof(buf), NULL, 0, NI_NAMEREQD);
+    buf[sizeof(buf)-1] = '\0';
 
     if (res == 0) {
       char **checkaddr;
@@ -899,11 +1293,27 @@ const char *pr_netaddr_get_dnsstr(pr_netaddr_t *na) {
 #endif /* HAVE_GETHOSTBYNAME2 */
 
       if (hent != NULL) {
+
+        pr_trace_msg(trace_channel, 10,
+          "checking addresses associated with host '%s'",
+          hent->h_name ? hent->h_name : "(null)");
+        
         switch (hent->h_addrtype) {
           case AF_INET:
             if (family == AF_INET) {
+
               for (checkaddr = hent->h_addr_list; *checkaddr; ++checkaddr) {
                 if (memcmp(*checkaddr, inaddr, hent->h_length) == 0) {
+                  char **alias;
+
+                  for (alias = hent->h_aliases; *alias; ++alias) {
+                    if (hent->h_name) {
+                      pr_trace_msg(trace_channel, 10,
+                        "host '%s' has alias '%s'", hent->h_name, *alias);
+                      netaddr_cache_set(*alias, na);
+                    }
+                  }
+
                   ok = TRUE;
                   break;
                 }
@@ -913,9 +1323,20 @@ const char *pr_netaddr_get_dnsstr(pr_netaddr_t *na) {
 
 #ifdef PR_USE_IPV6
           case AF_INET6:
-            if (family == AF_INET6) {
+            if (use_ipv6 && family == AF_INET6) {
+
               for (checkaddr = hent->h_addr_list; *checkaddr; ++checkaddr) {
                 if (memcmp(*checkaddr, inaddr, hent->h_length) == 0) {
+                  char **alias;
+
+                  for (alias = hent->h_aliases; *alias; ++alias) {
+                    if (hent->h_name) {
+                      pr_trace_msg(trace_channel, 10,
+                        "host '%s' has alias '%s'", hent->h_name, *alias);
+                      netaddr_cache_set(*alias, na);
+                    }
+                  }
+
                   ok = TRUE;
                   break;
                 }
@@ -925,7 +1346,18 @@ const char *pr_netaddr_get_dnsstr(pr_netaddr_t *na) {
 #endif /* PR_USE_IPV6 */
         }
 
-        name = ok ? buf : NULL;
+        if (ok) {
+          name = buf;
+          pr_trace_msg(trace_channel, 8,
+            "using DNS name '%s' for IP address '%s'", name,
+            pr_netaddr_get_ipstr(na));
+
+        } else {
+          name = NULL;
+          pr_trace_msg(trace_channel, 8,
+            "unable to verify any DNS names for IP address '%s'",
+            pr_netaddr_get_ipstr(na));
+        }
 
       } else
         pr_log_debug(DEBUG1, "notice: unable to resolve '%s': %s", buf,
@@ -936,10 +1368,12 @@ const char *pr_netaddr_get_dnsstr(pr_netaddr_t *na) {
     pr_log_debug(DEBUG10,
       "UseReverseDNS off, returning IP address instead of DNS name");
 
-  if (!name)
-    name = (char *) pr_netaddr_get_ipstr(na);
+  if (name) {
+    name = pr_netaddr_validate_dns_str(name);
 
-  name = pr_inet_validate(name);
+  } else {
+    name = (char *) pr_netaddr_get_ipstr(na);
+  }
 
   /* Copy the string into the pr_netaddr_t cache as well, so we only
    * have to do this once for this pr_netaddr_t.
@@ -948,6 +1382,10 @@ const char *pr_netaddr_get_dnsstr(pr_netaddr_t *na) {
   sstrncpy(na->na_dnsstr, name, sizeof(na->na_dnsstr));
   na->na_have_dnsstr = TRUE;
 
+  /* Update the netaddr object in the cache with the resolved DNS names. */
+  netaddr_cache_set(name, na);
+  netaddr_cache_set(pr_netaddr_get_ipstr(na), na);
+
   return na->na_dnsstr;
 }
 
@@ -956,6 +1394,11 @@ const char *pr_netaddr_get_localaddr_str(pool *p) {
   char buf[256] = {'\0'};
   struct hostent *host;
 
+  if (p == NULL) {
+    errno = EINVAL;
+    return NULL;
+  }
+
   if (gethostname(buf, sizeof(buf)-1) != -1) {
     buf[sizeof(buf)-1] = '\0';
 
@@ -966,15 +1409,16 @@ const char *pr_netaddr_get_localaddr_str(pool *p) {
     host = gethostbyname(buf);
 
     if (host)
-      return pr_inet_validate(pstrdup(p, host->h_name));
+      return pr_netaddr_validate_dns_str(pstrdup(p, host->h_name));
 
-    return pr_inet_validate(pstrdup(p, buf));
+    return pr_netaddr_validate_dns_str(pstrdup(p, buf));
   }
 
+  pr_trace_msg(trace_channel, 1, "gethostname(2) error: %s", strerror(errno));
   return NULL;
 }
 
-int pr_netaddr_loopback(const pr_netaddr_t *na) {
+int pr_netaddr_is_loopback(const pr_netaddr_t *na) {
   if (!na) {
     errno = EINVAL;
     return -1;
@@ -987,7 +1431,6 @@ int pr_netaddr_loopback(const pr_netaddr_t *na) {
 
 #ifdef PR_USE_IPV6
     case AF_INET6:
-
       /* XXX *sigh* Different platforms implement the IN6_IS_ADDR macros
        * differently.  For example, on Linux, those macros expect to operate
        * on s6_addr32, while on Solaris, the macros operate on struct in6_addr.
@@ -1059,15 +1502,23 @@ int pr_netaddr_is_v4mappedv6(const pr_netaddr_t *na) {
       return -1;
 
 #ifdef PR_USE_IPV6
-    case AF_INET6:
+    case AF_INET6: {
+      int res;
+
+      if (!use_ipv6) {
+        errno = EINVAL;
+        return -1;
+      }
 
 # ifndef LINUX
-      return IN6_IS_ADDR_V4MAPPED(
+      res = IN6_IS_ADDR_V4MAPPED(
         (struct in6_addr *) pr_netaddr_get_inaddr(na));
 # else
-      return IN6_IS_ADDR_V4MAPPED(
+      res = IN6_IS_ADDR_V4MAPPED(
         ((struct in6_addr *) pr_netaddr_get_inaddr(na))->s6_addr32);
 # endif
+      return res;
+    }
 #endif /* PR_USE_IPV6 */
   }
 
@@ -1118,3 +1569,46 @@ void pr_netaddr_set_sess_addrs(void) {
   sstrncpy(sess_remote_name, session.c->remote_name, sizeof(sess_remote_name));
   have_sess_remote_addr = TRUE;
 }
+
+unsigned char pr_netaddr_use_ipv6(void) {
+  if (use_ipv6)
+    return TRUE;
+
+  return FALSE;
+}
+
+void pr_netaddr_disable_ipv6(void) {
+#ifdef PR_USE_IPV6
+  use_ipv6 = 0;
+#endif /* PR_USE_IPV6 */
+}
+
+void pr_netaddr_enable_ipv6(void) {
+#ifdef PR_USE_IPV6
+  use_ipv6 = 1;
+#endif /* PR_USE_IPV6 */
+}
+
+void pr_netaddr_clear_cache(void) {
+  if (netaddr_tab) {
+    pr_trace_msg(trace_channel, 5, "emptying netaddr cache");
+    (void) pr_table_empty(netaddr_tab);
+    (void) pr_table_free(netaddr_tab);
+
+    /* Allocate a fresh table. */
+    netaddr_tab = pr_table_alloc(netaddr_pool, 0);
+  }
+}
+
+void init_netaddr(void) {
+  if (netaddr_pool) {
+    pr_netaddr_clear_cache();
+    destroy_pool(netaddr_pool);
+    netaddr_pool = NULL;
+  }
+
+  netaddr_pool = make_sub_pool(permanent_pool);
+  pr_pool_tag(netaddr_pool, "Netaddr API");
+
+  netaddr_tab = pr_table_alloc(netaddr_pool, 0);
+}
