@@ -1,6 +1,6 @@
 /*
  * ProFTPD - FTP server daemon
- * Copyright (c) 2004-2007 The ProFTPD Project team
+ * Copyright (c) 2004-2008 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -22,9 +22,8 @@
  * for OpenSSL in the source distribution.
  */
 
-/*
- * Configuration parser
- * $Id: parser.c,v 1.16 2007/03/28 03:49:33 castaglia Exp $
+/* Configuration parser
+ * $Id: parser.c,v 1.18 2008/09/04 01:49:53 castaglia Exp $
  */
 
 #include "conf.h"
@@ -96,11 +95,6 @@ static struct config_src *add_config_source(pr_fh_t *fh) {
   return cs;
 }
 
-static void add_server_ctxt(server_rec *s) {
-  parser_curr_server = (server_rec **) push_array(parser_servstack);
-  *parser_curr_server = s;
-}
-
 static char *get_config_word(pool *p, char *word) {
 
   /* Should this word be replaced with a value from the environment?
@@ -334,7 +328,7 @@ int pr_parser_parse_file(pool *p, const char *path, config_rec *start,
           "dispatching directive '%s' to module mod_%s", conftab->directive,
           conftab->m->name);
 
-        mr = call_module(conftab->m, conftab->handler, cmd);
+        mr = pr_module_call(conftab->m, conftab->handler, cmd);
         if (mr != NULL) {
           if (MODRET_ISERROR(mr)) {
 
@@ -615,12 +609,15 @@ server_rec *pr_parser_server_ctxt_open(const char *addrstr) {
    */
   xaset_insert_end(*parser_server_list, (xasetmember_t *) s);
   s->set = *parser_server_list;
-  if (addrstr)
+  if (addrstr) {
     s->ServerAddress = pstrdup(s->pool, addrstr);
+  }
 
   /* Default server port */
   s->ServerPort = pr_inet_getservport(s->pool, "ftp", "tcp");
 
-  add_server_ctxt(s);
+  parser_curr_server = (server_rec **) push_array(parser_servstack);
+  *parser_curr_server = s;
+
   return s;
 }
