@@ -1,6 +1,6 @@
 /*
  * ProFTPD - FTP server daemon
- * Copyright (c) 2003 The ProFTPD Project team
+ * Copyright (c) 2003-2008 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -23,7 +23,7 @@
  */
 
 /* ProFTPD xferlog(5) logging support.
- * $Id: xferlog.c,v 1.3 2005/06/14 01:23:45 castaglia Exp $
+ * $Id: xferlog.c,v 1.4 2008/01/05 06:23:10 castaglia Exp $
  */
 
 #include "conf.h"
@@ -61,6 +61,8 @@ int xferlog_write(long xfertime, const char *remhost, off_t fsize, char *fname,
     char abort_flag) {
 
   char buf[LOGBUFFER_SIZE] = {'\0'}, fbuf[LOGBUFFER_SIZE] = {'\0'};
+  int have_ident = FALSE;
+  char *rfc1413_ident = NULL;
   register unsigned int i = 0;
 
   if (xferlogfd == -1 || !remhost || !user || !fname)
@@ -72,6 +74,24 @@ int xferlog_write(long xfertime, const char *remhost, off_t fsize, char *fname,
   }
   fbuf[i] = '\0';
 
+  rfc1413_ident = pr_table_get(session.notes, "mod_ident.rfc1413-ident", NULL);
+  if (rfc1413_ident) {
+    have_ident = TRUE;
+
+    /* If the retrieved identity is "UNKNOWN", then change the string to be
+     * "*", since "*" is to be logged in the xferlog, as per the doc, when
+     * the authenticated user ID is not available.
+     */
+    if (strcmp(rfc1413_ident, "UNKNOWN") == 0)
+      rfc1413_ident = "*";
+
+  } else {
+    /* If an authenticated user ID is not available, log "*", as per the
+     * xferlog man page/spec.
+     */
+    rfc1413_ident = "*";
+  }
+
   snprintf(buf, sizeof(buf),
     "%s %ld %s %" PR_LU " %s %c _ %c %c %s ftp %c %s %c\n",
       pr_strtime(time(NULL)),
@@ -83,9 +103,8 @@ int xferlog_write(long xfertime, const char *remhost, off_t fsize, char *fname,
       direction,
       access_mode,
       user,
-      session.ident_lookups == TRUE ? '1' : '0',
-      (session.ident_lookups == TRUE && strcmp(session.ident_user,
-        "UNKNOWN")) ? session.ident_user : "*",
+      have_ident ? '1' : '0',
+      rfc1413_ident,
       abort_flag);
 
   buf[sizeof(buf)-1] = '\0';
