@@ -2,7 +2,7 @@
  * ProFTPD - FTP server daemon
  * Copyright (c) 1997, 1998 Public Flood Software
  * Copyright (c) 1999, 2000 MacGyver aka Habeeb J. Dihu <macgyver@tos.net>
- * Copyright (c) 2001, 2002, 2003 The ProFTPD Project team
+ * Copyright (c) 2001-2008 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -24,10 +24,9 @@
  * the source code for OpenSSL in the source distribution.
  */
 
-/* Shows a count of "who" is online via proftpd.  Uses the /var/run/proftpd*
- * log files.
+/* Shows a count of "who" is online via proftpd.  Uses the scoreboard file.
  *
- * $Id: ftpwho.c,v 1.18 2003/02/24 18:39:52 castaglia Exp $
+ * $Id: ftpwho.c,v 1.26 2008/02/10 02:29:22 castaglia Exp $
  */
 
 #include "utils.h"
@@ -41,9 +40,7 @@ struct scoreboard_class {
 #define OF_COMPAT		0x001
 #define OF_ONELINE		0x002
 
-static const char *config_filename = CONFIG_FILE_PATH;
-
-char *util_sstrncpy(char *, const char *, size_t);
+static const char *config_filename = PR_CONFIG_FILE_PATH;
 
 static char *percent_complete(off_t size, off_t done) {
   static char sbuf[32];
@@ -88,94 +85,53 @@ static const char *show_time(time_t *i) {
   return sbuf;
 }
 
-/* scan_config_file() is a kludge for 1.2 which does a very simplistic attempt
- * at determining what the "ScoreboardFile" directive is set to.  It will be
- * replaced in 1.3 with the abstracted configure system (hopefully).
- */
-static void scan_config_file(void) {
-  FILE *fp = NULL;
-  char buf[PR_TUNABLE_BUFFER_SIZE] = {'\0'};
-  char *cp, *file = NULL;
-
-  if (!config_filename || (fp = fopen(config_filename,"r")) == NULL)
-    return;
-
-  while (!file && fgets(buf, sizeof(buf) - 1, fp)) {
-    int i = strlen(buf);
-
-    if (i && buf[i - 1] == '\n')
-      buf[i-1] = '\0';
-
-    for (cp = buf; *cp && isspace((int) *cp); cp++);
-
-    if (*cp == '#' || !*cp)
-      continue;
-
-    i = strlen("ScoreboardFile");
-
-    if (strncasecmp(cp, "ScoreboardFile", i) != 0)
-      continue;
-
-    /* Found it! */
-    cp += i;
-
-    /* strip whitespace */
-    while (*cp && isspace((int) *cp))
-      cp++;
-
-    file = cp;
-
-    /* If the scoreboard file argument is quoted, dequote */
-    if (*cp == '"') {
-      char *src = cp;
-
-      cp++;
-      file++;
-
-      while (*++src) {
-        switch (*src) {
-          case '\\':
-            if (*++src)
-              *cp++ = *src;
-            break;
-
-          case '"':
-            src++;
-            break;
-
-          default:
-            *cp++ = *src;
-        }
-      }
-
-      *cp = '\0';
-    }
-  }
-
-  fclose(fp);
-
-  /* If we got something out of all this, go ahead and set it. */
-  if (file)
-    util_set_scoreboard(file);
-}
-
 static int check_scoreboard_file(void) {
-  struct stat sbuf;
+  struct stat st;
 
-  if (stat(util_get_scoreboard(), &sbuf) < 0)
+  if (stat(util_get_scoreboard(), &st) < 0)
     return -1;
 
   return 0;
 }
 
+static const char *show_uptime(time_t uptime_since) {
+  static char buf[128] = {'\0'};
+  time_t uptime_secs = time(NULL) - uptime_since;
+  int upminutes, uphours, updays;
+  int pos = 0;
+
+  memset(buf, '\0', sizeof(buf));
+
+  updays = (int) uptime_secs / (60 * 60 * 24);
+
+  if (updays)
+    pos += sprintf(buf + pos, "%d day%s, ", updays, (updays != 1) ? "s" : "");
+
+  upminutes = (int) uptime_secs / 60;
+
+  uphours = upminutes / 60;
+  uphours = uphours % 24;
+
+  upminutes = upminutes % 60;
+
+  if (uphours)
+    pos += sprintf(buf + pos, "%2d hr%s %02d min", uphours,
+      (uphours != 1) ? "s" : "", upminutes);
+  else
+    pos += sprintf(buf + pos, "%d min", upminutes);
+
+  return buf;
+}
+
 static struct option_help {
   const char *long_opt,*short_opt,*desc;
 } opts_help[] = {
-  { "--config", "-c", "specify full path to proftpd configuration file" },
-  { "--file", "-f", "specify full path to scoreboard file" },
-  { "--help", "-h", NULL },
-  { "--outform", "-o", "specify an output format" },
-  { "--verbose","-v", "display additional information for each connection" },
+  { "--config",	"-c",	"specify full path to proftpd configuration file" },
+  { "--file",	"-f",	"specify full path to scoreboard file" },
+  { "--help",	"-h",	NULL },
+  { "--outform","-o",	"specify an output format" },
+  { "--verbose","-v",	"display additional information for each connection" },
+  { "--server",	"-S",	"show users only for specified ServerName" },
   { NULL }
 };
 
@@ -186,6 +142,7 @@ static struct option opts[] = {
   { "help",    0, NULL, 'h' },
   { "outform", 1, NULL, 'o' },
   { "verbose", 0, NULL, 'v' },
+  { "server",  1, NULL, 'S' },
   { NULL,      0, NULL, 0   }
 };
 #endif /* HAVE_GETOPT_LONG */
@@ -196,7 +153,7 @@ static void show_usage(const char *progname, int exit_code) {
   printf("usage: %s [options]\n", progname);
   for (h = opts_help; h->long_opt; h++) {
 #ifdef HAVE_GETOPT_LONG
-    printf("  %s, %s\n", h->long_opt, h->short_opt);
+    printf("  %s, %s\n", h->short_opt, h->long_opt);
 #else /* HAVE_GETOPT_LONG */
     printf("  %s\n", h->short_opt);
 #endif
@@ -212,21 +169,24 @@ static void show_usage(const char *progname, int exit_code) {
 int main(int argc, char **argv) {
   pr_scoreboard_entry_t *score = NULL;
   pid_t mpid = 0;
+  time_t uptime = 0;
   unsigned int count = 0, total = 0;
   int c = 0, res = 0;
+  char *server_name = NULL;
   struct scoreboard_class classes[MAX_CLASSES];
   char *cp, *progname = *argv;
-  const char *cmdopts = "c:f:ho:v";
+  const char *cmdopts = "S:c:f:ho:v";
   unsigned char verbose = FALSE;
   unsigned long outform = 0;
 
   memset(classes, 0, MAX_CLASSES * sizeof(struct scoreboard_class));
 
-  if((cp = strrchr(progname,'/')) != NULL)
+  cp = strrchr(progname, '/');
+  if (cp != NULL)
     progname = cp+1;
 
   opterr = 0;
-  while((c =
+  while ((c =
 #ifdef HAVE_GETOPT_LONG
 	 getopt_long(argc, argv, cmdopts, opts, NULL)
 #else /* HAVE_GETOPT_LONG */
@@ -251,11 +211,11 @@ int main(int argc, char **argv) {
 
       case 'o':
         /* Check the given outform parameter. */
-        if (!strcmp(optarg, "compat")) {
+        if (strcmp(optarg, "compat") == 0) {
           outform |= OF_COMPAT;
           break;
 
-        } else if (!strcmp(optarg, "oneline")) {
+        } else if (strcmp(optarg, "oneline") == 0) {
           outform |= OF_ONELINE;
           break;
         }
@@ -263,9 +223,13 @@ int main(int argc, char **argv) {
         fprintf(stderr, "unknown outform value: '%s'\n", optarg);
         return 1;
 
+      case 'S':
+        server_name = strdup(optarg);
+        break;
+
       case '?':
-        fprintf(stderr, "unknown option: %c\n", (char)optopt);
-        show_usage(progname,1);
+        fprintf(stderr, "unknown option: %c\n", (char) optopt);
+        show_usage(progname, 1);
     }
   }
 
@@ -273,7 +237,9 @@ int main(int argc, char **argv) {
    * incorrect, try the config file kludge.
    */
   if (check_scoreboard_file() < 0) {
-    scan_config_file();
+    const char *file = util_scan_config(config_filename, "ScoreboardFile");
+    if (file)
+      util_set_scoreboard(file);
 
     if (check_scoreboard_file() < 0) {
       fprintf(stderr, "%s: %s\n", util_get_scoreboard(), strerror(errno));
@@ -284,7 +250,7 @@ int main(int argc, char **argv) {
   }
 
   count = 0;
-  if ((res = util_open_scoreboard(O_RDONLY, &mpid)) < 0) {
+  if ((res = util_open_scoreboard(O_RDONLY)) < 0) {
     switch (res) {
       case -1:
         fprintf(stderr, "unable to open scoreboard: %s\n", strerror(errno));
@@ -304,16 +270,28 @@ int main(int argc, char **argv) {
     }
   }
 
+  mpid = util_scoreboard_get_daemon_pid();
+  uptime = util_scoreboard_get_daemon_uptime();
+
   if (!mpid)
     printf("inetd FTP daemon:\n");
 
   else
-    printf("standalone FTP daemon [%d]:\n",(int) mpid);
+    printf("standalone FTP daemon [%u], up for %s\n", (unsigned int) mpid,
+      show_uptime(uptime));
 
-  while ((score = util_scoreboard_read_entry()) != NULL) {
-    unsigned char uploading = FALSE;
+  if (server_name)
+    printf("ProFTPD Server '%s'\n", server_name);
+
+  while ((score = util_scoreboard_entry_read()) != NULL) {
+    int downloading = FALSE, uploading = FALSE;
     register unsigned int i = 0;
 
+    /* If a ServerName was given, skip unless the scoreboard entry matches. */
+    if (server_name &&
+        strcmp(server_name, score->sce_server_label) != 0)
+      continue;
+
     if (!count++) {
       if (total)
         printf("   -  %d user%s\n\n", total, total > 1 ? "s" : "");
@@ -337,26 +315,36 @@ int main(int argc, char **argv) {
 
     total++;
 
-    if (strncmp(score->sce_cmd, "STOR", 4) == 0 ||
-        strncmp(score->sce_cmd, "STOU", 4) == 0 ||
-        strncmp(score->sce_cmd, "APPE", 4) == 0)
-      uploading = TRUE;
+    if (strcmp(score->sce_cmd, "RETR") == 0) {
+      downloading = TRUE;
+
+    } else {
+      if (strcmp(score->sce_cmd, "STOR") == 0 ||
+          strcmp(score->sce_cmd, "STOU") == 0 ||
+          strcmp(score->sce_cmd, "APPE") == 0) {
+        uploading = TRUE;
+      }
+    }
 
     if (outform & OF_COMPAT) {
-      if (score->sce_xfer_size) {
-        if (uploading)
-          printf("%5d %-6s (n/a) %s\n", (int) score->sce_pid,
-            show_time(&score->sce_begin_idle), score->sce_cmd);
-
-        else
-          printf("%5d %-6s (%s%%) %s\n", (int) score->sce_pid,
+      if ((downloading || uploading) &&
+          score->sce_xfer_size > 0) {
+        if (downloading)
+          printf("%5d %-6s (%s%%) %s %s\n", (int) score->sce_pid,
             show_time(&score->sce_begin_idle),
             percent_complete(score->sce_xfer_size, score->sce_xfer_done),
-            score->sce_cmd);
+            score->sce_cmd, score->sce_cmd_arg);
 
-      } else
-        printf("%5d %-6s %s\n", (int) score->sce_pid,
-          show_time(&score->sce_begin_idle), score->sce_cmd);
+        else
+          printf("%5d %-6s (n/a) %s %s\n", (int) score->sce_pid,
+            show_time(&score->sce_begin_idle), score->sce_cmd,
+            score->sce_cmd_arg);
+
+      } else {
+        printf("%5d %-6s %s %s\n", (int) score->sce_pid,
+          show_time(&score->sce_begin_idle), score->sce_cmd,
+          score->sce_cmd_arg);
+      }
 
       if (verbose) {
         if (score->sce_client_addr[0])
@@ -375,7 +363,7 @@ int main(int argc, char **argv) {
     if (strcmp(score->sce_user, "(none)")) {
 
       /* Is the client idle? */
-      if (!strcmp(score->sce_cmd, "(idle)")) {
+      if (strcmp(score->sce_cmd, "idle") == 0) {
 
         /* These printf() calls needs to be split up, as show_time() returns
          * a pointer to a static buffer, and pushing two invocations onto
@@ -389,16 +377,17 @@ int main(int argc, char **argv) {
           printf("\n");
 
       } else {
-        if (uploading)
-          printf("%5d %-8s [%6s] (n/a) %s", (int) score->sce_pid,
-            score->sce_user, show_time(&score->sce_begin_session),
-            score->sce_cmd);
-
-        else
-          printf("%5d %-8s [%6s] (%3s%%) %s", (int) score->sce_pid,
+        if (downloading) {
+          printf("%5d %-8s [%6s] (%3s%%) %s %s", (int) score->sce_pid,
             score->sce_user, show_time(&score->sce_begin_session),
             percent_complete(score->sce_xfer_size, score->sce_xfer_done),
-            score->sce_cmd);
+            score->sce_cmd, score->sce_cmd_arg);
+
+        } else {
+          printf("%5d %-8s [%6s] (n/a) %s %s", (int) score->sce_pid,
+            score->sce_user, show_time(&score->sce_begin_session),
+            score->sce_cmd, score->sce_cmd_arg);
+        }
 
         if (verbose) {
           printf("%sKB/s: %3.2f%s",
