@@ -2,7 +2,7 @@
  * ProFTPD - FTP server daemon
  * Copyright (c) 1997, 1998 Public Flood Software
  * Copyright (c) 1999, 2000 MacGyver aka Habeeb J. Dihu <macgyver@tos.net>
- * Copyright (c) 2001, 2002, 2003 The ProFTPD Project team
+ * Copyright (c) 2001-2008 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -26,11 +26,11 @@
 
 /* Scoreboard routines.
  *
- * $Id: utils.h,v 1.18 2003/10/21 01:44:47 castaglia Exp $
+ * $Id: utils.h,v 1.23 2008/02/10 02:29:22 castaglia Exp $
  */
 
-#ifndef UTIL_SCOREBOARD_H
-#define UTIL_SCOREBOARD_H
+#ifndef UTILS_UTILS_H
+#define UTILS_UTILS_H
 
 #include "config.h"
 #include "version.h"
@@ -43,7 +43,6 @@
 #include <stdlib.h>
 #include <string.h>
 #include <time.h>
-#include <sys/types.h>
 
 #ifdef HAVE_GETOPT_H
 #  include <getopt.h>
@@ -120,11 +119,11 @@ typedef struct {
   int sce_server_port;
   char sce_server_addr[80], sce_server_label[32];
 
-#ifdef USE_IPV6
+#ifdef PR_USE_IPV6
   char sce_client_addr[INET6_ADDRSTRLEN];
 #else
   char sce_client_addr[INET_ADDRSTRLEN];
-#endif /* USE_IPV6 */
+#endif /* PR_USE_IPV6 */
   char sce_client_name[PR_TUNABLE_SCOREBOARD_BUFFER_SIZE];
 
   char sce_class[32];
@@ -150,10 +149,12 @@ char *util_sstrncpy(char *, const char *, size_t);
 const char *util_get_scoreboard(void);
 int util_set_scoreboard(const char *);
 
+const char *util_scan_config(const char *, const char *);
+
 int util_close_scoreboard(void);
 int util_open_scoreboard(int);
 pid_t util_scoreboard_get_daemon_pid(void);
 time_t util_scoreboard_get_daemon_uptime(void);
-pr_scoreboard_entry_t *util_scoreboard_read_entry(void);
+pr_scoreboard_entry_t *util_scoreboard_entry_read(void);
 
-#endif /* UTIL_SCOREBOARD_H */
+#endif /* UTILS_UTILS_H */
